/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */
package org.biobeans.modules.biocore.actions;

import java.util.logging.Logger;
import java.util.prefs.PreferenceChangeEvent;
import java.util.prefs.PreferenceChangeListener;
import org.biobeans.modules.biocore.options.BioSettings;
import org.openide.nodes.Node;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.CookieAction;
import org.openide.util.actions.SystemAction;

/** Base action for operations on sequences.
 * 
 * The implementation decides if the action is enabled or not. 
 * Subclasses decide on the type of sequences this action is 
 * applicable by providing a combination of BioJava SeqIOConstant formats.
 * {@link #getSequenceFormat()}. The length of the sequence for which this action
 * is enabled can be set using getMinLength() and getMaxLength().
 *
 * @author  George Waldon
 */
public abstract class SequenceAction extends CookieAction {

    private static final Logger LOGGER = Logger.getLogger(SequenceAction.class.getName());
    
    private static final PreferenceChangeListener settingListener;
    
    /* Default maximum sequence length equals to 100,000.
     */
    public static int DEFAULTMAXLENGTH = 100000;

    public static int MAXLENGTH = DEFAULTMAXLENGTH;

    static {
        // Sequence setting listener
        settingListener = new PreferenceChangeListener() {
            @Override
            public void preferenceChange(PreferenceChangeEvent evt) {
                if (BioSettings.PROP_PARSING_LIMIT.equals(evt.getKey())) {
                    MAXLENGTH = BioSettings.getDefault().getParsingLimit();
                }
            }
        };
        BioSettings.attachListernerWeakly(settingListener);
        MAXLENGTH = BioSettings.getDefault().getParsingLimit();
    }

    @Override
    public org.openide.util.HelpCtx getHelpCtx() {
        return HelpCtx.DEFAULT_HELP;
    }    
 
    @Override
    protected boolean asynchronous() {
        return false;
    }

    @Override
    public boolean isEnabled() {
        // Clear caching, see {@link NodeAction}
        putProperty(SystemAction.PROP_ENABLED, null, false);
        boolean b = super.isEnabled();
        return b;
    }

    @Override
    protected boolean enable(Node[] activatedNodes) {
        boolean b = super.enable(activatedNodes);
        if (b == false) {
            return b;
        }
        return doEnable(activatedNodes);
    }

    /** Add further enabling options based on the activatedNodes.
     */
    abstract protected boolean doEnable(Node[] activatedNodes);

    @Override
    protected boolean surviveFocusChange() {
        return false;
    }

    /** Getter for the type of sequences that this action supports.
     *
     * @return int type of sequence in BioJava constants.
     */
    abstract protected int getSequenceFormat();

    /** Get the minimum length for a sequence to have this action enabled. By
     * default, return a value of -1 for which the length criteria is not 
     * considered.
     */
    protected int getMinLength() {
        return -1;
    }

    /** Get the maximum length for a sequence to have this action enabled. By
     * default, return DEFAULTMAXLENGTH; otherwise returns a value of -1 for 
     * which the length criteria is not considered.
     */
    protected int getMaxLength() {
        return MAXLENGTH;
    }

    /** Create a window title.
     */
    protected String windowTitle(String window) {
        String name = window;
        if (name == null || name.length() == 0) {
            name = NbBundle.getMessage(SequenceAction.class, "LBL_untitled");
        }
        name = name.trim();
        int index = name.indexOf(' ');
        if (index > 0) {
            name = name.substring(0, index);
        }
        return name;
    }
}
