/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */

package org.biobeans.modules.biocore.actions;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import org.biojava.bio.seq.Sequence;
import org.biobeans.modules.biocore.sequences.SeqProvider;
import org.biobeans.modules.biocore.sequences.SeqType;
import org.openide.nodes.Node;
import org.openide.util.Lookup;
import org.openide.util.RequestProcessor;
import org.openide.util.actions.CookieAction;

/** Base action for operations on sequences that are enabled by the SeqProvider cookie.
 * Cookies are resolved by SeqProvider position, the highest being selected. This
 * allows e.g. to access a feature sequence while the DataObject provides the parent
 * sequence.
 * 
 * The implementation decides if the action is enabled or not. 
 * Subclasses decide on the type of sequences this action is 
 * applicable by providing a combination of BioJava SeqIOConstant formats.
 * {@link #getSequenceFormat()}. The length of the sequence for which this action
 * is enabled can be set using getMinLength() and getMaxLength().
 *
 */
public abstract class SequenceProviderAction extends SequenceAction {
    
    /** SeqProvider */
    @Override
    protected Class[] cookieClasses() {
        return new Class[] {SeqProvider.class};
    }
    
    protected int getHighestPosition(Node[] activatedNodes) {
        int highpos = SeqProvider.DataPosition;
        for (Node activatedNode : activatedNodes) {
            Lookup context = activatedNode.getLookup();
            Collection<? extends SeqProvider> c = context.lookupAll(SeqProvider.class);
            for (SeqProvider provider : c) {
                int position = provider.getSeqPosition();
                if(position>highpos)
                    highpos = position;
            }
        }
        return highpos;
    }

    @Override
    protected boolean enable(Node[] activatedNodes) {
        if (activatedNodes.length == 0) {
            return false;
        }
        boolean b = super.enable(activatedNodes);
        // bypass filter on actual nodes
        //if (b == false) {
        //    return b;
        //}
        return doEnable(activatedNodes);
    }

    @Override
    protected boolean doEnable (Node[] activatedNodes) {
        int active = 0;
        int supported = 0;
        int highpos = getHighestPosition(activatedNodes);
        
        // In most case, we only need one sequence to be enabled and
        // for long, multiple files, checking the first sequence is enough.
        if(mode()==MODE_ANY) {
            for (Node activatedNode : activatedNodes) {
                Lookup context = activatedNode.getLookup();
                Collection<? extends SeqProvider> c = context.lookupAll(SeqProvider.class);
                Collection<SeqProvider> sp = new ArrayList<>();
                for (SeqProvider provider : c) {
                    int position = provider.getSeqPosition();
                    if(position==highpos) {
                        if(!sp.contains(provider)) {
                            sp.add(provider);
                            active++;
                            break;
                        }
                    }
                }
                supported += resolveSupported(sp);
                if(supported == 1) {
                    return true;
                }
                break;
            }
            active = 0;
            supported = 0;
        } 
        
        for (Node activatedNode : activatedNodes) {
            Lookup context = activatedNode.getLookup();
            Collection<? extends SeqProvider> c = context.lookupAll(SeqProvider.class);
            Collection<SeqProvider> sp = new ArrayList<>();
            for (SeqProvider provider : c) {
                int position = provider.getSeqPosition();
                if(position==highpos) {
                    if(!sp.contains(provider)) {
                        sp.add(provider);
                        active++;
                    }
                }
            }
            supported += resolveSupported(sp);
        }
        
        if (supported == 0)
            return false;
        
        if(highpos == SeqProvider.ComponentPosition) {
            if(supported!=1)
                return false;
        }
        
        int mode = mode ();
        return
            ((mode & CookieAction.MODE_ONE) != 0) ||
            (((mode & CookieAction.MODE_ALL) != 0) && (supported == active)) ||
            (((mode & CookieAction.MODE_EXACTLY_ONE) != 0) && (active == 1)) ||
            (((mode & CookieAction.MODE_SOME) != 0) && (supported < active));
        //return
        //    ((mode & CookieAction.MODE_ONE) != 0) ||
        //    (((mode & CookieAction.MODE_ALL) != 0) && (supported == activatedNodes.length)) ||
        //    (((mode & CookieAction.MODE_EXACTLY_ONE) != 0) && (activatedNodes.length == 1)) ||
        //    (((mode & CookieAction.MODE_SOME) != 0) && (supported < activatedNodes.length));
    }
    
    /** Indicate whether a particular sequence is supported by this action.
     *
     * @param seq the sequence to test; cannot be null
     * @return true if this action supports the sequence
     */
    protected boolean isSupported(Sequence seq) {
        assert seq != null:"There is no sequence to test for support";
        if(getMinLength()!=-1 && seq.length()<getMinLength())
            return false;
        if(getMaxLength()!=-1 && seq.length()>getMaxLength())
            return false;
        int type = SeqType.getFormatFromAlphabet(seq.getAlphabet());
        return (getSequenceFormat()&type) != 0;
    }
    
    /** Indicate whether a sequence array contains at least one sequence
     * supported by this action.
     *
     * @param seqarray the sequence array to search for supported sequences
     * @return true if the array contains at least one supported sequence
     */
    protected boolean isSupported(Sequence[] seqarray) {
        if(seqarray==null || seqarray.length==0)
            return false;
        for (Sequence seqarray1 : seqarray) {
            if (seqarray1 == null) {
                continue;
            }
            if (isSupported(seqarray1) == true) {
                return true;
            }
        }
        return false;
    }
    
    private int resolveSupported (Collection c) {
        int count = 0;
        Iterator it = c.iterator();
        while(it.hasNext()) {
            SeqProvider provider = (SeqProvider) it.next();
            Sequence[] sarr = provider.getSequences();
            if(isSupported(sarr)==true) count++;
        }
        return count;
    }    
     
    @Override
    protected void performAction(Node[] activatedNodes) {
        int highpos = getHighestPosition(activatedNodes);
        
        for (Node activatedNode : activatedNodes) {
            Lookup context = activatedNode.getLookup();
            Collection<? extends SeqProvider> c = context.lookupAll(SeqProvider.class);
            Collection<Sequence> seqcol = new ArrayList<>();
            for (SeqProvider provider : c) {
                int position = provider.getSeqPosition();
                if(position==highpos) {
                    Sequence[] sarr = provider.getSequences();
                    for(int i=0; i<sarr.length;i++) {
                        if(!seqcol.contains(sarr[i])) {
                            seqcol.add(sarr[i]);
                        }
                    }
                }
            }
            if(seqcol.size()>0) {
                Sequence[] seqs = seqcol.toArray(new Sequence[seqcol.size()]);
                invokeNodeAction(activatedNode,seqs,false);
            }
        }
    }
    
    static RequestProcessor RP = new RequestProcessor("Sequence Actions", 1);
    
    /** @param comp Indicate whether the action was invoked from a component.
     */
    final protected void invokeNodeAction (final Node node, final Sequence[] seqarray, final boolean comp) {
        if(!isSupported(seqarray))
            return;
        Runnable r = new Runnable() {
            @Override
            public void run() {
                performSeqAction(node,seqarray,comp);
            }
        };
        RP.post(r);
    }
    
    /** Actually do the job. This method is called asynchronously on a
     * specific request processor "Sequence Actions". The node might be null.
     * 
     * @param node The corresponding node; may be null (with ComponentSeqProvider).
     * @param seqarray an array of sequences to perform on.
     * @param comp Indicate whether this action was invoked from a component.
     */
    abstract protected void performSeqAction(Node node, Sequence[] seqarray, boolean comp);
 
}
