/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */
package org.biobeans.modules.biocore.seqfiles;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.PreferenceChangeEvent;
import java.util.prefs.PreferenceChangeListener;
import org.biobeans.modules.biocore.options.BioSettings;
import org.biobeans.modules.biocore.sequences.SeqProvider;
import org.biobeans.modules.biocore.sequences.SequenceCookie;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojavax.bio.seq.RichSequence;
import org.openide.filesystems.FileChangeAdapter;
import org.openide.filesystems.FileChangeListener;
import org.openide.filesystems.FileEvent;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import static org.openide.loaders.DataObject.PROP_NAME;
import static org.openide.loaders.DataObject.PROP_PRIMARY_FILE;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.CookieSet;
import org.openide.nodes.Node;
import org.openide.text.DataEditorSupport;

public abstract class SimpleSeqDataObject extends MultiDataObject implements CookieSet.Factory, SeqProvider {
    
    static final long serialVersionUID = 1L;

    private static final Logger LOGGER = Logger.getLogger(SimpleSeqDataObject.class.getName());

    // Maximum length for internal parsing
    private static int MAXLENGTH;

    transient private FileChangeListener fileChangeListener;
    transient private FileChangeListener fileChangeListenerHook;
    private PreferenceChangeListener settingListener;
    private transient boolean initialized;

    
    public SimpleSeqDataObject(FileObject pf, MultiFileLoader loader) throws DataObjectExistsException, IOException {
        super(pf, loader);
        init();
    }

    private void init() {
        CookieSet cookies = getCookieSet();
        cookies.add((Node.Cookie) DataEditorSupport.create(this, getPrimaryEntry(), cookies));

        SimpleSeqDataObject.PrimaryFileListener fl = new SimpleSeqDataObject.PrimaryFileListener();
	addPropertyChangeListener(fl);
        fileChangeListenerHook = fl;
	fileChangeListener = FileUtil.weakFileChangeListener(fileChangeListenerHook, getPrimaryEntry().getFile());
        addFileChangeListener(null);
        
        settingListener = new PreferenceChangeListener() {
            @Override
            public void preferenceChange(PreferenceChangeEvent evt) {
                if (BioSettings.PROP_PARSING_LIMIT.equals(evt.getKey())) {
                    MAXLENGTH = BioSettings.getDefault().getParsingLimit();
                    resetParsing();
                }
            }
        };
        BioSettings.attachListernerWeakly(settingListener);
        MAXLENGTH = BioSettings.getDefault().getParsingLimit();
        
        initialized = true;
    }
    
    protected abstract void resetParsing();
    
//    @Override
//    protected Node createNodeDelegate() {
//        return new DataNode(this, Children.LEAF, getLookup());
//    }

//    @Override
//    public Lookup getLookup() {
//        return getCookieSet().getLookup();
//    }

    @Override
    public <T extends Node.Cookie> T createCookie(Class<T> klass) {
        if (SequenceCookie.class.isAssignableFrom(klass)) {
            return klass.cast(this);
        } else if (SeqProvider.class.isAssignableFrom(klass)) {
            return klass.cast(this);
        }
        return null;
    }

    protected boolean canParse() {
        return getPrimaryEntry().getFile().getSize()<MAXLENGTH;
    }
    
    @Override
    public int getSeqPosition() {
        return SeqProvider.DataPosition;
    }

    @Override
    public Sequence[] getSequences() {
        if(!canParse()) {
            return new RichSequence[0];
        }
        try {
            return parseSequences();
        } catch (IOException | BioException ex) {
            LOGGER.log(Level.INFO, "", ex);
            return new RichSequence[0];
        }
    }
    
    abstract protected RichSequence[] parseSequences() throws IOException, BioException;

    
    /** Attaches a file change listener to the primary file.
     * @param previousPrimaryFile if not null.
     */
    private void addFileChangeListener(FileObject previousPrimaryFile) {
        if (previousPrimaryFile != null) {
            previousPrimaryFile.removeFileChangeListener(fileChangeListener);
        }
        getPrimaryEntry().getFile().addFileChangeListener(fileChangeListener);
    }
    
    protected void primaryFileChanged() {
        //Do nothing here
    }

    protected void primaryFileMoved(FileObject oldFile, FileObject newFile) {
        addFileChangeListener(oldFile);
    }

    private class PrimaryFileListener extends FileChangeAdapter implements PropertyChangeListener {
        @Override
        public void propertyChange(final PropertyChangeEvent evt) {
            if (!initialized)
                return;
            String propName = evt.getPropertyName();
            if (null != propName) switch (propName) {
                case PROP_PRIMARY_FILE:
                    primaryFileMoved((FileObject)evt.getOldValue(), (FileObject)evt.getNewValue());
                    break;
                case PROP_NAME:
                    primaryFileMoved(getPrimaryFile(), getPrimaryFile());
                    break;
            }
        }

        @Override
        public void fileChanged(FileEvent e) {
            if (!initialized)
                return;
            primaryFileChanged();
        }
    }
}
