/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */
package org.biobeans.modules.biocore.seqfiles.embl;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.logging.Logger;
import org.biobeans.modules.biocore.seqfiles.SimpleSeqDataObject;
import org.biobeans.modules.biocore.sequences.RichSequenceArrayIterator;
import org.biobeans.modules.biocore.sequences.SequenceCookie;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.SeqIOConstants;
import org.biojavax.RichObjectFactory;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiFileLoader;

/** Simple DataObject to represent an EMBL file. The sequences can be accessed
 * through either the SequenceCookie or the SeqProvider interface. Files
 * longer than 5,000,000 bytes are not parsed but sequences can still be accessed
 * on this DataObject through getSequenceIterator.
 *
 * @author George Waldon
 */
public class SimpleEMBLDataObject extends SimpleSeqDataObject implements SequenceCookie {

    static final long serialVersionUID = 1L;

    private static final Logger LOGGER = Logger.getLogger(SimpleEMBLDataObject.class.getName());

    //flag to prevent reparsing of unconsistent files
    private boolean noParsing = false;

    private volatile boolean modified = false;

    private transient RichSequence[] seqs;
    
    private final Object seqlock = new Object();

    public SimpleEMBLDataObject(FileObject pf, MultiFileLoader loader) throws DataObjectExistsException, IOException {
        super(pf, loader);
    }

    @Override
    public int getSequenceFormat() {
        return SeqIOConstants.DNA;
    }

    @Override
    public boolean canSetFormat() {
        return false;
    }

    @Override
    public void setSequenceFormat(int format, boolean noEventFiring) {
    }

    @Override
    public SequenceIterator getSequenceIterator() throws IOException, BioException {
        synchronized(seqlock) {
            if(!canParse()) {
                BufferedReader br = null;
                FileObject file = getPrimaryEntry().getFile();
                try {
                    br = new BufferedReader(new InputStreamReader(file.getInputStream()));
                } catch (FileNotFoundException ex) {
                    throw new IOException("EMBL file not found: " + ex, ex);
                }
                RichSequenceIterator rit = RichSequence.IOTools.readEMBLDNA(br, RichObjectFactory.getDefaultNamespace());
                return rit;
            } else {
                RichSequence[] ars = parseSequences();
                return new RichSequenceArrayIterator(ars);
            }
        }
    }

    @Override
    protected  RichSequence[] parseSequences() throws IOException, BioException {
        synchronized(seqlock) {
            if(noParsing==true && modified==false)
                return new RichSequence[0];
            boolean wasModified = modified;
            modified = false;
            if(wasModified==false && seqs!=null)
                return seqs;

            BufferedReader br = null;
            FileObject file = getPrimaryEntry().getFile();
            try {
                br = new BufferedReader(new InputStreamReader(file.getInputStream()));
            } catch (FileNotFoundException ex) {
                throw new IOException("EMBL file not found: " + ex, ex);
            }

            List<RichSequence> lrseq = new ArrayList<>();
            RichSequenceIterator rit = RichSequence.IOTools.readEMBLDNA(br, RichObjectFactory.getDefaultNamespace());
            try {
                while (rit.hasNext()) {
                    lrseq.add(rit.nextRichSequence());
                }
            } catch (NoSuchElementException ex) {
                noParsing = true;
            } catch (BioException bex) {
                noParsing = true;
                throw bex;
            }
            RichSequence[] ars = new RichSequence[lrseq.size()];
            seqs = lrseq.toArray(ars);
            return seqs;
        }
    }

    @Override
    protected void primaryFileChanged() {
        resetParsing();
    }
    
    @Override
    protected void resetParsing() {
        synchronized(seqlock) {
            noParsing = false;
            modified = true;
            seqs = null;
        }
    }
}
