/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */

package org.biobeans.modules.biocore.seqfiles.gck;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.NoSuchElementException;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.SeqIOConstants;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;
import org.biojavax.bio.seq.io.RichSequenceBuilderFactory;
import org.biojavax.bio.seq.io.RichSequenceFormat;
import org.biojavax.bio.seq.io.RichStreamReader;
import org.biobeans.modules.biocore.sequences.RichSequenceArrayIterator;
import org.biobeans.modules.biocore.sequences.SeqProvider;
import org.biobeans.modules.biocore.sequences.SequenceCookie;
import org.openide.filesystems.FileChangeAdapter;
import org.openide.filesystems.FileChangeListener;
import org.openide.filesystems.FileEvent;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.nodes.CookieSet;
import org.openide.nodes.Node;
import org.openide.nodes.Node.Cookie;
import org.openide.util.HelpCtx;
import org.openide.util.Lookup;

/** Represents a GCK object in the Repository.
 *
 * @author gwaldon
 */
public class GCKDataObject extends MultiDataObject
        implements CookieSet.Factory, SequenceCookie, SeqProvider {

    private static final Logger LOGGER = Logger.getLogger(GCKDataObject.class.getName());

    static final long serialVersionUID = 1L;

    /** WeakListener monitoring file modifications while the file is not
     *  opened in the Editor.
     */
    transient private FileChangeListener fileChangeListener;

    /** Keep the WeakListener alive.
     */
    transient private FileChangeListener fileChangeListenerHook;

    private transient boolean initialized;

    private transient RichSequence seq;

    private volatile boolean modified = false;

    public GCKDataObject(FileObject primaryFile, GCKDataLoader loader)
            throws DataObjectExistsException {
        super(primaryFile,loader);
        init();
    }

    private void init() {

        PrimaryFileListener fl = new PrimaryFileListener();
	addPropertyChangeListener(fl);
        fileChangeListenerHook = fl;
	fileChangeListener = FileUtil.weakFileChangeListener(fileChangeListenerHook, getPrimaryEntry().getFile());
        addFileChangeListener(null);
        initialized = true;
    }

    /** Attaches a file change listener to the primary file.
     * @param previousPrimaryFile if not null.
     */
    void addFileChangeListener(FileObject previousPrimaryFile) {
      if (previousPrimaryFile != null) {
        previousPrimaryFile.removeFileChangeListener(fileChangeListener);
      }
      getPrimaryEntry().getFile().addFileChangeListener(fileChangeListener);
    }

    protected void primaryFileMoved(FileObject oldFile, FileObject newFile) {
        addFileChangeListener(oldFile);
    }

    protected void primaryFileChanged() {
        modified = true;
    }

    // =============== Primary file monitoring ============
    private class PrimaryFileListener extends FileChangeAdapter implements PropertyChangeListener{
        @Override
        public void propertyChange(final PropertyChangeEvent evt) {
            if (!initialized)
                return;
            String propName = evt.getPropertyName();
            if (PROP_PRIMARY_FILE.equals(propName)) {
                primaryFileMoved((FileObject)evt.getOldValue(), (FileObject)evt.getNewValue());
            } else if (PROP_NAME.equals(propName)) {
                primaryFileMoved(getPrimaryFile(), getPrimaryFile());
            }
        }

        @Override
        public void fileChanged(FileEvent e) {
            if (!initialized)
                return;
            primaryFileChanged();
        }
    }

    @Override
    public HelpCtx getHelpCtx() {
        return HelpCtx.DEFAULT_HELP;
    }

    /** Provide node that represent the FastaDataObject
     */
    @Override
    protected Node createNodeDelegate () {
        return new GCKNode(this);
    }

    @Override
    public int getSequenceFormat() {
        return SeqIOConstants.DNA;
    }

    @Override
    public boolean canSetFormat() {
        return false;
    }

    @Override
    public void setSequenceFormat(int format, boolean noEventFiring) {
        LOGGER.info("Attempt to set the sequence format in GCK file");
    }

    @Override
    public int getSeqPosition() {
        return SeqProvider.DataPosition;
    }

    @Override
    public Sequence[] getSequences() {
        RichSequence rseq = parseFile();
        if(rseq==null)
            return new Sequence[0];
        else
            return new Sequence[] {rseq};
    }

    @Override
    public SequenceIterator getSequenceIterator() /*throws IOException, BioException*/ {
        Sequence[] seqarray = getSequences();
        RichSequence[] richarray = new RichSequence[seqarray.length];
        for(int i=0; i<seqarray.length;i++)
            richarray[i] = RichSequence.Tools.enrich(seqarray[i]);
        return new RichSequenceArrayIterator(richarray);
    }

    //flag to prevent constant reparsing of unconsistent files
    private boolean noParsing = false;

    private RichSequence parseFile() {
        if(noParsing==true)
            return null;
        boolean wasModified = modified;
        modified = false; //no return point
        if(wasModified==false && seq!=null)
            return seq;
        DataInputStream is = null;
        try {
            is = new DataInputStream(getPrimaryFile().getInputStream());
            RichSequenceFormat format = new GCKFormat();

            String name = getPrimaryFile().getName();
            name = name.replace(' ', '_');
            ((GCKFormat)format).setName(name);

            RichSequenceIterator seqit = new RichStreamReader(
                    new GCKFormat.DataReader(new InputStreamReader(is), is),
                    format,
                    RichSequence.IOTools.getDNAParser(),
                    RichSequenceBuilderFactory.FACTORY,
                    null);
            if (seqit.hasNext()) {
                seq = seqit.nextRichSequence();
            }
        } catch (NoSuchElementException | BioException | IOException ex) {
            noParsing = true;
        } finally {
            try {
                is.close();
            } catch (IOException ex) {
                LOGGER.log(Level.INFO, null, ex);
                noParsing = true;
            }
        }
        return seq;
    }

    @Override
    public Lookup getLookup() {
        return getCookieSet().getLookup();
    }

    @Override
    public <T extends Cookie> T createCookie(Class<T> klass) {
        if (SequenceCookie.class.isAssignableFrom(klass)) {
            return klass.cast(this);
        }
        if (SeqProvider.class.isAssignableFrom(klass)) {
            return klass.cast(this);
        }
        return null;
    }

}
