/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */

package org.biobeans.modules.biocore.seqfiles.raw;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.NoSuchElementException;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.modules.biocore.sequences.RichSequenceArrayIterator;
import org.biobeans.modules.biocore.sequences.SeqProvider;
import org.biobeans.modules.biocore.sequences.SequenceCookie;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.SeqIOConstants;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;
import org.biojavax.bio.seq.io.RichSequenceBuilderFactory;
import org.biojavax.bio.seq.io.RichSequenceFormat;
import org.biojavax.bio.seq.io.RichStreamReader;
import org.openide.filesystems.*;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.SaveAsCapable;
import org.openide.nodes.CookieSet;
import org.openide.nodes.Node;
import org.openide.nodes.Node.Cookie;
import org.openide.util.HelpCtx;
import org.openide.util.Lookup;

/** Represents a raw sequence object.
 *
 * @author George Waldon
 */
public class RawDataObject extends MultiDataObject
        implements CookieSet.Factory, SequenceCookie, SeqProvider {

    private static final Logger LOGGER = Logger.getLogger(RawDataObject.class.getName());

    static final long serialVersionUID = 1L;

    /** WeakListener monitoring file modifications while the file is not
     *  opened in the Editor.
     */
    transient private FileChangeListener fileChangeListener;

    transient private FileChangeListener fileChangeListenerHook;

    private transient boolean initialized;

    private transient RichSequence seq;

    private volatile boolean modified = false;

    private RawEditorSupport rawEditorSupport;

    private int sformat = SeqIOConstants.UNKNOWN;

    public RawDataObject(FileObject primaryFile, RawDataLoader loader)
            throws DataObjectExistsException {
        super(primaryFile,loader);
        init();
    }

    private void init() {
        CookieSet set = getCookieSet();
        set.add(RawEditorSupport.class,(CookieSet.Factory)this);
        set.assign(SaveAsCapable.class, new SaveAsCapable() {
            @Override
            public void saveAs( FileObject folder, String fileName ) throws IOException {
                RawEditorSupport es = getLookup().lookup( RawEditorSupport.class );
                es.saveAs( folder, fileName );
            }
        });

        PrimaryFileListener fl = new PrimaryFileListener();
	addPropertyChangeListener(fl);
        fileChangeListenerHook = fl;
	fileChangeListener = FileUtil.weakFileChangeListener(fileChangeListenerHook, getPrimaryEntry().getFile());
        addFileChangeListener(null);
        initialized = true;
    }

    /** Attaches a file change listener to the primary file.
     * @param previousPrimaryFile if not null.
     */
    void addFileChangeListener(FileObject previousPrimaryFile) {
      if (previousPrimaryFile != null) {
        previousPrimaryFile.removeFileChangeListener(fileChangeListener);
      }
      getPrimaryEntry().getFile().addFileChangeListener(fileChangeListener);
    }

    protected void primaryFileMoved(FileObject oldFile, FileObject newFile) {
        addFileChangeListener(oldFile);
    }

    protected void primaryFileChanged() {
        modified = true;
    }

    // =============== Primary file monitoring ============
    private class PrimaryFileListener extends FileChangeAdapter implements PropertyChangeListener {
        @Override
        public void propertyChange(final PropertyChangeEvent evt) {
            if (!initialized)
                return;
            String propName = evt.getPropertyName();
            if (null != propName) switch (propName) {
                case PROP_PRIMARY_FILE:
                    primaryFileMoved((FileObject)evt.getOldValue(), (FileObject)evt.getNewValue());
                    break;
                case PROP_NAME:
                    primaryFileMoved(getPrimaryFile(), getPrimaryFile());
                    break;
            }
        }

        @Override
        public void fileChanged(FileEvent e) {
            if (!initialized)
                return;
            primaryFileChanged();
        }
    }

    @Override
    public HelpCtx getHelpCtx() {
        return HelpCtx.DEFAULT_HELP;
    }

    /** Provide node that represent the RawDataObject
     */
    @Override
    protected Node createNodeDelegate () {
        return new RawNode(this);
    }

    @Override
    public int getSequenceFormat() {
        if (sformat == SeqIOConstants.UNKNOWN)
            getSequences(); // Trigger parsing if necessary
        return sformat;
    }

    @Override
    public boolean canSetFormat() {
        return false;
    }

    @Override
    public void setSequenceFormat(int format, boolean noEventFiring) {
        LOGGER.info("Attempt to set the sequence format in raw sequence file");
    }

    @Override
    public int getSeqPosition() {
        return SeqProvider.DataPosition;
    }

    @Override
    public Sequence[] getSequences() {
        RichSequence rseq = parseFile();
        if(rseq==null)
            return new Sequence[0];
        else
            return new Sequence[] {rseq};
    }

    @Override
    public SequenceIterator getSequenceIterator() /*throws IOException, BioException*/ {
        Sequence[] seqarray = getSequences();
        RichSequence[] richarray = new RichSequence[seqarray.length];
        for(int i=0; i<seqarray.length;i++)
            richarray[i] = RichSequence.Tools.enrich(seqarray[i]);
        return new RichSequenceArrayIterator(richarray);
    }

    //flag to prevent constant reparsing of unconsistent files
    private boolean noParsing = false;

    private synchronized RichSequence parseFile() {
        if(noParsing==true && modified==false)
            return null;
        boolean wasModified = modified;
        modified = false; //no return point
        if(wasModified==false && seq!=null)
            return seq;

        DataInputStream is = null;
        RichSequenceFormat format = (RichSequenceFormat) new RawFormat();
        String name = getPrimaryFile().getName();
        name = name.replace(' ', '_');
        ((RawFormat)format).setName(name);

        try {
            is = new DataInputStream(getPrimaryFile().getInputStream());
            RichSequenceIterator seqit = new RichStreamReader(
                    new BufferedReader(new InputStreamReader(is)),
                    format,
                    RichSequence.IOTools.getDNAParser(),
                    RichSequenceBuilderFactory.FACTORY,
                    null);
            if (seqit.hasNext()) {
                seq = seqit.nextRichSequence();
                sformat = SeqIOConstants.DNA;
            }
        } catch (NoSuchElementException | IOException ex) {
            noParsing = true;
        } catch (BioException ex) {
            try {
            is = new DataInputStream(getPrimaryFile().getInputStream());
            RichSequenceIterator seqit = new RichStreamReader(
                    new BufferedReader(new InputStreamReader(is)),
                    format,
                    RichSequence.IOTools.getProteinParser(),
                    RichSequenceBuilderFactory.FACTORY,
                    null);

                if (seqit.hasNext()) {
                    seq = seqit.nextRichSequence();
                    sformat = SeqIOConstants.AA;
                }
            }
            catch (BioException | IOException ex2) {
                noParsing = true;
            }
        } finally {
            try {
                is.close();
            } catch (IOException ex) {
                LOGGER.log(Level.INFO, null, ex);
                noParsing = true;
            }
        }
        return seq;
    }

    @Override
    public Lookup getLookup() {
        return getCookieSet().getLookup();
    }

    @Override
    public <T extends Cookie> T createCookie(Class<T> klass) {
        if (SequenceCookie.class.isAssignableFrom(klass)) {
            return klass.cast(this);
        }
        if (SeqProvider.class.isAssignableFrom(klass)) {
            return klass.cast(this);
        }
        if (klass.isAssignableFrom(RawEditorSupport.class)) {
            return klass.cast(getRawEditorSupport());
        }
        return null;
    }

    private synchronized RawEditorSupport getRawEditorSupport() {
        if ( rawEditorSupport == null ) {
            rawEditorSupport = new RawEditorSupport(this);
        }
        return rawEditorSupport;
    }

    // Package accessibility for RawEditorSupport:
    CookieSet getCookieSet0() {
        return getCookieSet();
    }
}
