/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */
package org.biobeans.modules.biocore.sequences;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import static org.biobeans.modules.biocore.sequences.FeatureUtil.getTerm;
import org.biojavax.Note;
import org.biojavax.RichAnnotation;
import org.biojavax.SimpleNote;
import org.biojavax.ontology.ComparableTerm;

public class AnnotationUtil {
    
    private AnnotationUtil() {};
    
    /** Compare RichAnnotation for identity.
     */
    public static boolean areRichAnnotationsIdentical(RichAnnotation ann1,RichAnnotation ann2) {
        Set<Note> set1 = ann1.getNoteSet();
        Set<Note> set2 = ann2.getNoteSet();
        if(set1.size()!=set2.size())
            return false;
        TreeSet<Note> seta,setb;
        if(set1 instanceof TreeSet)
            seta = (TreeSet<Note>)set1;
        else
            seta = new TreeSet<>(set1);
        if(set2 instanceof TreeSet)
            setb = (TreeSet<Note>)set2;
        else
            setb = new TreeSet<>(set2);
        return NoteUtil.areNoteSetsIdentical(seta,setb);
    }
    
    
    /** Retrieve all notes in a RichAnnotation having the same qualifier. 
     */
    public static List<Note> getNotes(RichAnnotation ran, String qualifier) {
        ComparableTerm term = getTerm(qualifier);
        List<Note> ns = new ArrayList<>();
        for (Note n : ran.getNoteSet()) {
            if(n.getTerm()==term) {
                ns.add(n);
            }
        }
        return ns;
    }

    /** Add a Note to a RichAnnotation. If a Note with the same qualifier/value
     * pair is found, no new Note is added. If a Note with the same qualifier
     * is found and change is allowed, its value is updated unless the
     * new value is empty, in which case the Note is removed. The rank of 
     * the new Note is set to the number of existing Notes + 1. 
     * 
     * @param ran  The RichAnnotation
     * @param qualifier A valid qualifier
     * @param value The note value
     * @param change Determine whether a Note with the same qualifier but a different
     * value is changed or a new Note is added.
     */
    public static void addNote(RichAnnotation ran, String qualifier, String value, boolean change) {
        List<Note> ns = AnnotationUtil.getNotes(ran,qualifier);
        for(Note note:ns) {
            String valnote = note.getValue();
            if(value.equals(valnote)) {
                return;
            }
        }
        if(change && ns.size()>0) {
            Note note = ns.iterator().next(); // change the first one
            if(value.isEmpty()) {
                ran.removeNote(note); // remove the first one
            } else {
                note.setValue(value);
            }
        } else if(!value.isEmpty() && !isSpace(value)) {  // create note
            int rank = ran.getNoteSet().size()+1;
            Note n = new SimpleNote(getTerm(qualifier),value,rank);
            ran.addNote(n);
        }
    }
    
    private static boolean isSpace(String value) {
        for(int i=0;i<value.length();i++) {
            if(!Character.isSpaceChar(value.charAt(i))) {
                return false;
            }
        }
        return true;
    }
    
    
}
