/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */
package org.biobeans.modules.biocore.sequences;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.FeatureHolder;
import org.biojava.ontology.InvalidTermException;
import org.biojava.utils.ChangeVetoException;
import org.biojavax.Note;
import org.biojavax.RankedCrossRef;
import org.biojavax.RichObjectFactory;
import org.biojavax.SimpleNote;
import org.biojavax.SimpleRankedCrossRef;
import org.biojavax.bio.seq.RichFeature;
import org.biojavax.bio.seq.RichFeatureRelationship;
import org.biojavax.bio.seq.RichLocation;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.SimpleRichFeature;
import org.biojavax.bio.seq.SimpleRichFeatureRelationship;
import org.biojavax.ontology.ComparableTerm;

public class FeatureUtil {
    
    private static final Logger LOGGER = Logger.getLogger(FeatureUtil.class.getName());
    
    private FeatureUtil() {};
    
    
    /** Get Term from String.
     * 
     */
    public static ComparableTerm getTerm(String term) {
        return RichObjectFactory.getDefaultOntology().getOrCreateTerm(term);
    }

    /** Retrieve a note in a RichFeature with a given qualifier; the first
     * found note is returned. 
     */
    public static Note getNote(RichFeature rf, String qualifier) {
        ComparableTerm term = getTerm(qualifier);
        for (Note n : rf.getNoteSet()) {
            if(term.equals(n.getTerm())) {
                return n;
            }
        }
        return null;
    }

    /** Retrieve all notes in a RichFeature having the same qualifier. 
     */
    public static List<Note> getNotes(RichFeature rf, String qualifier) {
        ComparableTerm term = getTerm(qualifier);
        List<Note> ns = new ArrayList<>();
        for (Note n : rf.getNoteSet()) {
            if(term.equals(n.getTerm())) {
                ns.add(n);
            }
        }
        return ns;
    }
    
    /** Add a note to a RichFeature. If a note with the same value
     * already exists, no new note is added. The rank of the new note
     * is set to the number of existing notes + 1.
     */
    public static void addNote(RichFeature rf, String value) {
        List<Note> ns = FeatureUtil.getNotes(rf,"note");
        for(Note note:ns) {
            if(value.equals(note.getValue())) {
                return;
            }
        }
        int rank = rf.getNoteSet().size()+1;
        Note n = new SimpleNote(getTerm("note"),value,rank);
        rf.getRichAnnotation().addNote(n);
    }
    
    /** Indicate whether a RichFeature is located on the negative strand.
     */
    public static boolean isNegativeFeature(RichFeature rf) {
        RichLocation rloc = (RichLocation) rf.getLocation();
        return rloc.getStrand()==RichLocation.Strand.NEGATIVE_STRAND;
    }
    
    /** Find the sequence start of a RichFeature.
     */
    public static int getFeatureLocationStart(RichFeature rf) {
        RichLocation rloc = RichLocation.Tools.enrich(rf.getLocation());
        int startpos = rloc.getMin();
        Iterator itloc = rloc.blockIterator();
        if (itloc.hasNext()) {
            RichLocation loc = (RichLocation) itloc.next();
           if(FeatureUtil.isNegativeFeature(rf)) {
                startpos = loc.getMax();
           } else {
               startpos = loc.getMin();
           }
        }
        return startpos;
    }
    
    /** Find the sequence end of a RichFeature.
     */
    public static int getFeatureLocationEnd(RichFeature rf) {
        RichLocation rloc = RichLocation.Tools.enrich(rf.getLocation());
        int endpos = rloc.getMax();
        
        Iterator itloc = rloc.blockIterator();
        if (itloc.hasNext()) {
            RichLocation loc = (RichLocation) itloc.next();
            while (itloc.hasNext()) {
                loc = (RichLocation) itloc.next();
            }
            if (FeatureUtil.isNegativeFeature(rf)) {
                endpos = loc.getMin();
            } else {
                endpos = loc.getMax();
            }
        }
        return endpos;
    }
    
    /** Calculate the length of the RichFeature.
     * 
     * @return length or 0 if not meaningful (parent not circular but should be)
     */
    public static int getFeatureLenth(RichFeature rf) {
        int start = getFeatureLocationStart(rf);
        int end = getFeatureLocationEnd(rf);
        if(start==end) {
            return 1;
        }
        FeatureHolder fh = rf.getParent();
        if(fh==null || !(fh instanceof RichSequence)) {
            return Math.abs(end-start)+ 1;
        }
        
        RichSequence rseq = (RichSequence) fh;
        
        RichLocation rloc = RichLocation.Tools.enrich(rf.getLocation());
        if (rloc.getStrand().equals(RichLocation.Strand.NEGATIVE_STRAND)) {
            if(end<start) {
                return start-end+1;
            } else {
                if(rseq.getCircular()) {
                    return start-end+1 + rseq.length();
                } else {
                    return 0;
                }
            }
        } else {
            if(start<end) {
                return end-start+1;
            } else {
                if(rseq.getCircular()) {
                    return end-start+1 + rseq.length();
                } else {
                    return 0;
                }
            }
        }
    }
    
    /** Clone a Feature into a RichFeature.
     */
    public static RichFeature cloneFeature(FeatureHolder parent, Feature feature) {
        RichFeature rf = RichFeature.Tools.enrich(feature);
        RichFeature clone = null;
        try {
            //Does not clone annotations, relationships, and crossrefs.
            clone = new SimpleRichFeature(parent,rf.makeTemplate());
            RichLocation loc = LocationUtil.cloneLocation(rf.getLocation());
            loc.setFeature(clone);
            clone.setLocation(loc);
            clone.setName(rf.getName());
            clone.setRank(rf.getRank());
            
            //notes
            Set<Note> notes = rf.getNoteSet();
            Iterator<Note> it = notes.iterator();
            Set<Note> n = new TreeSet<>();
            while(it.hasNext()) {
                Note note = it.next();
                n.add(NoteUtil.cloneNote(note));
            }
            clone.setNoteSet(n);
            
            //rankedcrossrefs
            Set<RankedCrossRef> rankedcrossrefs = rf.getRankedCrossRefs();
            Iterator<RankedCrossRef> rit = rankedcrossrefs.iterator();
            Set<RankedCrossRef> c = new TreeSet<>();
            while(rit.hasNext()) {
                RankedCrossRef ref = rit.next();
                c.add(new SimpleRankedCrossRef(
                        ref.getCrossRef(),
                        ref.getRank()));
            }
            clone.setRankedCrossRefs(c);
            
            //feature relationships
            Set<RichFeatureRelationship> relations = rf.getFeatureRelationshipSet();
            Iterator<RichFeatureRelationship> frit = relations.iterator();
            Set<RichFeatureRelationship> r = new TreeSet<>();
            while(frit.hasNext()) {
                RichFeatureRelationship rel = frit.next();
                RichFeatureRelationship frclone =  new SimpleRichFeatureRelationship(
                        clone,
                        rel.getObject(), 
                        rel.getTerm(), 
                        rel.getRank());
                r.add(frclone);
            }
            clone.setFeatureRelationshipSet(r);

        } catch (InvalidTermException | ChangeVetoException ex) {
            LOGGER.log(Level.INFO, ex.getMessage());
        }
        return clone;
    }
    
    
}
