/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */
package org.biobeans.modules.biocore.sequences;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Logger;
import static org.biobeans.modules.biocore.sequences.NoteUtil.cloneNote;
import org.biojava.bio.BioException;
import org.biojava.bio.symbol.Location;
import org.biojavax.Note;
import org.biojavax.bio.seq.CompoundRichLocation;
import org.biojavax.bio.seq.EmptyRichLocation;
import org.biojavax.bio.seq.MultiSourceCompoundRichLocation;
import org.biojavax.bio.seq.Position;
import org.biojavax.bio.seq.RichLocation;
import org.biojavax.bio.seq.RichLocation.Strand;
import static org.biojavax.bio.seq.RichLocation.Tools.flatten;
import org.biojavax.bio.seq.SimplePosition;
import org.biojavax.bio.seq.SimpleRichLocation;


public class LocationUtil {
    
    private static final Logger LOGGER = Logger.getLogger(LocationUtil.class.getName());
    
    private LocationUtil() {};
    
    /** This comparator sorts locations by CrossRef, then
     * strand, then term, then min, then max but never rank.
     */
    public static class LocationComparator implements Comparator<RichLocation> {
    
        @Override
        public int compare(RichLocation o1, RichLocation o2) {
            if (o1.getCrossRef() != null || o2.getCrossRef() != null) {
                if (o1.getCrossRef() != null && o2.getCrossRef() != null) {
                    return o1.getCrossRef().compareTo(o2.getCrossRef());
                } else {
                    return -1;
                }
            }
            if (!o1.getStrand().equals(o2.getStrand())) {
                return o1.getStrand().compareTo(o2.getStrand());
            }
            if (o1.getTerm() != null || o2.getTerm() != null) {
                if (o1.getTerm() != null && o2.getTerm() != null && !o1.getTerm().equals(o2.getTerm())) {
                    return o1.getTerm().compareTo(o2.getTerm());
                } else {
                    return -1;
                }
            }
            if (o1.getMin() != o2.getMin()) {
                return o1.getMin() - o2.getMin();
            }
            return o1.getMax() - o2.getMax();
        }
    }
    
    /** Find the sequence start of a location.
     * .
     * @param rloc
     * @return the start sequence position
     */
    public static int getLocationStart(RichLocation rloc) {
        int startpos = rloc.getMin();
        Iterator itloc = rloc.blockIterator();
        if (itloc.hasNext()) {
            RichLocation loc = (RichLocation) itloc.next();
           if(rloc.getStrand().equals(RichLocation.Strand.NEGATIVE_STRAND)) {
                startpos = loc.getMax();
           } else {
               startpos = loc.getMin();
           }
        }
        return startpos;
    }
    
    /** Find the sequence end of a location.
     * 
     * @param rloc
     * @return the end sequence position
     */
    public static int getLocationEnd(RichLocation rloc) {
        int endpos = rloc.getMax();
        
        Iterator itloc = rloc.blockIterator();
        if (itloc.hasNext()) {
            RichLocation loc = (RichLocation) itloc.next();
            while (itloc.hasNext()) {
                loc = (RichLocation) itloc.next();
            }
            if(rloc.getStrand().equals(RichLocation.Strand.NEGATIVE_STRAND)) {
                endpos = loc.getMin();
            } else {
                endpos = loc.getMax();
            }
        }
        return endpos;
    }
    
    /** Assert that a location is of a certain type, on the same strand of
     * a sequence with all the blocks in consecutive order, starting from rank 1. 
     * 
     * @param rloc A RichLocation
     * @param cancross If true, the location can go across the origin once.
     * @return true if all blocs are in consecutive order.
     */
    public static boolean verifyLocationOrder(RichLocation rloc, boolean cancross) {
        boolean ok = true;
        int cirlen = rloc.getCircularLength();
        int clen = cirlen!=0? cirlen:rloc.getMax()+1; // An arbitraty lengh of the sequence
        int curpos = getLocationStart(rloc);
        Strand strand = rloc.getStrand();
        RichLocation curloc = null;
        RichLocation prevloc = null;
        boolean cross = false;
        
        if(rloc instanceof EmptyRichLocation)
            return true;
        if(rloc instanceof MultiSourceCompoundRichLocation)
            return false;

        Iterator<Location> itloc = rloc.blockIterator();
        int rank = 1;
        if(strand.equals(RichLocation.Strand.NEGATIVE_STRAND)) {
            while (itloc.hasNext()) {
                curloc = RichLocation.Tools.enrich(itloc.next());
                if(!(curloc.getRank()==rank++)) {
                    return false;
                }
                if(cirlen != curloc.getCircularLength()) {
                    return false;
                }
                if(prevloc==null) {
                    prevloc = curloc;
                } else {
                    if(prevloc.overlaps(curloc)) {
                        return false;
                    }
                }
                if(strand!= curloc.getStrand())
                    return false;
                int curmin = curloc.getMin();
                int curmax = curloc.getMax();
                if (curloc.getMin() > curloc.getMax()) {
                    //Inverted location
                    return false;
                }
                if (!(curmax <= curpos)) {
                    if(!cancross)
                        return false;
                    if(cross) {
                        //Cross only once
                        return false;
                    }
                    // CounterRotate one circular length
                    curmin-=clen;
                    curmax-=clen;
                    cross = true;
                }
                if(!(curmax <= curpos)) {
                    ok = false;
                    break;
                }
                curpos = curmin;
                if(cross) {
                    // Rest position for incoming blocs
                    curpos += clen;
                }
            }
            
        } else {
            while (itloc.hasNext()) {
                curloc = RichLocation.Tools.enrich(itloc.next());
                if(!(curloc.getRank()==rank++)) {
                    return false;
                }
                if(cirlen != curloc.getCircularLength()) {
                    return false;
                }
                if(prevloc==null) {
                    prevloc = curloc;
                } else {
                    if(prevloc.overlaps(curloc)) {
                        return false;
                    }
                }
                if(strand!= curloc.getStrand())
                    return false;
                int curmin = curloc.getMin();
                int curmax = curloc.getMax();
                if (curloc.getMin() > curloc.getMax()) {
                    //Inverted location
                    return false;
                }
                if (!(curmin >= curpos)) {
                    if(!cancross)
                        return false;
                    if(cross) {
                        //Cross only once
                        return false;
                    }
                    // Rotate one circular length
                    curmin+=clen;
                    curmax+=clen;
                    cross = true;
                }
                if (!(curmin >= curpos) ) {
                        ok = false;
                        break;
                }
                curpos = curmax;
                if(cross) {
                    // Rest position for incoming blocs
                    curpos -= clen;
                }
            }
        }
        if (curloc!= null && curloc!=prevloc && prevloc.overlaps(curloc)) {
            return false;
        }
        return ok;
    }
    
    /** Order the blocks of a location of a certain type, on the same strand
     * of a sequence and not crossing the origin, and merge adjacent
     * blocks. Filter out erroneous circular length and strand in blocks.
     * 
     * @param rloc The RichLocation to order...
     * @return A RichLocation with properly ranked and fused blocks.
     */
    public static RichLocation orderLocation(RichLocation rloc) {
        if(!(rloc instanceof CompoundRichLocation)) {
            return rloc;
        }
           
        int circularlen = rloc.getCircularLength();
        Strand strand = rloc.getStrand();
        boolean isRCStrand = strand==Strand.NEGATIVE_STRAND;

        //Get the members
        List<Location> blockList = new ArrayList<>(flatten(rloc));

        //Order without the rank
        Collection<RichLocation> rlocs = new TreeSet(new LocationComparator());
        for (Location block : blockList) {
            rlocs.add(RichLocation.Tools.enrich(block));
        }

        //Merge
        List<RichLocation> newblocks = new ArrayList<>();
        RichLocation lastBlock = null;
        for (RichLocation rblock : rlocs) {
            if (lastBlock != null) {
                if (lastBlock.getMax() + 1 >= rblock.getMin()) {
                    lastBlock = new SimpleRichLocation(
                            lastBlock.getMinPosition(),
                            lastBlock.getMax() > rblock.getMax()
                                    ? lastBlock.getMaxPosition() : rblock.getMaxPosition(),
                            0, strand);
                    lastBlock.setCircularLength(circularlen);
                } else {
                    if (strand != lastBlock.getStrand()) {
                        lastBlock = new SimpleRichLocation(
                                lastBlock.getMinPosition(),
                                lastBlock.getMaxPosition(),
                                0, strand);
                    }
                    if(isRCStrand) {
                        newblocks.add(0,lastBlock);
                    } else {
                        newblocks.add(lastBlock);
                    }
                    lastBlock = rblock;
                }
            } else {
                lastBlock = rblock;
            }
        }
        if (lastBlock != null) {
            if (isRCStrand) {
                newblocks.add(0, lastBlock);
            } else {
                newblocks.add(lastBlock);
            }

        }

        //Rank
        int rank = 1;
        for (RichLocation loc : newblocks) {
            loc.setRank(rank++);
        }

        //Construct
        if (newblocks.isEmpty()) {
            return RichLocation.EMPTY_LOCATION;
        } else if (newblocks.size() == 1) {
            return newblocks.get(0);
        }
        RichLocation result = new CompoundRichLocation(newblocks);
        result.setRank(1);
        return result;

    }
    
    /** Exclude a contiguous location from another location. Locations must not
     *  have multiple sources and must seat on the same strand.
     *
     * @param location  The location to exclude from.
     * @param exclude   The location to be excluded, always contiguous.
     * @return          The resulting new location.
     */
    public static RichLocation excludeLocation(RichLocation location, RichLocation exclude) {
        if (location instanceof EmptyRichLocation) return location;
        if (exclude instanceof EmptyRichLocation) return location;
        
        if(!exclude.isContiguous()) {
            String mes = "Attempt to exclude non-contiguous location: "
                    + location.toString()
                    + " Excluding: "
                    + exclude.toString();
            LOGGER.info(mes);
            assert(false) : mes;
            return location;
        }
        
        Collection<Location> members = new ArrayList<>();
        members.addAll(RichLocation.Tools.flatten(location));
        members.add(exclude);
        if(RichLocation.Tools.isMultiSource(members)) {
            String mes = "Attempt to exclude multisource locations: "
                    + location.toString()
                    + " Excluding: "
                    + exclude.toString();
            LOGGER.info(mes);
            assert(false) : mes;
            return location;
        }
        
        RichLocation intersect = (RichLocation) location.intersection(exclude);
        if(intersect instanceof EmptyRichLocation) return location;

        int cirlen = location.getCircularLength();
        members = RichLocation.Tools.flatten(location);
        Collection<RichLocation> bucket = new ArrayList<>();
        for (Iterator i = members.iterator(); i.hasNext();) {
            RichLocation member = (RichLocation) i.next();
            RichLocation minex = (RichLocation) member.intersection(intersect);
            if (!(minex instanceof EmptyRichLocation)) {
                int ourModStart = 0;
                int ourModEnd = 0;
                int theirModStart = 0;
                int theirModEnd = 0;
                if (cirlen > 0) {
                    // Modulate our start/end to shortest possible equivalent region
                    int parts[] = RichLocation.Tools.modulateCircularLocationPair(member, minex, cirlen);
                    ourModStart = parts[0];
                    ourModEnd = parts[1];
                    theirModStart = parts[2];
                    theirModEnd = parts[3];
                } else {
                    ourModStart = member.getMin();
                    ourModEnd = member.getMax();
                    theirModStart = minex.getMin();
                    theirModEnd = minex.getMax();
                }
                if (ourModStart == theirModStart && ourModEnd == theirModEnd) {
                    // areas are identical
                } else if (ourModStart == theirModStart || ourModEnd == theirModEnd) {
                    // areas overlap on either end
                    Position startPos = null;
                    if(theirModStart == ourModStart) {
                        startPos = new SimplePosition(theirModEnd+1);
                    } else
                        startPos = new SimplePosition(ourModStart);
                    Position endPos = null;
                    if(theirModEnd == ourModEnd) {
                        endPos = new SimplePosition(theirModStart-1);
                    } else
                        endPos = new SimplePosition(ourModEnd);
                    RichLocation templ = new SimpleRichLocation(startPos, endPos, 0, location.getStrand(), location.getCrossRef());
                    if(cirlen!=0) templ.setCircularLength(cirlen);
                    bucket.add(templ);
                } else {
                    // or minexs is contained and two locations are created
                    RichLocation templ = new SimpleRichLocation(
                            new SimplePosition(ourModStart),
                            new SimplePosition(theirModStart-1),
                            0, location.getStrand(), location.getCrossRef());
                    if(cirlen!=0) templ.setCircularLength(cirlen);
                    bucket.add(templ);
                    templ = new SimpleRichLocation(
                            new SimplePosition(theirModEnd+1),
                            new SimplePosition(ourModEnd),
                            0, location.getStrand(), location.getCrossRef());
                    if(cirlen!=0) templ.setCircularLength(cirlen);
                    bucket.add(templ);
                }
                
            } else
                bucket.add(member);
        }
        
        if(bucket.isEmpty()) {
            return RichLocation.EMPTY_LOCATION;
        } else if(bucket.size()==1) {
            RichLocation loc = (RichLocation) bucket.toArray(new Location[0])[0];
            loc.setRank(1);
            return loc;
        }

        int rank = 1;
        for(RichLocation rl:bucket)
            rl.setRank(rank++);
        RichLocation result = new CompoundRichLocation(bucket);
        result.setRank(1);
        return result;
    }

    /** Join two locations that are from a single source and on the same strand.
     *
     * @param loc1 A RichLocation
     * @param loc2 A RichLocation
     * @return The joined location. 
     * @throws org.biojava.bio.BioException 
     */
    public static RichLocation joinLocation(RichLocation loc1, 
            RichLocation loc2) throws BioException {
        
        if(loc1 instanceof EmptyRichLocation 
                && loc2 instanceof EmptyRichLocation){
            return RichLocation.EMPTY_LOCATION;
        }
        if(loc1 instanceof EmptyRichLocation) {
            return cloneLocation(loc2);
        }
        if(loc2 instanceof EmptyRichLocation) {
            return cloneLocation(loc1);
        }
        if (loc1 instanceof MultiSourceCompoundRichLocation
                || loc2 instanceof MultiSourceCompoundRichLocation) {
            throw new BioException("Error joining multisource locations");
        }
        Strand strand = loc1.getStrand();
        if(strand != loc2.getStrand()) {
            throw new BioException("Error joining locations on different strands");
        }
        
        if(loc1.getCircularLength()!=loc2.getCircularLength()) {
            throw new BioException("Error joining locations of different circularity");
        }
        
        RichLocation jloc = RichLocation.Tools.enrich(loc1.union(loc2));
        //Simplest solution; we do not know where this location really starts
        jloc = orderLocation(jloc);
        //Union are very error prone by reusing members that are part of other
        //locations and then changing their rank. It is safer to clone just
        //everything.
        jloc = cloneLocation(jloc);
        jloc.setRank(1);
        return jloc;

    }
    
    /** Create the reverse-complement location of a location, i.e. the equivalent 
     * location on the reverse complement strand.
     * 
     @param rloc The RichLocation to reverse-complement.
     @param length The length of the sequence this location belongs to.
     * @return 
     */
    public static RichLocation getReverseComplementLocation(RichLocation rloc,int length) {
        Collection<Location> members = RichLocation.Tools.flatten(rloc);
        Collection<RichLocation> newblocks = new ArrayList();
        for(Location member:members) {
            RichLocation oldBlock = RichLocation.Tools.enrich(member);
            Position oldend = oldBlock.getMaxPosition();
            Position newstart = new SimplePosition(
                    oldend.getFuzzyStart(),
                    oldend.getFuzzyEnd(),
                    length - oldend.getEnd() +1,
                    length - oldend.getStart() +1,
                    oldend.getType());
            Position oldstart = oldBlock.getMinPosition();
            Position newend = new SimplePosition(
                    oldstart.getFuzzyStart(),
                    oldstart.getFuzzyEnd(),
                    length - oldstart.getEnd() +1,
                    length - oldstart.getStart() +1,
                    oldstart.getType());
            RichLocation.Strand strand = oldBlock.getStrand();
            if(strand.equals(RichLocation.Strand.POSITIVE_STRAND)) {
                strand = RichLocation.Strand.NEGATIVE_STRAND;
            } else if(strand.equals(RichLocation.Strand.NEGATIVE_STRAND)) {
                strand = RichLocation.Strand.POSITIVE_STRAND;
            } else {
                strand = RichLocation.Strand.UNKNOWN_STRAND;
            }
            RichLocation newBlock = new SimpleRichLocation(newstart,newend,0,strand,oldBlock.getCrossRef());
            newBlock.setCircularLength(oldBlock.getCircularLength());
            newblocks.add(newBlock);
        }
        
        if(newblocks.isEmpty()) {
            return RichLocation.EMPTY_LOCATION;
        } else if(newblocks.size()==1) {
            RichLocation loc = (RichLocation) newblocks.toArray(new Location[0])[0];
            loc.setRank(1);
            return loc;
        }
        
        int rank = 0;
        for(RichLocation rl:newblocks)
            rl.setRank(++rank);
        if(RichLocation.Tools.isMultiSource(members)) {
            return new MultiSourceCompoundRichLocation(newblocks);
        } else {
            return new CompoundRichLocation(newblocks);
        }
    }

    /** Create the complement location of a location, i.e. the same location  
     * taken on the other strand, member order is conserved.
     @param rloc The RichLocation to complement
     @return the complement location.
     */
    public static RichLocation getComplementLocation(RichLocation rloc) {
        Collection<Location> members = RichLocation.Tools.flatten(rloc);
        Set<RichLocation> newblocks = new TreeSet<>(new LocationComparator());
        for(Location member:members) {
            RichLocation oldBlock = RichLocation.Tools.enrich(member);
            Position oldstart = oldBlock.getMinPosition();
            Position oldend = oldBlock.getMaxPosition();
            RichLocation.Strand strand = oldBlock.getStrand();
            if(strand.equals(RichLocation.Strand.POSITIVE_STRAND)) {
                strand = RichLocation.Strand.NEGATIVE_STRAND;
            } else if(strand.equals(RichLocation.Strand.NEGATIVE_STRAND)) {
                strand = RichLocation.Strand.POSITIVE_STRAND;
            } else {
                strand = RichLocation.Strand.UNKNOWN_STRAND;
            }
            RichLocation newBlock = new SimpleRichLocation(oldstart,oldend,0,strand,oldBlock.getCrossRef());
            newBlock.setCircularLength(oldBlock.getCircularLength());
            newblocks.add(newBlock);
        }
        
        if(newblocks.isEmpty()) {
            return RichLocation.EMPTY_LOCATION;
        } else if(newblocks.size()==1) {
            RichLocation loc = (RichLocation) newblocks.toArray(new Location[0])[0];
            loc.setRank(1);
            return loc;
        }
        
        int rank = 0;
        for(RichLocation rl:newblocks)
            rl.setRank(++rank);
        if(RichLocation.Tools.isMultiSource(members)) {
            return new MultiSourceCompoundRichLocation(newblocks);
        } else {
            return new CompoundRichLocation(newblocks);
        }
    }
    
    /** Translate a location simply, without crossing the origin for circular
     * locations, without crossing boundaries for linear locations (i.e. not
     * moving the origin below 1 for negative moves).
     * 
     * @param rloc The RichLocation to translate
     * @param move A meaningful translation value, can be negative.
     */
    private static RichLocation _translateLocation(RichLocation rloc,int move) {
        Collection<Location> members = RichLocation.Tools.flatten(rloc);
        Set<RichLocation> newblocks = new TreeSet<>(new LocationComparator());
        for(Location member:members) {
            RichLocation oldBlock = RichLocation.Tools.enrich(member);
            Position oldstart = oldBlock.getMinPosition();
            Position newstart = new SimplePosition(
                    oldstart.getFuzzyStart(),
                    oldstart.getFuzzyEnd(),
                    oldstart.getEnd() + move,
                    oldstart.getStart() + move,
                    oldstart.getType());
            
            Position oldend = oldBlock.getMaxPosition();
            Position newend = new SimplePosition(
                    oldend.getFuzzyStart(),
                    oldend.getFuzzyEnd(),
                    oldend.getEnd() + move,
                    oldend.getStart() + move,
                    oldend.getType());
            RichLocation.Strand strand = oldBlock.getStrand();
            RichLocation newBlock = new SimpleRichLocation(newstart,newend,0,strand,oldBlock.getCrossRef());
            newBlock.setCircularLength(oldBlock.getCircularLength());
            newblocks.add(newBlock);
        }
        
        if(newblocks.isEmpty()) {
            return RichLocation.EMPTY_LOCATION;
        } else if(newblocks.size()==1) {
            RichLocation loc = (RichLocation) newblocks.toArray(new Location[0])[0];
            loc.setRank(1);
            return loc;
        }
        
        int rank = 0;
        for(RichLocation rl:newblocks)
            rl.setRank(++rank);
        if(RichLocation.Tools.isMultiSource(members)) {
            return new MultiSourceCompoundRichLocation(newblocks);
        } else {
            return new CompoundRichLocation(newblocks);
        }
    }
    
    /** Move a location around a circular sequence.
     * @param rloc The RichLocation to translate
     * @param move A translation value.
     * @return 
     */
    public static RichLocation translateLocation(RichLocation rloc,int move) {
        if(rloc instanceof EmptyRichLocation) {
            return rloc;
        }
        int circularlength = rloc.getCircularLength();
        RichLocation.Strand strand = rloc.getStrand();
        boolean simple = false;
        if(circularlength==0) {
            simple = true;
        } 
        
        if(move<0) {
            if( (-move)<rloc.getMin()) {
                simple = true;
            } else if(circularlength==0) {
                throw new RuntimeException("Impossible to translate this location to negative indices");
            }
        } else {
            if((rloc.getMax()+move)<=circularlength) {
                simple = true;
            }
        }
        if(simple) {
            return _translateLocation(rloc,move);
        }
        
        // Convert negative move to positive rotation
        while(move<0) {
            move+=circularlength;
        }
        
        Collection<Location> members = RichLocation.Tools.flatten(rloc);
        List<Location> newblocks = new ArrayList<>();
        for(Location member:members) {
            RichLocation oldBlock = RichLocation.Tools.enrich(member);
            
            Position oldstart = oldBlock.getMinPosition();
            Position newstart = new SimplePosition(
                    oldstart.getFuzzyStart(),
                    oldstart.getFuzzyEnd(),
                    RichLocation.Tools.modulateCircularIndex(oldstart.getEnd() + move,circularlength),
                    RichLocation.Tools.modulateCircularIndex(oldstart.getStart() + move,circularlength),
                    oldstart.getType());
            
            Position oldend = oldBlock.getMaxPosition();
            Position newend = new SimplePosition(
                    oldend.getFuzzyStart(),
                    oldend.getFuzzyEnd(),
                    RichLocation.Tools.modulateCircularIndex(oldend.getEnd() + move,circularlength),
                    RichLocation.Tools.modulateCircularIndex(oldend.getStart() + move,circularlength),
                    oldend.getType());
            
            if(newstart.getStart()>newend.getStart()) {
                // we are across the origin
                Position midA = new SimplePosition(circularlength);
                Position midB = new SimplePosition(1);
                RichLocation newBlockA = new SimpleRichLocation(newstart,midA,0,strand,oldBlock.getCrossRef());
                newBlockA.setCircularLength(circularlength);
                
                RichLocation newBlockB = new SimpleRichLocation(midB,newend,0,strand,oldBlock.getCrossRef());
                newBlockB.setCircularLength(circularlength);
                if(strand==RichLocation.Strand.NEGATIVE_STRAND) {
                    newblocks.add(newBlockB);
                    newblocks.add(newBlockA);
                } else {
                    newblocks.add(newBlockA);
                    newblocks.add(newBlockB);
                }
            } else {
                RichLocation newBlock = new SimpleRichLocation(newstart,newend,0,strand,oldBlock.getCrossRef());
                newBlock.setCircularLength(circularlength);
                newblocks.add(newBlock);
            }
        }
        
        // not working because origin change creates contiguous not overlapping locations
        //members = RichLocation.Tools.merge(newblocks);
        if(strand==RichLocation.Strand.NEGATIVE_STRAND) {
            int prevmin = 0;
            for (int p = 0; p < newblocks.size(); p++) {
                Location loc = newblocks.get(p);
                int max = loc.getMax();
                if(prevmin!=0 && max==prevmin-1) {
                    Location prevBlock = newblocks.get(p-1);
                    RichLocation newBlock = new SimpleRichLocation(
                            new SimplePosition(loc.getMin()),
                            new SimplePosition(prevBlock.getMax()),
                            0,
                            strand,
                            ((RichLocation)prevBlock).getCrossRef());
                    newBlock.setCircularLength(circularlength);
                    newblocks.set(p-1,newBlock);
                    newblocks.remove(p);
                    break; // Only one interrupted block
                }
                prevmin = loc.getMin();
            }
        } else {
            int prevmax = 0;
            for (int p = 0; p < newblocks.size(); p++) {
                Location loc = newblocks.get(p);
                int min = loc.getMin();
                if(prevmax!=0 && min==prevmax+1) {
                    Location prevBlock = newblocks.get(p-1);
                    RichLocation newBlock = new SimpleRichLocation(
                            new SimplePosition(prevBlock.getMin()),
                            new SimplePosition(loc.getMax()),
                            0,
                            strand,
                            ((RichLocation)prevBlock).getCrossRef());
                    newBlock.setCircularLength(circularlength);
                    newblocks.set(p-1,newBlock);
                    newblocks.remove(p);
                    break; // Only one interrupted block
                }
                prevmax = loc.getMax();
            }
        }
        members = newblocks;
        
        int rank = 0;
        for(Location rl:members) {
            ((RichLocation)rl).setRank(++rank);
        }
        if(members.isEmpty()) {
            return RichLocation.EMPTY_LOCATION;
        } else if(members.size()==1) {
            return (RichLocation) newblocks.toArray(new Location[0])[0];
        }
        if(RichLocation.Tools.isMultiSource(members)) {
            return new MultiSourceCompoundRichLocation(members);
        } else {
            return new CompoundRichLocation(members);
        }
    }
    
    /** Clone a Position.
     * 
     * @param pos The Position to clone
     * @return The cloned location
     */
    public static Position clonePosition(Position pos) {
        if (pos == Position.EMPTY_POSITION) {
            return Position.EMPTY_POSITION;
        }

        Position clone = new SimplePosition(
                pos.getFuzzyStart(),
                pos.getFuzzyEnd(),
                pos.getStart(),
                pos.getEnd(),
                pos.getType());

        return clone;
    }
    

    /** Clone a Location into a RichLocation.
     * @param loc The location to clone
     * @return The cloned location
     * 
     */
    public static RichLocation cloneLocation(Location loc) {
        RichLocation rloc = RichLocation.Tools.enrich(loc);
        RichLocation clone = null;
        if(rloc instanceof MultiSourceCompoundRichLocation ) {
            List<RichLocation> members = new ArrayList<>();
            Iterator c = rloc.blockIterator();
            while (c.hasNext()) members.add(cloneLocation((RichLocation) c.next()));
            clone = new MultiSourceCompoundRichLocation(rloc.getTerm(),members);
            clone.setRank(rloc.getRank());
        } else if(rloc instanceof CompoundRichLocation ) {
            List<RichLocation> members = new ArrayList<>();
            Iterator c = rloc.blockIterator();
            while (c.hasNext()) members.add(cloneLocation((RichLocation) c.next()));
            clone = new CompoundRichLocation(rloc.getTerm(),members);
            clone.setRank(rloc.getRank());
        } else if(rloc instanceof SimpleRichLocation ) {
            clone = new SimpleRichLocation(
                    clonePosition(rloc.getMinPosition()),
                    clonePosition(rloc.getMaxPosition()),
                    rloc.getRank(),
                    rloc.getStrand(),
                    rloc.getCrossRef());
            clone.setTerm(rloc.getTerm());
            clone.setCircularLength(rloc.getCircularLength());
            //notes
            Set<Note> notes = rloc.getNoteSet();
            Iterator<Note> it = notes.iterator();
            Set<Note> n = new TreeSet<>();
            while(it.hasNext()) {
                Note note = it.next();
                n.add(cloneNote(note));
            }
            clone.setNoteSet(n);
        } else if (rloc instanceof EmptyRichLocation) {
            clone = RichLocation.EMPTY_LOCATION;
        }
        return clone;
    }
    
    /** Test whether a location contains
     * the origin but does not start or end on it.
     *
     * @param loc a  RichLocation
     */
    public static boolean crossOrigin(RichLocation loc) {
        if(!loc.isContiguous()) {
            Iterator blocs = loc.blockIterator();
            boolean result = false;
            while(blocs.hasNext()) {
                RichLocation rl = (RichLocation) blocs.next();
                result |= crossOrigin(rl);
            }
            return result;
        }
        int len = loc.getCircularLength();
        if(len==0)
            return false;
        int[] i = RichLocation.Tools.modulateCircularLocation(loc.getMin(), loc.getMax(), len);
        return i[1]>len;
    }
    
    /** Test whether a location is circular and made of 2 blocks
     * touching each other at the origin (contiguous across the origin).
     */
    public static boolean isLocationContigousAcrossOrigin (RichLocation rloc) {
        if(rloc.isContiguous()) {
            return false;
        }
        int cirlen = rloc.getCircularLength();
        if(cirlen==0) {
            return false;
        }
        Iterator itloc = rloc.blockIterator();
        if (itloc.hasNext()) {
            RichLocation loc = (RichLocation) itloc.next();
            if(loc.getMax()==cirlen) {
                if (itloc.hasNext()) {
                    loc = (RichLocation) itloc.next();
                    if(loc.getMin()==1) {
                        if (!itloc.hasNext()) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }
}
