/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */

package org.biobeans.modules.biocore.sequences;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import org.biojava.bio.seq.Sequence;
import org.biojavax.Namespace;
import org.biojavax.RichObjectFactory;
import org.biojavax.SimpleNamespace;
import org.biojavax.bio.seq.RichSequence;
import org.openide.util.NbBundle;

/** Sequence identifier associated with the parameters Namespace,
 * name, accession, and version. These properties are not null 
 * and are immutable. Equality is the combination of Namespace, 
 * name, accession and version, like for a BioEntry. When not specified,
 * the Namespace is set to a generic Namespace whose name is an empty String,
 * the accession to an empty String, and the name to an empty string as well. 
 * This Identifier also contains a description and an identifier number 
 * that are used for the writing or generated during the parsing of FASTA files.
 * 
 * @author George Waldon
 */
public class RichIdentifier extends Identifier implements Serializable {
    
    static final long serialVersionUID = 1L;
    
    /** The local Namespace "lcl".
     */
    private static final Namespace LOCAL = (Namespace)RichObjectFactory.getObject(
            SimpleNamespace.class,new Object[]{"lcl"});
    
    private Namespace ns;
    private String name;
    private String accession;
    private int version = 0;
    private String description = null;
    private String id = null;
    
    /** Create a RichIdentifier from a RichSequence.
     *
     * @param seq a RichSequence
     * @return the identifier
     */
    public static RichIdentifier create(RichSequence seq) {
        if(null==seq)
            return new RichIdentifier(LOCAL,"","",0);
        RichIdentifier rid =
                new RichIdentifier(seq.getNamespace(),seq.getName(),seq.getAccession(),seq.getVersion());
        String des = seq.getDescription();
        if( des!=null && des.length()!=0 ) {
            if(des.indexOf('\n')!=-1)
                des = des.replace('\n', ' ');
            rid.setDescription(des);
        }
        String i = seq.getIdentifier();
        if(i!=null && i.length()!=0)
            rid.setIdentifier(i);
        return rid;
    }
    
    /** Create a RichIdentifier from a Sequence. If the input sequence is not a
     *  RichSequence, the Namespace is set to {@link #LOCAL} and the 
     * accession to an empty String and the version to 0.
     *
     * @param seq a sequence
     * @return the identifier
     */
    public static RichIdentifier create(Sequence seq) {
        if(seq instanceof RichSequence)
            return create((RichSequence)seq);
        RichIdentifier rid = new RichIdentifier(
                LOCAL,
                seq.getName(),
                "",
                0);
        return rid;
    }
    
     /** Create a RichIdentifier.
      *
      * @param ns Namespace, if null, set to {@link #LOCAL}.
      * @param name the name as String, if null, set to "".
      * @param accession the accession as String, if null, set to "".
      * @param version the version as integer
      * @return the identifier
      */
    public static RichIdentifier create(Namespace ns, String name, String accession, int version) {
        if(ns==null || ns.getName().length()==0) 
            ns = LOCAL;
        if(name==null)
            name = "";
        if(accession==null)
            accession = "";
        RichIdentifier rid = new RichIdentifier(ns,name,accession,version);
        return rid;
    }
    
    /**  Parser for description line of FASTA files
     *
     * @param a String is understood as followed:
     *  gi|<identifier>|<namespace>|<accession>.<version>|<name> <description>
     * or
     *  gi|<identifier>|<namespace>|<accession>|<name> <description>
     * or
     *  <namespace>|<accession>.<version>|<name> <description>
     * or
     *  <namespace>|<accession>|<name> <description>
     * or
     *  <name> <description>
     * or
     *  <name>
     *
     * @return a RichIdentifier
     */
    public static RichIdentifier createFromDescription(String des) {
        String delim = "\\|";
        String ns = "";
        String name_description = "";
        String accession_version = "";
        String name = "";
        String description = "";
        String accession = "";
        String identifier = "";
        int version = 0;
        
        if(des.length()==0 || des.equals("|||"))
            return new RichIdentifier( LOCAL, "", "", 0);
        
        String[] tokens = des.split(delim,5);
        String start = tokens[0];
        start = start.trim();
        if(start.equals("gi")) {
            if(tokens.length==5) {
                identifier = tokens[1];
                ns = tokens[2];
                accession_version = tokens[3];
                name_description = tokens[4];
            } else if(tokens.length==4) {
                //something missing, guess = namespace
                identifier = tokens[1];
                accession_version = tokens[2];
                name_description = tokens[3];
            } else if(tokens.length==3){
                //something missing, guess = namespace + accession
                identifier = tokens[1];
                name_description = tokens[2];
            }
                
        } else if(tokens.length==1) {
            //name description
            name_description = tokens[0];
        } else if(tokens.length==2) {
            // RichSequence creates <accession>|<name>
            accession_version = tokens[0];
            name_description = tokens[1];
            // This could be true
            //ns = tokens[0];
            //accession_version = tokens[1];
        } else {
            ns = tokens[0];
            accession_version = tokens[1];
            name_description = tokens[2];
            int more = tokens.length - 3;
            while(more>0) {
               name_description += tokens[tokens.length-more];
               more--;
            }
        }
        
        Namespace space = LOCAL;
        ns = ns.trim();
        if(ns.length()!=0) {
            space = (Namespace)RichObjectFactory.getObject(
                    SimpleNamespace.class,new Object[]{ns});
        }
        
        if(name_description.length()!=0) {
            int pos = name_description.indexOf(' ');
            if(pos==-1) {
                name = name_description;
                name = name.trim();
            } else if(pos==0) {
                name_description = name_description.trim();
                description = name_description;
            } else {
                name = name_description.substring(0,pos);
                description = name_description.substring(pos+1,name_description.length());
                description = description.trim();
            }
        }
        
        int pos = accession_version.lastIndexOf('.');
        if(pos>=0){
            accession = accession_version.substring(0,pos);
            if(pos<(accession_version.length()-1) )  { // . at the end
                try {
                    version = Integer.parseInt(accession_version.substring(pos+1,accession_version.length()));
                } catch (NumberFormatException ex) {
                    // Something went wrong, don't try to find a version
                    accession = accession_version;
                    version = 0;
                }
            }
        } else
            accession = accession_version;
        accession = accession.trim();
        
        RichIdentifier rid = new RichIdentifier(
                space,
                name,
                accession,
                version);
        if(description.length()!=0)
            rid.setDescription(description);
        if(identifier.length()!=0)
            rid.setIdentifier(identifier);
        return rid;
    }
    
    /**  Parser for String formatted as "<namespace>|<accession>.<version>|<name>"
     *
     * @return a RichIdentifier
     */
    public static RichIdentifier createFromString(String s) throws IllegalArgumentException {
        String delim = "\\|";
        String ns = "";
        String name = "";
        String accession_version = "";
        String accession = "";
        int version = 0;
        
        if(s.length()==0 || s.equals("|||"))
            return new RichIdentifier( LOCAL, "", "", 0);

        String[] tokens = s.split(delim,3);
        if(tokens.length==1) {
            //name
            name = tokens[0];
        } else if(tokens.length==2) {
            //space + accession
            ns = tokens[0];
            accession_version = tokens[1];
        } else {
            ns = tokens[0];
            accession_version = tokens[1];
            name = tokens[2];
        }
        
        //we assume one dot if any, separating accession from version
        int pos = accession_version.indexOf('.');
        if(pos>=0){
            accession = accession_version.substring(0,pos);
            try {
                version = Integer.parseInt(accession_version.substring(pos+1,accession_version.length()));
            } catch (NumberFormatException ex) {
                throw new IllegalArgumentException("Cannot parse version from this string");
            }
        } else
            accession = accession_version;
        Namespace space = LOCAL;
        ns = ns.trim();
        if(ns.length()!=0) {
            space = (Namespace)RichObjectFactory.getObject(
                    SimpleNamespace.class,new Object[]{ns});
        }
        RichIdentifier rid = new RichIdentifier(
                space,
                name,
                accession,
                version);
        return rid;
    }
    
    /** Create a RichIdentifier from a simple name.
     * 
     * @param name
     * @return 
     */
    public static RichIdentifier createFromName(String name) {
        Identifier id = Identifier.create(name);
        return enrich(id);
    }
    
    /** Enrich an Identifier.
     * 
     * @param name
     * @return 
     */
    public static RichIdentifier enrich(Identifier id) {
        if(id instanceof RichIdentifier)
            return (RichIdentifier)id;
        RichIdentifier rid = new RichIdentifier(
                LOCAL,
                id.getName(),
                "",
                0);
        return rid;
    }

    /**
     * Creates a new RichIdentifier representing a sequence in the given namespace
     * with the given name, accession and version.
     * @param ns The namespace for this new bioentry (not null).
     * @param name The name for this new bioentry (not null).
     * @param accession The accession for this new bioentry (not null).
     * @param version The version for this new bioentry.
     */
    private RichIdentifier(Namespace ns, String name, String accession, int version) {
        super(name);
        if (name==null) 
            name = "";
        if (accession==null) 
            accession = "";
        if (ns==null) 
            ns = LOCAL;
        this.ns = ns;
        this.name = name;
        this.accession = accession;
        this.version = version;
    }
    
    /** Get the localized display name of identifier.
     *
     * @return a name for display, by default from to name; if name is null, the
     * accession is used.
     */
    public String getDisplayName() {
        String dname = getName();
        if(dname==null || dname.length()==0)
            dname = getAccession();
        if(dname==null || dname.length()==0)
            dname = NbBundle.getMessage(RichIdentifier.class, "LBL_alternatename");
        return dname;
    }
    
    /**
     * @return a String in the form 
     * "gi|<identifier>|<namespace>|<accession>.<version>|<name> <description>"
     */
    private String getCompleteLine() {
        String des = "";
        if(id!=null && id.length()!=0)
            des += "gi|" + id + "|";
        des += ns.getName() + "|";
        String acc = (accession==null)? "":accession;
        des += acc + "." + String.valueOf(version)+ "|";
        if(name!=null && name.length()!=0)
            des += name;
        if(description!=null && description.length()!=0)
            des += " " + description;
        return des;
    }
    
    /** Create a description line, useful e.g. for  printing FASTA files.
     *
     * @return a String in the form 
     * "gi|<identifier>|<namespace>|<accession>.<version>|<name> <description>"
     * or
     * <namespace>|<accession>.<version>|<name> <description> if the gi number
     * is null
     * or 
     * <accession>.<version>|<name> <description> when there is no gi number and 
     * the namespace is local
     * or
     * <name> <description> if <version> is 0 and <accession> equals <name>
     * or
     * <name> <description> when the accession is also null
     */
    public String getDescriptionLine() {
        if(id==null && ns.getName().equals("lcl")) {
            String line =  "";
            if(accession!=null && accession.length()!=0 ) {
                // accession = name and no versioning
                if(version==0 && accession.equals(name)) {
                    line += name;
                    if(description!=null && description.length()!=0)
                        line += " " + description;
                    return line;
                }
                line += accession;
                if(version>0)
                    line += "." + version;
                line += "|";
            }
            if(name!=null && name.length()!=0)
                line += name;
            if(description != null && description.length() != 0)
                line += " " + description;
            return line;
        }
        return getCompleteLine();
    }

    /**
     * @return a String in the form "<namespace>|<accession>.<version>|<name>"
     * The LOCAL namespace "lcl" is not incorporated, only <name> is returned
     * if the accession is null, and <accession>.<version> is
     * returned if accession and name are identical.
     */
    @Override
    public String toString() {
        String des = "";
        if( !(ns.getName().equals("lcl")) ) {
            des += ns.getName() + "|";
            des += accession + "." + String.valueOf(version)+ "|";
            des += name;
        } else if (accession==null || accession.length()==0)  {
            des += name;
        } else if( accession.equals(name) ) {
            des += accession + "." + String.valueOf(version);
        } else {
            des += accession + "." + String.valueOf(version)+ "|";
            des += name;
        }

        return des;
    }
    
    public String getIdentifier() {
        return id;
    }
    
    public void setIdentifier(String id) {
        this.id = id;
    }
    
    public Namespace getNamespace() {
        return ns;
    }
    
    public String getAccession() {
        return accession;
    }
    
    public int getVersion() {
        return version;
    }
    
    public String getDescription() {
        return description;
    }
    
    public void setDescription(String description) {
        this.description = description;
    }
    
    /** Same as BioEntries.
     */
    @Override
    public int compareTo(Object o) {
        RichIdentifier other = (RichIdentifier)o;
        if (!this.ns.equals(other.getNamespace())) 
            return this.ns.compareTo(other.getNamespace());
        if (!this.name.equals(other.getName())) 
            return this.name.compareTo(other.getName());
        if (!this.accession.equals(other.getAccession())) 
            return this.accession.compareTo(other.getAccession());
        return this.version-other.getVersion();
    }
    
    /** Equality on
     * 1) Namespace
     * 2) name
     * 3) accession
     * 4) version
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) 
            return true;
        if (o==null || !(o instanceof RichIdentifier)) 
            return false;
           RichIdentifier other = (RichIdentifier)o;
           return (this.ns.equals(other.getNamespace()) &&
                   this.name.equals(other.getName()) &&
                   this.accession.equals(other.getAccession()) &&
                   this.version==other.getVersion());
    }
    
    @Override
    public int hashCode() {
        int code = 17;
        code = 37*code + this.ns.hashCode();
        code = 37*code + this.name.hashCode();
        code = 37*code + this.accession.hashCode();
        code = 37*code + this.version;
        return code;
    }
    
    @Override
    public Object clone() {
        RichIdentifier newId = new RichIdentifier(getNamespace(),getName(),getAccession(),getVersion());
        newId.setDescription(this.getDescription());
        newId.setIdentifier(this.getIdentifier());
        return newId;
    }
    
    private void readObject(ObjectInputStream ois)
    throws IOException, ClassNotFoundException {
        String s = (String)ois.readObject();
        RichIdentifier rid = createFromDescription(s);
        ns = rid.getNamespace();
        name = rid.getName();
        accession = rid.getAccession();
        version = rid.getVersion();
        description = rid.getDescription();
        id = rid.getIdentifier();
    }

    private void writeObject(ObjectOutputStream oos)
    throws IOException {
        oos.writeObject(this.getDescriptionLine());
    } 
}
