/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */

package org.biobeans.modules.biocore.sequences;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.StringReader;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.FeatureFilter;
import org.biojava.bio.seq.FeatureHolder;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.RNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.Alphabet;
import org.biojava.bio.symbol.Edit;
import org.biojava.bio.symbol.IllegalAlphabetException;
import org.biojava.bio.symbol.SimpleSymbolList;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeListener;
import org.biojava.utils.ChangeType;
import org.biojava.utils.ChangeVetoException;
import org.biojava.utils.Unchangeable;
import org.biojavax.Comment;
import org.biojavax.Namespace;
import org.biojavax.Note;
import org.biojavax.RankedCrossRef;
import org.biojavax.RankedDocRef;
import org.biojavax.RichAnnotation;
import org.biojavax.RichObjectFactory;
import org.biojavax.SimpleComment;
import org.biojavax.SimpleNote;
import org.biojavax.SimpleRankedCrossRef;
import org.biojavax.SimpleRankedDocRef;
import org.biojavax.bio.BioEntryRelationship;
import org.biojavax.bio.SimpleBioEntryRelationship;
import org.biojavax.bio.seq.RichLocation;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;
import org.biojavax.bio.seq.SimpleRichSequence;
import org.biojavax.bio.seq.io.GenbankFormat;
import org.biojavax.bio.seq.io.RichSequenceBuilderFactory;
import org.biojavax.bio.seq.io.RichSequenceFormat;
import org.biojavax.bio.seq.io.RichStreamReader;
import org.biojavax.bio.seq.io.RichStreamWriter;
import org.biojavax.bio.taxa.NCBITaxon;
import org.openide.util.NbBundle;
import org.openide.util.NbPreferences;

public class RichSequenceFactory {
    
    private static final Logger LOGGER = Logger.getLogger(RichSequenceFactory.class.getName());
    
    public static String TaxonomyRoot = "9ZZZZ"; //Root of taxonomy in UniProt format used for synthetic or unknown origin construct.
    
    public static final String PROP_ACCESSIONNUMBER = "accession";
    
    public static String DEFAULT_ACCESSION = "0";
    
    static Pattern p = Pattern.compile("^\\d\\d.(JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC).\\d\\d\\d\\d$");

    /** no instantiation */
    private RichSequenceFactory() {
    }
    
    
    private static Preferences prefs() {
        return NbPreferences.forModule(RichSequenceFactory.class);
    }
    
    /** Return a unique integer that can be used to number accession.
     */
    private static synchronized int getNextAccessionNumber() {
        int cur = Integer.parseInt(prefs().get(PROP_ACCESSIONNUMBER, DEFAULT_ACCESSION));
        cur++;
        prefs().put(PROP_ACCESSIONNUMBER, String.valueOf(cur));
        return cur;
    }

    /** Create a string that can be used as accession for a sequence based on
     * an input string by appending a unique number at the end; if the resulting
     * String is longer than 16 characters, the input string will be truncated.
     * The method will remove previous number if a '-' character is found.
     * @param from
     * @return
     */
    public static String createAccession(String from) {

        String acc = String.valueOf(getNextAccessionNumber());
        int len = acc.length();
        // remove if possible remnant of previous accession
        int pos = from.lastIndexOf("-");
        if(pos!=-1) {
            if(from.equals("-")) {
                return acc.toUpperCase();
            } else if(pos==from.length()-1) {
                from = from.substring(0, pos);
            } else { 
                String after = from.substring(pos+1);
                boolean isNumber = true;
                for(int i = 0; i<after.length(); i++) {
                    if(!Character.isDigit(after.charAt(i))) {
                        isNumber = false;
                        break;
                    }
                }
                if(isNumber) {
                    if(from.length()-pos >= len)
                        from = from.substring(0, pos);
                }
            }
        }
        return (from.substring(0, Math.min(from.length(), 24-len-1)) + "-" + acc).toUpperCase();
    }

    /** Create a RichSequence with a different Identifier; a reference 
     * to the original sequence is kept internally and no sequence
     *  attributes are cloned. The division of the shadow is null; the
     * date note returns today.
     *
     * @param seq Sequence to make a shadow from
     * @param rid Identifier
     * @return A serializable RichSequence
     */
    public static SerializableRichSequence createRichSequenceShadow(RichSequence seq, RichIdentifier rid) {
        return new DelegateRichSequence(seq,rid);
    }
    
    /** Create a serializable RichSequence.
     *
     * @param seq Sequence to make a shadow from
     * @return A serializable RichSequence
     */
    public static SerializableRichSequence createRichSequenceShadow(RichSequence seq) {
        return new DelegateRichSequence(seq,RichIdentifier.create(seq));
    }

    private static class DelegateRichSequence implements SerializableRichSequence {
        private transient RichSequence seq;
        private RichIdentifier rid;

        DelegateRichSequence(RichSequence seq, RichIdentifier rid) {
            this.seq = seq;
            this.rid = rid;
        }

        private void readObject(ObjectInputStream ois)
                throws IOException, ClassNotFoundException {
            try {
                String s1 = (String) ois.readObject();
                int type = SeqType.getTypeForString(s1);
                
                String s2 = (String) ois.readObject();
                SymbolTokenization token = null;
                if(type==SeqType.TYPE_DNA)
                    token = DNATools.getDNA().getTokenization("token");
                else if(type==SeqType.TYPE_RNA)
                    token = RNATools.getRNA().getTokenization("token");
                else
                    token = ProteinTools.getTAlphabet().getTokenization("token");
                RichSequenceFormat genbank = new GenbankFormat();
                RichSequenceBuilderFactory factory = RichSequenceBuilderFactory.THRESHOLD;
                //Namespace serial = (Namespace) RichObjectFactory.getObject(
                //        SimpleNamespace.class,
                //       new Object[]{"bjserial"});
                Namespace serial = RichObjectFactory.getDefaultNamespace();
                
                //BufferedReader br = new BufferedReader(
                //            new CharArrayReader(s3.toCharArray(),0,s3.length()));
                BufferedReader input = new BufferedReader(new StringReader(s2));
                RichStreamReader seqsIn = new RichStreamReader(input,genbank,token,factory,serial);
                seq = seqsIn.nextRichSequence();

            } catch (BioException ex) {
                LOGGER.log(Level.INFO, "", ex);
            }
        }

        private void writeObject(ObjectOutputStream oos)
                throws IOException {

            int type = SeqType.convertSequencetoType(seq);
            oos.writeObject(SeqType.getStringType(type));
            RichSequenceFormat genbank = new GenbankFormat();
            //Namespace serial = (Namespace) RichObjectFactory.getObject(
            //        SimpleNamespace.class,
            //        new Object[]{"bjserial"});
            Namespace serial = RichObjectFactory.getDefaultNamespace();
            
            //StringOutputStream sos = new StringOutputStream();
            //RichStreamWriter seqsOut = new RichStreamWriter(sos, genbank);
            //RichSequenceIterator it = new SingleRichSeqIterator(seq);
            //seqsOut.writeStream(it,serial);
            //oos.writeObject(sos.toString());
             RichStreamWriter seqsOut = new RichStreamWriter(oos,genbank);
             RichSequenceIterator it = new RichSequence.IOTools.SingleRichSeqIterator(this);
             seqsOut.writeStream(it,RichObjectFactory.getDefaultNamespace()); 
        }

        @Deprecated
        @Override
        public void addChangeListener(ChangeListener cl) {
            seq.addChangeListener(cl);
        }

        @Override
        public void addChangeListener(ChangeListener cl, ChangeType ct) {
            seq.addChangeListener(cl, ct);
        }

        @Deprecated
        @Override
        public void removeChangeListener(ChangeListener cl) {
            seq.removeChangeListener(cl);
        }

        @Override
        public void removeChangeListener(ChangeListener cl, ChangeType ct) {
            seq.removeChangeListener(cl, ct);
        }

        @Override
        public boolean isUnchanging(ChangeType ct) {
            return seq.isUnchanging(ct);
        }

        @Override
        public Namespace getNamespace() {
            return rid.getNamespace();
        }

        @Override
        public String getName() {
            return rid.getName();
        }

        @Override
        public String getAccession() {
            return rid.getAccession();
        }

        @Override
        public String getIdentifier() {
            return seq.getIdentifier();
        }

        @Override
        public void setIdentifier(String identifier) throws ChangeVetoException {
            seq.setIdentifier(identifier);
        }

        @Override
        public String getDivision() {
            return seq.getDivision();
        }

        @Override
        public void setDivision(String division) throws ChangeVetoException {
            seq.setDivision(division);
        }

        @Override
        public String getDescription() {
            return rid.getDescription();
        }

        @Override
        public void setDescription(String description) throws ChangeVetoException {
            rid.setDescription(description);
        }

        @Override
        public int getVersion() {
            return rid.getVersion();
        }

        @Override
        public NCBITaxon getTaxon() {
            return seq.getTaxon();
        }

        @Override
        public void setTaxon(NCBITaxon taxon) throws ChangeVetoException {
            seq.setTaxon(taxon);
        }

        @Override
        public Set<RankedDocRef> getRankedDocRefs() {
            return seq.getRankedDocRefs();
        }

        @Override
        public Set<Comment> getComments() {
            return seq.getComments();
        }

        @Override
        public Set<BioEntryRelationship> getRelationships() {
            return seq.getRelationships();
        }

        @Override
        public void addRankedDocRef(RankedDocRef docref) throws ChangeVetoException {
            seq.addRankedDocRef(docref);
        }

        @Override
        public void removeRankedDocRef(RankedDocRef docref) throws ChangeVetoException {
            seq.removeRankedDocRef(docref);
        }

        @Override
        public void addComment(Comment comment) throws ChangeVetoException {
            seq.addComment(comment);
        }

        @Override
        public void removeComment(Comment comment) throws ChangeVetoException {
            seq.removeComment(comment);
        }

        @Override
        public void addRelationship(BioEntryRelationship relation) throws ChangeVetoException {
            seq.addRelationship(relation);
        }

        @Override
        public void removeRelationship(BioEntryRelationship relation) throws ChangeVetoException {
            seq.removeRelationship(relation);
        }

        @Override
        public Set<Note> getNoteSet() {
            Set<Note> notes = seq.getNoteSet();
            Set<Note> ns = new TreeSet<>();
            Iterator<Note> itn = notes.iterator();
            while (itn.hasNext()) {
                Note n = itn.next();
                String value = n.getValue();
                if (value.length() == 11) { //dd-MMM-yyyy
                    Matcher m = p.matcher(value);
                    if (m.matches()) {
                        Date now = new Date();
                        SimpleDateFormat format = new SimpleDateFormat("dd-MMM-yyyy");
                        n = new SimpleNote(
                                n.getTerm(),
                                format.format(now).toUpperCase(),
                                n.getRank());
                    }
                }
                ns.add(n);
            }
            return ns;
        }

        @Override
        public void setNoteSet(Set<Note> notes) throws ChangeVetoException {
            seq.setNoteSet(notes);
        }

        @Override
        public Annotation getAnnotation() {
            return seq.getAnnotation();
        }

        @Override
        public RichAnnotation getRichAnnotation() {
            return seq.getRichAnnotation();
        }

        @Override
        public Set<RankedCrossRef> getRankedCrossRefs() {
            return seq.getRankedCrossRefs();
        }

        @Override
        public void setRankedCrossRefs(Set<RankedCrossRef> crossrefs) throws ChangeVetoException {
            seq.setRankedCrossRefs(crossrefs);
        }

        @Override
        public void addRankedCrossRef(RankedCrossRef crossref) throws ChangeVetoException {
            seq.addRankedCrossRef(crossref);
        }

        @Override
        public void removeRankedCrossRef(RankedCrossRef crossref) throws ChangeVetoException {
            seq.removeRankedCrossRef(crossref);
        }

        @Override
        public int compareTo(Object o) {
            return seq.compareTo(o);
        }

        @Override
        public Double getSeqVersion() {
            return seq.getSeqVersion();
        }

        @Override
        public void setSeqVersion(Double seqVersion) throws ChangeVetoException {
            seq.setSeqVersion(seqVersion);
        }

        @Override
        public Set<Feature> getFeatureSet() {
            return seq.getFeatureSet();
        }

        @Override
        public void setFeatureSet(Set<Feature> features) throws ChangeVetoException {
            seq.setFeatureSet(features);
        }

        @Override
        public void setCircular(boolean circular) throws ChangeVetoException {
            seq.setCircular(circular);
        }

        @Override
        public boolean getCircular() {
            return seq.getCircular();
        }

        @Override
        public SymbolList getInternalSymbolList() {
            return seq.getInternalSymbolList();
        }

        @Override
         public String getURN() {
            return seq.getURN();
        }

        @Override
        public Alphabet getAlphabet() {
            return seq.getAlphabet();
        }

        @Override
        public int length() {
            return seq.length();
        }

        @Override
        public Symbol symbolAt(int index) throws IndexOutOfBoundsException {
            return seq.symbolAt(index);
        }

        @Override
        public List<Symbol> toList() {
            return seq.toList();
        }

        @Override
        public Iterator<Symbol> iterator() {
            return seq.iterator();
        }

        @Override
        public SymbolList subList(int start, int end) throws IndexOutOfBoundsException {
            return seq.subList(start, end);
        }

        @Override
        public String seqString() {
            return seq.seqString();
        }

        @Override
        public String subStr(int start, int end) throws IndexOutOfBoundsException {
            return seq.subStr(start, end);
        }

        @Override
        public void edit(Edit edit) throws IndexOutOfBoundsException, IllegalAlphabetException, ChangeVetoException {
            seq.edit(edit);
        }

        @Override
        public int countFeatures() {
            return seq.countFeatures();
        }

        @Override
        public Iterator<Feature> features() {
            return seq.features();
        }

        @Override
        public FeatureHolder filter(FeatureFilter fc, boolean recurse) {
            return seq.filter(fc, recurse);
        }

        @Override
        public FeatureHolder filter(FeatureFilter filter) {
            return seq.filter(filter);
        }

        @Override
        public Feature createFeature(Feature.Template ft) throws BioException, ChangeVetoException {
            return seq.createFeature(ft);
        }

        @Override
        public void removeFeature(Feature f) throws ChangeVetoException, BioException {
            seq.removeFeature(f);
        }

        @Override
        public boolean containsFeature(Feature f) {
            return seq.containsFeature(f);
        }

        @Override
        public FeatureFilter getSchema() {
            return seq.getSchema();
        }

    }

    /** Create a sequence selection that implements SelectionSequence.
     *
     * @param seq      The sequence that represents the selection;
     * @param rid      An identifier to create a shadow of it;
     * @param parent   The parent sequence;
     * @param location A contiguous location on the parent sequence that represents
     *                 the selection.
     * @return null if the selection is not contiguous, otherwise a SelectionSequence.
     */
    public static RichSequence createSelectionSequence(RichSequence seq, RichIdentifier rid, RichSequence parent, RichLocation loc) {
        assert loc.isContiguous();
        if(!loc.isContiguous())
            return null;
        return new DelegateSelectionSequence(seq,rid,parent,loc);
    }

    private static class DelegateSelectionSequence extends DelegateRichSequence implements SelectionSequence {
        private RichSequence seq;
        private RichSequence parent;
        private RichLocation location;

        DelegateSelectionSequence(RichSequence seq, RichIdentifier rid, RichSequence parent, RichLocation location) {
            super(seq,rid);
            this.parent  = parent;
            this.location = location;
            assert location.isContiguous();
        }
        
        @Override
        public RichLocation getSelectionLocation() {
            return location;
        }

        @Override
        public RichSequence getParent() {
            return parent;
        }

        private void readObject(ObjectInputStream ois)
                throws IOException, ClassNotFoundException {
            
// See createEmptyRichSequence(alphabet)
//            int type = SeqType.convertSequencetoType(seq);
//            Alphabet alphabet = null;
//            if (type == SeqType.TYPE_DNA) {
//                alphabet = DNATools.getDNA();
//            } else if (type == SeqType.TYPE_RNA) {
//                alphabet = RNATools.getRNA();
//            } else {
//                alphabet = ProteinTools.getTAlphabet();
//            }
//            parent = RichSequenceFactory.createEmptyRichSequence(alphabet);

            parent = RichSequenceFactory.createEmptyRichSequence();
            location = RichLocation.EMPTY_LOCATION;
        }

        private void writeObject(ObjectOutputStream oos)
                throws IOException {
        }
        
    }
     
    /** Create an empty RichSequence with an empty SymbolList.
     * @return An empty RichSequence
     */
    public static RichSequence createEmptyRichSequence() {
        RichIdentifier rid = RichIdentifier.enrich(Identifier.create(""));
        RichSequence rseq = new SimpleRichSequence(
                    rid.getNamespace(),
                    rid.getName(),
                    rid.getAccession(),
                    rid.getVersion(),
                    SymbolList.EMPTY_LIST,
                    0.0);
        return rseq;
    }

//Error prone. BioJava always assumes empty lists use SymbolList.EMPTY_LIST
//    /** Create an empty RichSequence of a given Alphabet.
//     * 
//     * @param alphabet the alphabet
//     * @return A RichSequence
//     */
//    public static RichSequence createEmptyRichSequence(Alphabet alphabet) {
//        RichIdentifier rid = RichIdentifier.enrich(Identifier.create(""));
//        RichSequence rseq = new SimpleRichSequence(
//                    rid.getNamespace(),
//                    rid.getName(),
//                    rid.getAccession(),
//                    rid.getVersion(),
//                    createEmptySymbolList(alphabet),
//                    0.0);
//        return rseq;
//    }
    
    /** Create an empty SymbolList of a given Alphabet.
     */
    public static SymbolList createEmptySymbolList(Alphabet alphabet) {
        return new EmptySymbolList(alphabet);
    }
    
    private static class EmptySymbolList extends Unchangeable implements SymbolList, Serializable {

        private Alphabet alphabet;

        private EmptySymbolList() {
            this(Alphabet.EMPTY_ALPHABET);
        }
        
        private EmptySymbolList(Alphabet alphabet) {
            this.alphabet = alphabet;
        }

        @Override
        public Alphabet getAlphabet() {
            return alphabet;
        }

        @Override
        public int length() {
            return 0;
        }

        @Override
        public Symbol symbolAt(int index) throws IndexOutOfBoundsException {
            throw new IndexOutOfBoundsException("Attempted to retrieve symbol from empty list at " + index);
        }

        @Override
        public List toList() {
            return Collections.EMPTY_LIST;
        }

        @Override
        public Iterator iterator() {
            return Collections.EMPTY_LIST.iterator();
        }

        @Override
        public SymbolList subList(int start, int end) throws IndexOutOfBoundsException {
            return new EmptySymbolList(alphabet);
        }

        @Override
        public String seqString() {
            return "";
        }

        @Override
        public String subStr(int start, int end) throws IndexOutOfBoundsException {
            throw new IndexOutOfBoundsException(
                    "You can not retrieve part of an empty symbol list");
        }

        @Override
        public void edit(Edit edit) throws IndexOutOfBoundsException, ChangeVetoException {
            throw new ChangeVetoException(
                    "You can't edit the empty symbol list");
        }
    }
    
    public static String getDivision(SymbolList sym) {
        String division = TaxonomyRoot;
        if(sym instanceof RichSequence) {
            RichSequence rs = (RichSequence) sym;
            String div = rs.getDivision();
            if(div!=null && !div.isEmpty())
                division = div;
        }
        if(sym instanceof SelectionSequence) {
            SelectionSequence sseq = (SelectionSequence) sym;
            String div = sseq.getDivision();
            if(div!=null && !div.isEmpty())
                division = div;
        }
        return division;
    }
    
    /** Create a SimpleRichSequence at the date note of today.
     * 
     */
    public static RichSequence createSimpleRichSequence(RichIdentifier rid, SymbolList symList) {
        RichSequence rseq = new SimpleRichSequence(rid.getNamespace(),rid.getName(),rid.getAccession(),rid.getVersion(),symList,null);
        rseq.setDescription(rid.getDescription());
        RichAnnotation notes = rseq.getRichAnnotation();
        Date now = new Date();
        SimpleDateFormat format = new SimpleDateFormat("dd-MMM-yyyy");
        Note n = new SimpleNote(
                RichSequence.Terms.getDateCreatedTerm(),
                format.format(now).toUpperCase(),
                notes.getNoteSet().size()+1);
        notes.addNote(n);
        n = new SimpleNote(
                RichSequence.Terms.getDateUpdatedTerm(),
                format.format(now).toUpperCase(),
                notes.getNoteSet().size()+1);
        notes.addNote(n);
        
        // Division
        rseq.setDivision(RichSequenceFactory.getDivision(symList));
        return rseq;
    }
    
    /** Get reverse complement SymbolList.
     *
     * @param sym a non-empty DNA or RNA SymbolList or an empty SymbolList.
     */
    public static SymbolList getReverseComplement(SymbolList sym) {
        if(sym==null || sym.length()==0) 
            return SymbolList.EMPTY_LIST;
        SymbolList revcomp = null;
        try {
            if(SeqType.isDNA(sym))
                revcomp = DNATools.reverseComplement(sym);
            else if(SeqType.isRNA(sym))
                revcomp = RNATools.reverseComplement(sym);
            else
                return SymbolList.EMPTY_LIST;
        } catch (IllegalAlphabetException ex) {
            return SymbolList.EMPTY_LIST;
        }
        return revcomp;
    }
    
    /** Get complement SymbolList.
     *
     * @param sym a non-empty DNA or RNA SymbolList or an empty SymbolList.
     */
    public static SymbolList getComplement(SymbolList sym) {
        if(sym==null || sym.length()==0)
            return SymbolList.EMPTY_LIST;
        SymbolList comp = null;
        try {
            if(SeqType.isDNA(sym))
                comp = DNATools.complement(sym);
            else if(SeqType.isRNA(sym))
                comp = RNATools.complement(sym);
            else
                return SymbolList.EMPTY_LIST;
        } catch (IllegalAlphabetException ex) {
            return SymbolList.EMPTY_LIST;
        }
        return comp;
    }
    
    /** Complement a sequence and annotate it. */
    public static RichSequence getComplementSequence(Sequence seq) {
        RichIdentifier rid = RichIdentifier.create(seq);
        String dname = rid.getDisplayName();
        String nname = dname;
        nname = RichSequenceFactory.createAccession(nname);
        String des = rid.getDescription();
        SymbolList sym = RichSequenceFactory.getComplement(seq);
        String mes = NbBundle.getMessage(RichSequenceFactory.class, "MES_comp");
        mes += " " + dname;
        des = mes; // Do not add previous description when sequence is manipulated
        RichIdentifier rid2 = RichIdentifier.create(
                rid.getNamespace(),
                nname,
                nname, 1);
        rid2.setDescription(des);
        RichSequence rseq = RichSequenceFactory.createSimpleRichSequence(rid2, sym);
        RichSequenceFactory.addComment(rseq, des, true);
        // Division from the parent if any
        rseq.setDivision(RichSequenceFactory.getDivision(seq));
        return rseq;
    }
    
    /** Get reverse complement SymbolList.
     *
     * @param sym a non-empty DNA or RNA SymbolList or an empty SymbolList.
     */
    public static SymbolList getReverse(SymbolList sym) {
        if (sym == null || sym.length() == 0)
            return SymbolList.EMPTY_LIST;
        List<Symbol> syms = sym.toList();
        int len = syms.size();
        Symbol[] sys = new Symbol[len];
        for (int i = 0; i < len; i++) {
            sys[len - i-1] = syms.get(i);
        }
        return new SimpleSymbolList(sys, len, sym.getAlphabet());

    }
    
    /** Reverse Reverse a sequence and annotate it. */
    public static RichSequence getReverseSequence(Sequence seq) {
        RichIdentifier rid = RichIdentifier.create(seq);
        String dname = rid.getDisplayName();
        String nname = dname;
        nname = RichSequenceFactory.createAccession(nname);
        String des = rid.getDescription();
        SymbolList sym = RichSequenceFactory.getReverse(seq);
        String mes = NbBundle.getMessage(RichSequenceFactory.class, "MES_rev");
        mes += " " + dname;
        des = mes; // Do not add previous description when sequence is manipulated
        RichIdentifier rid2 = RichIdentifier.create(
                rid.getNamespace(),
                nname,
                nname, 1);
        rid2.setDescription(des);
        RichSequence rseq = RichSequenceFactory.createSimpleRichSequence(rid2, sym);
        RichSequenceFactory.addComment(rseq, des, true);
        // Division from the parent if any
        rseq.setDivision(RichSequenceFactory.getDivision(seq));
        return rseq;
    }
    
    /** Clone a RichSequence.
     * @param sequence The sequence to clone
     * @return The cloned sequence.
     */
    public static RichSequence cloneSequence(Sequence sequence) {
        RichSequence seq = RichSequence.Tools.enrich(sequence);
        SymbolList symList;
        if(seq.getInternalSymbolList()==SymbolList.EMPTY_LIST) {
            symList = SymbolList.EMPTY_LIST;
        } else {
            symList = new SimpleSymbolList(seq.getInternalSymbolList());
        }
        RichSequence s = new SimpleRichSequence(
                        seq.getNamespace(),
                        seq.getName(),
                        seq.getAccession(),
                        seq.getVersion(),
                        symList,
                        seq.getSeqVersion()
                        );
        s.setDescription(seq.getDescription());
        s.setDivision(seq.getDivision());
        s.setIdentifier(seq.getIdentifier());
        s.setCircular(seq.getCircular());
        s.setTaxon(seq.getTaxon());
        
        Set features = seq.getFeatureSet();
        Iterator it = features.iterator();
        Set<Feature> f = new TreeSet<>();
        while(it.hasNext()) {
            f.add(FeatureUtil.cloneFeature(s,(Feature)it.next()));
        }
        s.setFeatureSet(f);
        
        Set<Note> notes = seq.getNoteSet();
        Iterator<Note> nit = notes.iterator();
        Set<Note> n = new TreeSet<>();
        while(nit.hasNext()) {
            Note note = nit.next();
            n.add(NoteUtil.cloneNote(note));
        }
        s.setNoteSet(n);
        
        Set comments = seq.getComments();
        it = comments.iterator();
        while(it.hasNext()) {
            Comment com = (Comment)it.next();
            s.addComment(new SimpleComment(
                    com.getComment(),
                    com.getRank()));
        }
        
        Set rankedcrossrefs = seq.getRankedCrossRefs();
        it = rankedcrossrefs.iterator();
        while(it.hasNext()) {
            RankedCrossRef ref = (RankedCrossRef)it.next();
            s.addRankedCrossRef(new SimpleRankedCrossRef(
                    ref.getCrossRef(), //same as (CrossRef)RichObjectFactory.getObject(SimpleCrossRef.class,new Object[]{...
                    ref.getRank()));
        }
        
        Set rankeddocrefs = seq.getRankedDocRefs();
        it = rankeddocrefs.iterator();
        while(it.hasNext()) {
            RankedDocRef ref = (RankedDocRef)it.next();
            s.addRankedDocRef(new SimpleRankedDocRef(
                    ref.getDocumentReference(), //same as (DocRef)RichObjectFactory.getObject(SimpleDocRef.class,new Object[]{...
                    ref.getStart(),
                    ref.getEnd(),
                    ref.getRank())); 
        }
        
        Set relations = seq.getRelationships();
        it = relations.iterator();
        while(it.hasNext()) {
            BioEntryRelationship rel = (BioEntryRelationship)it.next();
            BioEntryRelationship clonerel =  new SimpleBioEntryRelationship(
                    s, 
                    rel.getSubject(), 
                    rel.getTerm(), 
                    rel.getRank());
            s.addRelationship(clonerel);
        }
        return s;
    }
    
    /** Add some text to the Comment of a RichSequence. A Comment will be 
     * created if the sequence has no Comment or the first Comment will be used.
     * Text will be added up front.
     * 
     * @param rseq The input sequence
     * @param comment  The String to add
     * @param separate If true, a '\n' character will be appended after.
     * @return 
     */
    public static boolean addComment(RichSequence rseq, String comment, boolean separate) {
        Set<Comment> coms = rseq.getComments();
        Comment newcom = null;
        Comment oldcom = null;
        if(comment==null) {
            comment = ""; // Cannot be null
        }
        if(coms.isEmpty()) {
            newcom = new SimpleComment(comment,1);
        } else {
            Iterator<Comment> cit = coms.iterator();
            oldcom = cit.next();
            String ctext = oldcom.getComment();
            if(comment.endsWith("\n")) {
                ctext = comment + ctext;
            } else if(separate) {
                ctext = comment + "\n" + ctext;
            } else {
                ctext = comment + ctext;
            }
            
            // Old fashion
            //if(ctext.endsWith("\n")) {
            //    ctext += comment;
            //} else if(separate) {
            //    ctext += "\n" + comment;
            //} else {
            //    ctext += comment;
            //}
            
            newcom = new SimpleComment(ctext,1);
        }
        try {
            if(oldcom!=null)
                rseq.removeComment(oldcom);
            rseq.addComment(newcom);
        } catch (ChangeVetoException cve) {
            return false;
        }
        return true;
    }
}
