/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or bbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * bbbuild/licenses/cddl-gpl-2-cp.txt.  This particular file is subject to
 * the "Classpath" exception as provided in the GPL Version 2 section of the
 * License file that accompanied this code. If applicable, add the following
 * below the License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 * Portions Copyrighted 2017 George Waldon
 */

package org.biobeans.modules.biocore.sequences;


import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.RNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.openide.util.NbBundle;

import org.biojava.bio.seq.io.SeqIOConstants;
import org.biojava.bio.symbol.Alphabet;
import org.biojava.bio.symbol.AlphabetManager;
import org.biojava.bio.symbol.FiniteAlphabet;


/** Helper for sequences based on the four main symbolic formats
 * (DNA, RNA, AA, and UNKNOWN) with localization, alphabet creation
 * and type determination; also provide conversion with SeqIOConstants.
 * In most cases, SeqIOConstants constants are preferred, but
 * when unambigous values are needed, seqtypes are better. 
 *
 * @author  George Waldon
 */
public class SeqType {
    
    /** Representation of Unknown type (value 0).
     */
    public static final int TYPE_UNKNOWN = 0;
    
    /** Representation of DNA (value 1).
     */
    public static final int TYPE_DNA = 1;
    
    /** Representation of RNA (value 2).
     */
    public static final int TYPE_RNA = 2;
    
    /** Representation of AA (value 3).
     */
    public static final int TYPE_PROTEIN = 3;

    /** Localized strings to describe SeqType.
     */
    private static final String[] tags;
    
    /** Internal strings to describe the four main symbolic formats
     * (DNA, RNA, AA, and UNKNOWN).
     *
     *  Used to set file attribute, manage ComboBox, etc.
     */
    private static final String[] seqtype;
    
    static {
        tags = new String[4];
        tags[TYPE_UNKNOWN] = NbBundle.getMessage(SeqType.class, "TYPE_unknown");
        tags[TYPE_DNA] =     NbBundle.getMessage(SeqType.class, "TYPE_dna");
        tags[TYPE_RNA] =     NbBundle.getMessage(SeqType.class, "TYPE_rna");
        tags[TYPE_PROTEIN] = NbBundle.getMessage(SeqType.class, "TYPE_protein");

        seqtype = new String[4];
        seqtype[TYPE_UNKNOWN] = "unknown";
        seqtype[TYPE_DNA] =     "DNA";
        seqtype[TYPE_RNA] =     "RNA";
        seqtype[TYPE_PROTEIN] = "Protein";
    }
    
    /**
     *
     * @return number of SeqType (= 4).
     */
    public static int getSize() {
        return seqtype.length;
    } 
    
    /** Validity check for SeqType.
     *
     * @return true if value corresponds to known SeqType
     */
    public static boolean checkTypeValue(int value) {
        return !(value<TYPE_UNKNOWN || value>TYPE_PROTEIN);
    }
    
    /** Get all SeqType description.
     *
     * @return localized tags that describe SeqType.
     */
    public static String[] getTags() {
        return tags;
    }
    
    /** Get a description for a SeqType.
     * 
     * @param a SeqType
     * @return the localized description of the sequence type
     */
    public static String getTagFromType(int type) {
        try {
            return tags[type];
        }
        catch (ArrayIndexOutOfBoundsException aob) {
            return tags[TYPE_UNKNOWN];
        }
    }
    
    /** Get a description for a sequence format.
     * 
     * @param a BioJava SeqIOConstants sequence format
     * @return the localized description of the sequence type
     */
    public static String getTagFromFormat(int format) {
        return getTagFromType(convertFormattoType(format));
    }
    
    /** Convert a BioJava SeqIOConstants format into a SeqType. If the input encodes
     * several formats, use {@Link #convertFormattoTypeArray()} instead.
     *
     * @param a BioJava SeqIOConstants sequence format
     * @return a SeqType
     */
    public static int convertFormattoType (int format) {
        if( (format & SeqIOConstants.DNA)!= 0) {
            return TYPE_DNA;
        } else if( (format & SeqIOConstants.RNA)!= 0) {
            return TYPE_RNA;
        } else if( (format & SeqIOConstants.AA)!= 0) {
            return TYPE_PROTEIN;
        } else
            return TYPE_UNKNOWN;
    }
    
    /** Convert a BioJava sequence format into an array of SeqTypes.
     *  If no formats are found, an array containing  <CODE> Seqtype.TYPE_UNKNOWN </CODE>
     * is returned.
     *
     * This is used e.g. when sequences of varied formats are queried.
     *
     * @param a BioJava sequence format
     * @return a SeqType
     */    
    public static int[] convertFormattoTypeArray(int format) {
        int n = 0;
        if( (format & SeqIOConstants.DNA)!= 0)
            n++;
        if( (format & SeqIOConstants.RNA)!= 0)
            n++;
        if( (format & SeqIOConstants.AA)!= 0)
            n++;
        if(n==0)
            return new int[] {TYPE_UNKNOWN};
            
        int[] intarr = new int[n];
        n = 0;
        if( (format & SeqIOConstants.DNA)!= 0) {
            intarr[n] = TYPE_DNA;
            n++;
        }
        if( (format & SeqIOConstants.RNA)!= 0) {
            intarr[n] = TYPE_RNA;
            n++;
        }
        if( (format & SeqIOConstants.AA)!= 0) {
            intarr[n] = TYPE_PROTEIN;
            n++;
        }
        return intarr;
    }

    
    /** Get a sequence format for a SeqType.
     * 
     * @param   a SeqType
     * @return a BioJava <code>SeqIOConstants</code> sequence format
     */
    public static int convertTypetoFormat (int type) {
        switch(type) {
            case TYPE_DNA:      return SeqIOConstants.DNA;
            case TYPE_RNA:      return SeqIOConstants.RNA;
            case TYPE_PROTEIN:  return SeqIOConstants.AA;
        }
        return SeqIOConstants.UNKNOWN;
    }
    
    /** Get strings for all SeqType.
     *
     * @return SeqTypes as array of Strings.
     */
    public static String[] getStringTypes() {
        return seqtype;
    }
    
    /** Get a string for a SeqType.
     *
     * @return string for SeqType.
     */
    public static String getStringType(int type) {
        try {
            return seqtype[type];
        }
        catch (ArrayIndexOutOfBoundsException aob) {
            return seqtype[TYPE_UNKNOWN];
        }
    }
    
    /** Get SeqType for a string.
     *
     * @return SeqType for string.
     */
    public static int getTypeForString(String type) {
        for(int i=0; i<getSize(); i++) {
            if( type.equals(seqtype[i]))
                return i;
        }
        return TYPE_UNKNOWN;
    }
    
    /** Get localized alphabet name.
     *
     * @return localizable name for Alphabet
     */
    public static String getAlphabetName(FiniteAlphabet fa) {
        if(fa==null)
            return tags[TYPE_UNKNOWN];
        String name = fa.getName();
        switch (name) {
            case "DNA":
                return tags[TYPE_DNA];
            case "PROTEIN":
            case "PROTEIN-TERM":
                return tags[TYPE_PROTEIN];
            case "RNA":
                return tags[TYPE_RNA];
        }
        return tags[TYPE_UNKNOWN];
    }
    
    /** Get Alphabet for SeqType.
     * 
     * @return Alphabet corresponding to the SeqType
     * @throw java.util.NoSuchElementException if type is
     *  out of bound, or is TYPE_UNKNOWN, or Alphabet is not found
     */
    public static Alphabet getAlphabetForType(int type)
    throws java.util.NoSuchElementException{
        if( type == SeqType.TYPE_DNA) {
            return AlphabetManager.alphabetForName("DNA");
        } else if( type == SeqType.TYPE_PROTEIN) {
            return AlphabetManager.alphabetForName("PROTEIN-TERM");
        } else if( type == SeqType.TYPE_RNA) {
            return AlphabetManager.alphabetForName("RNA");
        } else {
            throw new java.util.NoSuchElementException();
        }
    }
    
    /** Get alphabet for sequence format; if input is a mix of
     * several formats, result might be incorrect.
     *
     * @param format A SeqIOConstants constant
     * @return Alphabet corresponding to the SeqIOConstants format
     * @throw java.util.NoSuchElementException if type is
     *  out of bound, or is UNKNOWN, or Alphabet is not found
     */
    public static Alphabet getAlphabetForFormat(int format)
    throws java.util.NoSuchElementException{
        if( (format&SeqIOConstants.DNA)!=0 ) {
            return AlphabetManager.alphabetForName("DNA");
        } else if( (format&SeqIOConstants.AA)!=0 ) {
            return AlphabetManager.alphabetForName("PROTEIN-TERM");
        } else if( (format&SeqIOConstants.RNA)!=0 ) {
            return AlphabetManager.alphabetForName("RNA");
        } else {
            throw new java.util.NoSuchElementException();
        }
    }
    
    /** get SeqType for alphabet.
     * 
     * @param alph input Alphabet
     *@return SeqType corresponding to alphabet name
     */
    public static int getTypeFromAlphabet(Alphabet alph) {
        if(alph==null)
            return SeqType.TYPE_UNKNOWN;
        String name = alph.getName();
        if(name.equals("DNA"))
            return SeqType.TYPE_DNA;
        if(name.equals("RNA"))
            return SeqType.TYPE_RNA;
        if(name.equals("PROTEIN") || name.equals("PROTEIN-TERM"))
            return SeqType.TYPE_PROTEIN;
        return SeqType.TYPE_UNKNOWN;
    }
    
    /** get format for alphabet.
     * 
     * @param alph input Alphabet
     *@return  {@link SeqIOConstants} constant corresponding to alphabet name
     */
    public static int getFormatFromAlphabet(Alphabet alph) {
        if(alph==null)
            return SeqIOConstants.UNKNOWN;
        String name = alph.getName();
        if(name.equals("DNA"))
            return SeqIOConstants.DNA;
        if(name.equals("RNA"))
            return SeqIOConstants.RNA;
        if(name.equals("PROTEIN") || name.equals("PROTEIN-TERM"))
            return SeqIOConstants.AA;
        return SeqIOConstants.UNKNOWN;
    }
   
    /** Get the format that describes multiple sequences.
     * 
     * @param seqs an array of {@link Sequence}
     * @return int represent a combination of all the formats that these sequences
     *  represent.
     */
    public static int convertSequencestoFormat(Sequence[] seqs) {
        int format = SeqIOConstants.UNKNOWN;
        if(seqs!=null) {
            for (Sequence seq : seqs) {
                format &= getFormatFromAlphabet(seq.getAlphabet());
            }
        }
        return format;
    }
    
    /** Get SeqType for a symbol list. 
     * 
     * @param sym
     * @return 
     */
    public static int convertSequencetoType(SymbolList sym) {
        if(isDNA(sym))
            return TYPE_DNA;
        if(isRNA(sym))
            return TYPE_RNA;
        if(isPROTEIN(sym))
            return TYPE_PROTEIN;
        return TYPE_UNKNOWN;
    }
    
    /** Test whether a SymbolList is made of DNA.
     */
    public static boolean isDNA(SymbolList sym) {
        if(sym==null)
            throw new IllegalArgumentException("The symbol list was null");
        Alphabet alph = sym.getAlphabet();
        if(alph==null)
            throw new IllegalArgumentException("The alphabet was null");
        int type = getTypeFromAlphabet(alph);
        if(type==TYPE_DNA)
            return true;
        return false;
    }
    
    /** Test whether a Symbol is part of DNA alphabet.
     */
    public static boolean isDNA(Symbol s) {
        if(s==null)
            throw new IllegalArgumentException("The symbol was null");
        Alphabet alph = DNATools.getDNA();
        return alph.contains(s);
    }
    
    /** Test whether a SymbolList is made of RNA.
     */
    public static boolean isRNA(SymbolList sym) {
        if(sym==null)
            throw new IllegalArgumentException("The symbol list was null");
        Alphabet alph = sym.getAlphabet();
        if(alph==null)
            throw new IllegalArgumentException("The alphabet was null");
        int type = getTypeFromAlphabet(alph);
        if(type==TYPE_RNA)
            return true;
        return false;
    }
    
    /** Test whether a Symbol is part of RNA alphabet.
     */
    public static boolean isRNA(Symbol s) {
        if(s==null)
            throw new IllegalArgumentException("The symbol was null");
        Alphabet alph = RNATools.getRNA();
        return alph.contains(s);
    }
    
    /** Test whether a SymbolList is made of PROTEIN.
     */
    public static boolean isPROTEIN(SymbolList sym) {
        if(sym==null)
            throw new IllegalArgumentException("The symbol list was null");
        Alphabet alph = sym.getAlphabet();
        if(alph==null)
            throw new IllegalArgumentException("The alphabet was null");
        int type = getTypeFromAlphabet(alph);
        if(type==TYPE_PROTEIN)
            return true;
        return false;
    }
    
    /** Test whether a Symbol is part of protein alphabet.
     */
    public static boolean isPROTEIN(Symbol s) {
        if(s==null)
            throw new IllegalArgumentException("The symbol was null");
        Alphabet alph = ProteinTools.getTAlphabet();
        return alph.contains(s);
    }
}