/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
*/
package org.biobeans.modules.bistro.actions.imports;

import java.awt.Toolkit;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Action;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.task.ActivityEvent;
import org.biobeans.modules.bistro.task.ActivityTask;
import org.biobeans.modules.bistro.task.ActivityTaskListener;
import org.biobeans.modules.bistro.services.BioService;
import org.biobeans.modules.bistro.services.BioServiceRegistry;
import org.biobeans.modules.bistro.task.Counter;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.openide.awt.StatusDisplayer;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.util.Lookup;
import org.openide.util.Mutex;
import org.openide.util.NbBundle;
import org.openide.util.actions.CookieAction;

/**
 *
 * @author George Waldon
 */
public abstract class ImportNCBIFolderAbstractAction extends CookieAction {

    private static final Logger LOGGER = Logger.getLogger(ImportNCBIFolderAbstractAction.class.getName());

    abstract String getTaskName();
    
    abstract String getFileExtension();
    
    abstract String getDialogLabel();
    
    @Override
    public boolean asynchronous() {
        return false;
    }
    
    @Override
    protected int mode() {
        return CookieAction.MODE_EXACTLY_ONE;
    }

    @Override
    protected Class<?>[] cookieClasses() {
        return new Class[] {DataFolder.class};
    }

    protected DataFolder selectedFolder( Lookup context ) {   
        DataFolder targetFolder = null;
        //Try to find the target folder
        targetFolder = context.lookup( DataFolder.class );
        if ( targetFolder == null ) {
            //Try with DataObject
            DataObject obj = context.lookup( DataObject.class );
            if ( obj != null) {
                targetFolder = obj.getFolder();
            }
        }
        return targetFolder;
    }
    
    @Override
    protected void performAction(final Node[] activatedNodes) {
        Mutex.EVENT.writeAccess(new Runnable() {
            
            @Override
            public void run() {
                Node des = activatedNodes[0].cloneNode();
                final DataFolder targetFolder = selectedFolder(des.getLookup());
                if(targetFolder==null) {
                    Toolkit.getDefaultToolkit().beep();
                    StatusDisplayer.getDefault().setStatusText(
                            NbBundle.getMessage(ImportGenBankFolderAction.class, "MES_nofolder"));
                    return;
                }
                
                final ActivityTask.Termination tm = new ActivityTask.Termination();
                final ProgressHandle progr = ProgressHandleFactory.createHandle(
                        NbBundle.getMessage(ImportGenBankFolderAction.class, "LBL_NCBIimport"),
                        tm,
                        (Action)null
                        );
                progr.start();
                
                
                Runnable r = new Runnable() {
                    @Override
                    public void run() {
                        BioServiceRegistry registry = Lookup.getDefault().lookup(BioServiceRegistry.class);
                        final BioService service = registry.findServicebyID(getTaskName());
                        ActivityTask task = service.execute();
                        task.addTaskListener(new Transferlistener(targetFolder, progr));
                        task.setTermination(tm);
                    }

                };
                
                Counter.pour(r);
            } 
        });
    }
    
    private class Transferlistener implements ActivityTaskListener {

        DataFolder oriFolder;
        ProgressHandle progr;
        String documentName;
        String seq;

        private Transferlistener(DataFolder targetFolder,ProgressHandle progr) {
            oriFolder = targetFolder;
            this.progr = progr;
        }

        @Override
        public void activityUpdate(ActivityEvent event){
            
            String id = event.getId();
            DataContainer result = event.getData();
            
            if(id.equals("id")) {
                 documentName = (String) result.getData();
            } else if(id.equals("outputText")) {
                Object thing = result.getData();
                if (thing==null || !(thing instanceof String) ) {
                    StatusDisplayer.getDefault().setStatusText(
                            NbBundle.getMessage(ImportGenBankFolderAction.class, "MES_errorfetching"));
                    Toolkit.getDefaultToolkit().beep();
                } else {
                    seq = (String) thing;
                    if (seq.length() < 10) { //bad fetch return "\n"-only containing files
                        StatusDisplayer.getDefault().setStatusText(
                                NbBundle.getMessage(ImportGenBankFolderAction.class, "MES_nothingfetched"));
                        Toolkit.getDefaultToolkit().beep();
                        //TODO somehoe we need to terminate this task and make sure memm=ory is cleared
                        return;
                    }
                }
            }
            
            if(seq!=null && documentName!=null) {
                
                progr.finish();

                if(result==null) {
                    LOGGER.info("Result is null after task finished."); //NOI18N
                    return;
                }
                
                OutputStream out = null;
                PrintStream pout = null;
                FileLock lock = null;
                FileObject newFile = null;
                try {
                    //find a free document name
                    String exportDoc = FileUtil.findFreeFileName(oriFolder.getPrimaryFile(), documentName, getFileExtension());
                    newFile = oriFolder.getPrimaryFile().createData(exportDoc + "." + getFileExtension()); //NOI18N
                    lock = newFile.lock();
                    out = newFile.getOutputStream(lock);
                    pout = new PrintStream(out, false);
                    pout.print(seq);
                } catch (IOException ex) {
                    LOGGER.log(Level.WARNING, "", ex);
                } finally {
                    IOException iox = null;
                    try {
                        if (out != null) {
                            out.flush();
                            out.close();
                        }
                        if (pout != null) {
                            pout.flush();
                            pout.close();
                        }
                    } catch (IOException ex) {
                        iox = ex;
                    }
                    if (lock != null) {
                        lock.releaseLock();
                    }
                    if (iox != null) {
                        LOGGER.log(Level.WARNING, "", iox);
                    }
                }
            }
        }

        @Override
        public void taskFinished(ActivityTask task){
            progr.finish();
        }
    }
}
