/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
*/
package org.biobeans.modules.bistro.activity;

import org.biobeans.modules.bistro.task.ActivityEvent;
import org.biobeans.modules.bistro.task.ActivityTask;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.modules.bistro.activity.rest.HTTPException;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.task.CancelledByUserException;
import org.biobeans.modules.bistro.task.Counter;
import org.biobeans.modules.bistro.task.TaskExecutionException;
import org.openide.awt.StatusDisplayer;
import org.openide.util.NbBundle;

/**
 * @param <ConfigType> type of configuration object to be used to hold
 * configuration information
 */
public abstract class AbstractActivity<ConfigType> implements Activity<ConfigType> {
    
    private static final Logger LOGGER = Logger.getLogger(AbstractActivity.class.getName());
    
    private Set<Port> inputPorts;
    
    private Set<Port> outputPorts;
    
    private Map<String, DataContainer> inputModel;
    
    private Map<ActivityTask,Map<String, DataContainer>> taskMap;
    
    public AbstractActivity() {
        initialize();
    }

    private void initialize() {
        inputPorts = new HashSet<>();
        outputPorts = new HashSet<>();
        taskMap = new HashMap<>();
        inputModel = new HashMap<>();
    }
    
    @Override
    public Set<Port> getInputPorts() {
        return inputPorts;
    }
    
    @Override
    public Port getInputPort(String name) {
        for(Port p: inputPorts) {
            if(name.equals(p.getId()))
                return p;
        }
        return null;
    }
    
    protected void addInputPort(InputPort port) {
        inputPorts.add(port);
        inputModel.put(port.getId(), null);
    }
    
    protected void removeInputPorts() {
        inputPorts.clear();
    }
    
    @Override
    public Set<Port> getOutputPorts() {
        return outputPorts;
    }

    @Override
    public Port getOutputPort(String name) {
        for(Port p: outputPorts) {
            if(name.equals(p.getId()))
                return p;
        }
        return null;
    }
    
    protected void addOutputPort(OutputPort port) {
        outputPorts.add(port);
    }
    
    protected void removeOutputPorts() {
        outputPorts.clear();
    }
    
    private synchronized void cancelTask(ActivityTask task) {
        taskMap.remove(task);
    }
    
    @Override
    public synchronized void notifyActivity(ActivityEvent ae) {
        final ActivityTask task = ae.getTask();
        Map<String, DataContainer> inputMap = taskMap.get(task);
        if(task.isFinished()) {
            taskMap.remove(task);
            return;
        }
        if(inputMap == null) {
            inputMap = new HashMap<>();
            inputMap.putAll(inputModel);
        }
        
        inputMap.put(ae.getId(), ae.getData());
        
        //Wait til the map if full
        for(String pname:inputMap.keySet()) {
            if(inputMap.get(pname)==null) {
                return;
            }
        }
        
        //Remove mapping for this task
        final Map<String, DataContainer> map = inputMap;
        taskMap.remove(task);
        
        //Execute map
        Runnable r = new Runnable() {
            @Override
            public void run() {
                try {
                    Map<String, DataContainer> outputMap = execute(map);
                    for(String portid:outputMap.keySet()) {
                        Port p = getOutputPort(portid);
                        if(p!=null) {
                            ActivityEvent de = new ActivityEvent(task,p.getId(),outputMap.get(portid));
                            ((OutputPort)p).outputData(de);
                        }
                    }
                } catch (TaskExecutionException ex) {
                    if(ex instanceof HTTPException) {
                        String message = "Could not process requet: HTTP Exception with error code " + ((HTTPException)ex).getStatusCode();
                        StatusDisplayer.getDefault().setStatusText(message);
                    } else if(!(ex instanceof CancelledByUserException)) {
                        LOGGER.log(Level.INFO, "Exception during activity execution", ex);
                    }
                    task.cancel();
                    cancelTask(task);
                }
            }
            
        };
        Counter.percolate(r);
    }
    
     protected abstract Map<String, DataContainer> execute(Map<String, DataContainer> inputMap) throws TaskExecutionException;
}
