/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.activity.localactivities.biojava;

import java.io.BufferedReader;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.logging.Logger;
import org.biobeans.modules.biocore.sequences.RichIdentifier;
import org.biobeans.modules.biocore.sequences.RichSequenceFactory;
import org.biobeans.modules.biocore.sequences.SerializableRichSequence;
import org.biobeans.modules.bistro.activity.AbstractActivity;
import org.biobeans.modules.bistro.activity.ActivityConfigurationException;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.data.DataContainerAdapter;
import org.biobeans.modules.bistro.task.TaskExecutionException;
import org.biobeans.modules.bistro.activity.SimpleInputPort;
import org.biobeans.modules.bistro.activity.SimpleOutputPort;
import org.biojava.bio.BioException;
import org.biojavax.RichObjectFactory;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;

/**
 * This processor converts a FASTA record into a BioJava RichSequence.
 *
 * @author George Waldon
 *
 * @input inputText a FASTA record
 */
public class FASTAToBiojavaActivity extends AbstractActivity<FASTAToBiojavaActivityConfigurationBean> {

    private static final Logger LOGGER = Logger.getLogger(FASTAToBiojavaActivity.class.getName());

    public FASTAToBiojavaActivity() {
    }
    
    private FASTAToBiojavaActivityConfigurationBean configBean;

    @Override
    public FASTAToBiojavaActivityConfigurationBean getConfiguration() {
        return this.configBean;
    }

    @Override
    public void configure(FASTAToBiojavaActivityConfigurationBean configBean)
            throws ActivityConfigurationException {
        this.configBean = configBean;
        removeInputPorts();
        removeOutputPorts();
        addInputPort(new SimpleInputPort("inputText","text/plain",0,this));
        addOutputPort(new SimpleOutputPort("outputSeq","application/biojavax-seq",0,this));
    }

    @Override
    public Map<String, DataContainer> execute(Map<String, DataContainer> inputMap) throws TaskExecutionException { 
        DataContainerAdapter inAdapter = new DataContainerAdapter(inputMap);
        String inseq = inAdapter.getString("inputText");
        StringReader sr = new StringReader(inseq);
        BufferedReader br = new BufferedReader(sr); 
        RichSequenceIterator seqs = null;
        if(!configBean.isProteinParser()) {
            seqs = RichSequence.IOTools.readFastaDNA(br,RichObjectFactory.getDefaultNamespace());
        } else {
            seqs = RichSequence.IOTools.readFastaProtein(br,RichObjectFactory.getDefaultNamespace());
        }
        RichSequence rseq = null;
        if (seqs.hasNext()) {
            try {
                rseq = seqs.nextRichSequence();
            } catch (NoSuchElementException | BioException ex) {
                rseq = null;
            }
        }
        if(rseq==null) {
            rseq = RichSequenceFactory.createEmptyRichSequence();
        }
        SerializableRichSequence srseq = RichSequenceFactory.createRichSequenceShadow(
                rseq, 
                RichIdentifier.create(rseq));
        Map<String, DataContainer> outputMap = new HashMap<>();
        DataContainerAdapter dc = new DataContainerAdapter(outputMap);
        dc.putSerializable("outputSeq", srseq);
        return outputMap;
    }

}
