package org.biobeans.modules.bistro.activity.rest;

import org.biobeans.modules.bistro.activity.ActivityConfigurationException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.modules.bistro.activity.AbstractActivity;
import org.biobeans.modules.bistro.activity.SimpleInputPort;
import org.biobeans.modules.bistro.activity.SimpleOutputPort;
import org.biobeans.modules.bistro.activity.rest.HTTPRequestHandler.HTTPRequestResponse;
import org.biobeans.modules.bistro.activity.rest.URISignatureHandler.URISignatureParsingException;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.data.DataContainerAdapter;
import org.biobeans.modules.bistro.data.factory.DataContainerFactory;
import org.biobeans.modules.bistro.task.TaskExecutionException;

/**
 * Generic REST activity that is capable to perform all four HTTP methods.
 *
 * @author Sergejs Aleksejevs, Geroge Waldon
 */
public class RESTActivity extends AbstractActivity<RESTActivityConfigurationBean> {

    private static final Logger LOGGER = Logger.getLogger(RESTActivity.class.getName());

    public static enum HTTP_METHOD {
        GET, POST, PUT, DELETE
    };

    // Default choice of data format (especially, for outgoing data)
    public static enum DATA_FORMAT {
        String(String.class), Binary(byte[].class);

        private final Class dataFormat;

        DATA_FORMAT(Class dataFormat) {
            this.dataFormat = dataFormat;
        }

        public Class getDataFormat() {
            return this.dataFormat;
        }

        public String getMIMEType() {
            if (dataFormat.equals(String.class)) {
                return "text/plain";
            } else {
                return "application/octet-stream";
            }
        }
    };

    // These ports are default ones; additional ports will be dynamically
    // generated from the
    // URI signature used to configure the activity
    private static final String IN_BODY = "inputBody";
    private static final String OUT_RESPONSE_BODY = "responseBody";
    private static final String OUT_RESPONSE_HEADERS = "responseHeaders";
    private static final String OUT_STATUS = "status";
    private static final String OUT_REDIRECTION = "redirection";
    private static final String OUT_COMPLETE_URL = "actualURL";

    // Configuration bean for this activity - essentially defines a particular
    // instance
    // of the activity through the values of its parameters
    private RESTActivityConfigurationBean configBean;

    @Override
    public RESTActivityConfigurationBean getConfiguration() {
        return this.configBean;
    }

    @Override
    public void configure(RESTActivityConfigurationBean configBean)
            throws ActivityConfigurationException {
        // Check configBean is valid - mainly check the URI signature for being
        // well-formed and other details being present and valid;
        //
        // NB! The URI signature will still be valid if there are no
        // placeholders at all - in this
        // case for GET and DELETE methods no input ports will be generated and
        // a single input
        // port for input message body will be created for POST / PUT methods.
        if (!configBean.isValid()) {
            throw new ActivityConfigurationException(
                    "Bad data in the REST activity configuration bean - "
                    + "possible causes are: missing or ill-formed URI signature, missing or invalid MIME types for the "
                    + "specified HTTP headers ('Accept' | 'Content-Type'). This should not have happened, as validation "
                    + "on the UI had to be performed prior to accepting this configuration.");
        }

        // Store for getConfiguration()
        this.configBean = configBean;

        // (Re)create input/output ports depending on configuration
        configurePorts();
    }

    protected void configurePorts() {
        // In case we are being reconfigured - remove existing ports first to
        // avoid duplicates
        //TODO activity should listen to their beans for modifications.
        removeInputPorts();
        removeOutputPorts();

        // ---- CREATE INPUTS ----
        // all input ports are dynamic and depend on the configuration
        // of the particular instance of the REST activity
        // POST and PUT operations send data, so an input for the message body
        // is required
        if (hasMessageBodyInputPort()) {
            addInputPort(new SimpleInputPort(IN_BODY, configBean.getOutgoingDataFormat().getMIMEType(), 0, this));
        }

        // now process the URL signature - extract all placeholders and create
        // an input port for each
        Map<String, Class<?>> activityInputs = new HashMap<>();
        List<String> placeholders = URISignatureHandler
                .extractPlaceholders(configBean.getUrlSignature());
        String acceptsHeaderValue = configBean.getAcceptsHeaderValue();
        if (acceptsHeaderValue != null && !acceptsHeaderValue.isEmpty()) {
            try {
                List acceptsPlaceHolders = URISignatureHandler
                        .extractPlaceholders(acceptsHeaderValue);
                acceptsPlaceHolders.removeAll(placeholders);
                placeholders.addAll(acceptsPlaceHolders);
            } catch (URISignatureParsingException e) {
                LOGGER.log(Level.INFO, "Error in URL parsing", e);
            }
        }
        for (ArrayList<String> httpHeaderNameValuePair : configBean.getOtherHTTPHeaders()) {
            try {
                List headerPlaceHolders = URISignatureHandler
                        .extractPlaceholders(httpHeaderNameValuePair.get(1));
                headerPlaceHolders.removeAll(placeholders);
                placeholders.addAll(headerPlaceHolders);
            } catch (URISignatureParsingException e) {
                LOGGER.log(Level.INFO, "Error in URL parsing", e);
            }
        }
        for (String placeholder : placeholders) {
            // these inputs will have a dynamic name each;
            // the data type is string as they are the values to be
            // substituted into the URL signature at the execution time
            addInputPort(new SimpleInputPort(placeholder, "text/plain", 0, this));
            activityInputs.put(placeholder, String.class);
        }

        // all inputs have now been configured - store the resulting set-up in
        // the config bean;
        // this configuration will be reused during the execution of activity,
        // so that existing
        // set-up could simply be referred to, rather than "re-calculated"
        configBean.setActivityInputs(activityInputs);

        // ---- CREATE OUTPUTS ----
        // all outputs are of depth 0 - i.e. just a single value on each;
        // output ports for Response Body and Status are static - they don't
        // depend on the configuration of the activity;
        addOutputPort(new SimpleOutputPort(OUT_RESPONSE_BODY, "text/plain", 0, this));
        addOutputPort(new SimpleOutputPort(OUT_STATUS, "text/plain", 0, this));
        if (configBean.getShowActualUrlPort()) {
            addOutputPort(new SimpleOutputPort(OUT_COMPLETE_URL, "text/plain", 0, this));
        }
        if (configBean.getShowResponseHeadersPort()) {
            addOutputPort(new SimpleOutputPort(OUT_RESPONSE_HEADERS, "text/plain", 1, this));
        }

        // Redirection port may be hidden/shown
        if (configBean.getShowRedirectionOutputPort()) {
            addOutputPort(new SimpleOutputPort(OUT_REDIRECTION, "text/plain", 0, this));
        }
    }

    /**
     * Uses HTTP method value of the config bean of the current instance of
     * RESTActivity.
     *
     * @see RESTActivity#hasMessageBodyInputPort(HTTP_METHOD)
     */
    public boolean hasMessageBodyInputPort() {
        return (RESTActivity
                .hasMessageBodyInputPort(configBean.getHttpMethod()));
    }

    /**
     * Return value of this method has a number of implications - various input
     * ports and configuration options for this activity are applied based on
     * the selected HTTP method.
     *
     * @param httpMethod HTTP method to make the decision for.
     * @return True if this instance of the REST activity uses HTTP POST / PUT
     * methods; false otherwise.
     */
    public static boolean hasMessageBodyInputPort(HTTP_METHOD httpMethod) {
        return (httpMethod == HTTP_METHOD.POST || httpMethod == HTTP_METHOD.PUT);
    }

    /**
     * This generic activity can deal with any of the four HTTP methods and
     * executes pre-configured instance of REST activity. It resolves inputs of
     * the activity and registers its outputs; the real invocation of the HTTP
     * request is performed by
     * {@link HTTPRequestHandler#initiateHTTPRequest(String, RESTActivityConfigurationBean, String)}
     * .
     */
    @Override
    protected Map<String, DataContainer> execute(Map<String, DataContainer> inputMap) throws TaskExecutionException {

        DataContainerAdapter indata = new DataContainerAdapter(inputMap);

        // ---- RESOLVE INPUTS ----
        // RE-ASSEMBLE REQUEST URL FROM SIGNATURE AND PARAMETERS
        // (just use the configuration that was determined in
        // configurePorts() - all ports in this set are required)
        Map<String, String> urlParameters = new HashMap<>();
        for (String inputName : configBean.getActivityInputs().keySet()) {
            try {
                if (inputMap.containsKey(inputName)) {
                    urlParameters.put(inputName, indata.getString(inputName));
                }
            } catch (Exception e) {
                // problem occurred while resolving the inputs
                throw new TaskExecutionException("REST activity was unable to resolve all necessary inputs "
                        + "that contain values for populating the URI signature placeholders", e);
            }
        }
        String completeURL = URISignatureHandler.generateCompleteURI(
                configBean.getUrlSignature(), urlParameters, configBean
                .getEscapeParameters());

        // OBTAIN THE INPUT BODY IF NECESSARY
        // ("IN_BODY" is treated as *optional* for now)
        Object inputMessageBody = null;
        if (hasMessageBodyInputPort() && inputMap.containsKey(IN_BODY)) {
            if (configBean.getOutgoingDataFormat().getDataFormat().equals((String.class))) {
                inputMessageBody = indata.getString(IN_BODY);
            } else {
                throw new TaskExecutionException("TODO: add encoding for Byte[]");
            }
        }

        // ---- DO THE ACTUAL SERVICE INVOCATION ----
        HTTPRequestResponse requestResponse = HTTPRequestHandler
                .initiateHTTPRequest(completeURL, configBean,
                        inputMessageBody, urlParameters);

        // test if an internal failure has occurred
        if (requestResponse.hasException()) {
            throw new TaskExecutionException("Internal error has occurred while trying to execute the REST activity",
                    requestResponse.getException());
        }

        // ---- REGISTER OUTPUTS ----
        Map<String, DataContainer> outputs = new HashMap<>();

        DataContainer responseBodyRef = null;

        if (requestResponse.hasServerError()) {
            // test if a server error has occurred -- if so, return
            // output as an error document
            String errormes = "";
            if (requestResponse.getResponseBody() == null) {
                // No response body - register empty string
            } else if (requestResponse.getResponseBody() instanceof String) {
                errormes = (String) requestResponse.getResponseBody();
            } else if (requestResponse.getResponseBody() instanceof byte[]) {
                // Do the only thing we can - try to convert to
                // UTF-8 encoded string
                // and hope we'll get back something intelligible
                String str = null;
                try {
                    str = new String(((byte[]) requestResponse
                            .getResponseBody()), "UTF-8");
                } catch (UnsupportedEncodingException e) {
                    LOGGER.log(Level.INFO, "Failed to reconstruct the response body byte[] into string using UTF-8 encoding", e);
                    str = new String(((byte[]) requestResponse.getResponseBody())); // try with no encoding, probably will get garbage
                }
                errormes = str;
            } else {
                // Do what we can - call toString() method and hope
                // for the best
                errormes = requestResponse.getResponseBody().toString();
            }
            
            HTTPException ex = new HTTPException(errormes);
            ex.setStatusCode(requestResponse.getStatusCode());
            throw ex;
        } else if (requestResponse.getResponseBody() != null) {
            // some response data is available
            responseBodyRef = DataContainerFactory.create(requestResponse.getResponseBody());
        } else {
            // no data was received in response to the request - must
            // have been just a response header...
            responseBodyRef = DataContainerFactory.create("");
        }
        outputs.put(OUT_RESPONSE_BODY, responseBodyRef);
        outputs.put(OUT_STATUS, DataContainerFactory.create(requestResponse.getStatusCode()));

        if (configBean.getShowActualUrlPort()) {
            outputs.put(OUT_COMPLETE_URL, DataContainerFactory.create(completeURL));
        }
        if (configBean.getShowResponseHeadersPort()) {
            DataContainer headerdata = DataContainerFactory.create(requestResponse.getHeadersAsStrings());
            headerdata.setDepth(1);
            outputs.put(OUT_RESPONSE_HEADERS, headerdata);
        }

        if (configBean.getShowRedirectionOutputPort()) {
            outputs.put(OUT_REDIRECTION, DataContainerFactory.create(requestResponse.getRedirectionURL()));
        }

        // return map of output data, with empty index array as this is
        // the only and final result (this index parameter is used if
        // pipelining output)
        return outputs;
    }
}
