/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.data;

import java.beans.IntrospectionException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.modules.bistro.data.factory.DataContainerFactory;
import org.biobeans.modules.bistro.data.factory.DataContainerXMLFactory;
import org.biobeans.modules.bistro.data.iterator.BaclavaIterator;
import org.biobeans.modules.bistro.dataflow.SemanticMarkup;
import org.jdom.Element;

/**
 * A simple wrapper around an arbitrary Collection object which allows lookup
 * and storage of any metadata within the collection or its children. In
 * addition, there is an object of metadata concerning the DataContainer itself.
 * You obtain a DataContainer by invoking the create operation on the
 * DataContainerFactory in the factory subpackage, this is to allow the factory
 * to sensibly configure such things as types and underlying collections.
 *
 * @author Tom Oinn
 * @author EMBL-EBI
 */
public class DataContainer implements Cloneable, Serializable {

    private static final Logger LOGGER = Logger.getLogger(DataContainer.class.getName());

    protected Object theData;
    
    private int depth;

    protected HashMap<Object, WeakReference<SemanticMarkup>> metadataMap = new HashMap<>();

    protected SemanticMarkup myMarkup;

    protected HashMap<Object, String> lsid = new HashMap<>();

    public DataContainer(DataContainer other) {
        theData = other.theData;
        metadataMap.putAll(other.metadataMap);
        myMarkup = new SemanticMarkup(other.myMarkup);
        lsid.putAll(other.lsid);
    }

    /**
     * Construct a new DataContainer from the supplied XML Jdom Element.
     * Delegates to the DataContainerXMLFactory for almost all the real work
     * here.
     *
     * @param e
     */
    public DataContainer(Element e) {
        myMarkup = new SemanticMarkup(this);
        theData = DataContainerXMLFactory.configureDataContainer(e, this);
    }
    
    public void setDepth(int depth) {
        this.depth = depth;
    }
    
    public int getDepth() {
        return depth;
    }

    /**
     * Set the LSID of the named object to the specified value.
     */
    public void setLSID(Object target, String id) {
        if (id != null) {
            lsid.put(target, id);
        }
    }

    /**
     * Get the LSID of the named object, returns the empty string if there is no
     * such mapping
     *
     * @param target
     * @return
     */
    public String getLSID(Object target) {
        String lsidString = lsid.get(target);
        return (lsidString != null) ? lsidString : "";
    }

    /**
     * Get the object with the supplied LSID or return null if there isn't one
     *
     * @param LSID
     * @return
     */
    public Object getDataWithLSID(String LSID) {
        for (Object key : lsid.keySet()) {
            String value = lsid.get(key);
            // logger.debug("LSID value found : "+value);
            if (value.equals(LSID)) {
                return key;
            }
        }
        return null;
    }

    /**
     * Return an array of all the LSIDs that this DataContainer's LSID map
     * contains as values
     *
     * @return
     */
    public String[] getAllLSIDs() {
        return lsid.values().toArray(new String[0]);
    }

    /**
     * Create and bind a new SemanticMarkup object to the DataContainer itself,
     * it's not totally clear there's a need for this but it does no harm so why
     * not?
     *
     * @param o
     */
    public DataContainer(Object o) {
        if (o == null) {
            throw new RuntimeException(
                    "Attempt to create a null data object, definitely not allowed!");
        }
        if (o instanceof ArrayList == false) {
            theData = DataContainerFactory.convertObject(o);
        } else {
            theData = o;
        }
        myMarkup = new SemanticMarkup(this);
    }

//    /**
//     * Get a display icon for this DataContainer, currently this is based on the
//     * MIME type from the syntactic type string.
//     * @return 
//     */
//    public ImageIcon getIcon() {
//        String baseMIMEType = (getSyntacticType().split("'")[1].toLowerCase())
//                .split("/")[0];
//        return new ImageIcon(ClassLoader
//                .getSystemResource("org/embl/ebi/escience/baclava/icons/"
//                        + baseMIMEType + ".png"));
//    }
    /**
     * Return the SemanticMarkup object associated with the DataContainer itself
     *
     * @return
     */
    public SemanticMarkup getMetadata() {
        return myMarkup;
    }

    /**
     * Get the underlying data, this is the first level of the data document.
     *
     * @return
     */
    public Object getData() {
        return theData;
    }

    /**
     * Set the underlying data object, this is the first level of the data
     * document.
     */
    private void setData(Object data) {
        theData = data;
        setLSID(theData, "");
//        doInternalLSIDFill(theData, SYSTEM_DEFAULT_LSID_PROVIDER);
    }

    /**
     * Get the syntax type of this DataContainer. The type string is based
     * around application of the collection type constructors to a base MIME
     * type. For example, t(s('text/plain')) is a tree of sets of TEXT/PLAIN
     * items. The MIME type may be a comma separated list of types. Possible
     * type constructors are t(..) for trees, s(..) for sets, l(..) for lists
     * and p(..) for partial orders.
     * <p>
     * The type string "null" represents and empty DataContainer and is the
     * default value returned if the collection is empty.
     *
     * @return
     */
    public String getSyntacticType() {
        return getSyntacticTypeForObject(theData);
    }

    public String getMostInterestingMIMETypeForObject(Object o) {
        String typeString = getSyntacticTypeForObject(o);
        // logger.debug("Got types : "+typeString);
        String mimeTypes = typeString.split("'")[1].toLowerCase();
        for (String interestingType : DataType.BASICTYPES) {
            if (mimeTypes.matches(".*" + interestingType + ".*")) {
                return interestingType;
            }
        }
        try {
            return mimeTypes.split(",")[0];
        } catch (Exception ex) {
            return "null";
        }
    }

    public String getSyntacticTypeForObject(Object o) {
        if (o instanceof Collection) {
            if (((Collection) o).isEmpty()) {
                if (o instanceof Set) {
                    return "s('null')";
                }
                return "l('null')";
            }
            // Change to take the _last_ item of the List if it's a List rather
            // than using the iterator to take the first item
            Object innerObject;
            if (o instanceof List) {
                List innerList = (List) o;
                innerObject = innerList.get(innerList.size() - 1);
            } else {
                // Pull the first object out of the collection and recurse
                innerObject = ((Collection) o).iterator().next();
            }
            if (o instanceof Set) {
                return ("s(" + getSyntacticTypeForObject(innerObject) + ")");
            } else if (o instanceof List) {
                return ("l(" + getSyntacticTypeForObject(innerObject) + ")");
            }
            // No idea what the collection is, return the most general
            // type constructor for a partial order
            return ("p(" + getSyntacticTypeForObject(innerObject) + ")");
        }
        // Not a collection, first see if there is any metadata
        // associated with the object that we can use to determine
        // the mime types
        try {
            SemanticMarkup markup = getMetadataForObject(o, false);
            List mimeTypeList = markup.getMIMETypeList();
            if (mimeTypeList.isEmpty()) {
                // If there is no MIME information in the markup object
                // then we have to revert to guesswork.
                throw new NoMetadataFoundException();
            }
            // Return a comma seperated list of MIME types within
            // single quotes.
            StringBuilder sb = new StringBuilder();
            sb.append("'");
            for (Iterator i = mimeTypeList.iterator(); i.hasNext();) {
                String mimeType = (String) i.next();
                sb.append(mimeType);
                if (i.hasNext()) {
                    sb.append(",");
                }
            }
            sb.append("'");
            return sb.toString();
        } catch (NoMetadataFoundException nmfe) {
            StringBuilder sb = new StringBuilder();
            // Try to annotate with mime types based on data object type
            if (o instanceof String) {
                getMetadata().addMIMEType("text/plain");
            } else if (o instanceof byte[]) {
                getMetadata().addMIMEType("application/octet-stream");
            } else {
                // Special magic MIME type for Java class
                // Destroy all existing mime types first if this is the
                // case!
                getMetadata().clearMIMETypes();
                getMetadata().addMIMEType("java/" + o.getClass().getName());
            }
            for (Iterator i = getMetadata().getMIMETypeList().iterator(); i
                    .hasNext();) {
                sb.append((String) i.next());
                if (i.hasNext()) {
                    sb.append(",");
                }
            }

            // Try to annotate with mime types based on data object type
            String specifiedMIMETypes = sb.toString();
            return ("'" + specifiedMIMETypes + "'");
        }
    }

    /**
     * Copy the markup from the supplied DataContainer object to this one. This
     * is mainly used when the data has been repackaged, to preserve eg. LSID
     * values.
     *
     * @param source
     */
    public void copyMetadataFrom(DataContainer source) {
        lsid.putAll(source.lsid);
        metadataMap.putAll(source.metadataMap);
    }

    /**
     * Link the metadata of this object to the specified DataContainer, similar
     * in effect to the copy operation above in most cases but doesn't deep copy
     * the metadata.
     *
     * @param source
     */
    public void linkMetadataFrom(DataContainer source) {
        lsid = source.lsid;
        metadataMap = source.metadataMap;
    }

    /**
     * Get the SemanticMarkup associated with an object in this DataContainer.
     * If there is no such metadata available the behavious depends upon the
     * value of the supplyDefaults parameter. If false, then a
     * NoMetadataFoundException is thrown, if true a new SemanticMarkup object
     * is created, stored in the dictionary and returned to the caller.
     *
     * @param theObject
     * @param supplyDefault
     * @return
     */
    public SemanticMarkup getMetadataForObject(Object theObject,
            boolean supplyDefault) throws NoMetadataFoundException {
        WeakReference<SemanticMarkup> ref = metadataMap.get(theObject);
        if (ref != null && ref.get() != null) {
            return ref.get();
        }
        if (supplyDefault == false) {
            throw new NoMetadataFoundException("No metadata available");
        }
        // Create a new markup object and store
        // it bound to the object specified
        SemanticMarkup theMarkup = new SemanticMarkup(theObject);
        metadataMap.put(theObject, new WeakReference<>(theMarkup));
        return theMarkup;
    }

    /**
     * Return the JDom Element corresponding to this DataContainer represented
     * as XML
     */
    public Element getElement() {
        return DataContainerXMLFactory.getElement(this);
    }

    /**
     * Extract a child object as a DataContainer. This assumes that the object
     * referenced is actually within this DataContainer object, behaviour is
     * undefined otherwise. Although we could check this and impose it as a
     * constraint the performance hit of traversing the entire original
     * DataContainer collection structure and doing potentially expensive
     * equivalence computations is probably not worth it. Use with care.
     *
     * @param child
     * @return a view on an object contained within this DataContainer as a new
     * DataContainer
     */
    public DataContainer extractChild(Object child) {
        DataContainer result = new DataContainer(this);
        result.theData = child;
        return result;
    }

    /**
     * Iterate over all immediate children. If there are no children, return an
     * iterator over nothing. All children will be viewed as DataContainer
     * instances.
     *
     * @return an Iterator over all children
     */
    @SuppressWarnings("unchecked")
    public Iterator<DataContainer> childIterator() {
        if (theData instanceof Collection) {
            List<DataContainer> dataContainerList = new ArrayList<>();
            for (Iterator i = ((Collection) theData).iterator(); i
                    .hasNext();) {
                DataContainer newContainer = new DataContainer(i.next());
                newContainer.metadataMap = metadataMap;
                newContainer.myMarkup = new SemanticMarkup(myMarkup);
                newContainer.lsid = lsid;
                dataContainerList.add(newContainer);
                // dataContainerList.add(extractChild(i.next()));
            }
            return dataContainerList.iterator();
        }
        return Collections.EMPTY_LIST.iterator();
    }

    /**
     * Given a desired type, return the BaclavaIterator that provides
     * DataContainer objects of this type. If the desired collection structure
     * is not contained by this DataContainer then an exception is thrown.
     *
     * @param desiredType
     * @return
     * @exception IntrospectionException thrown if the supplied type is not
     * contained within the current DataContainer type.
     */
    public BaclavaIterator iterator(String desiredType)
            throws IntrospectionException {
        String type = null;
        String currentType = null;
        try {
            type = desiredType.split("\\'")[0];
        } catch (ArrayIndexOutOfBoundsException aioobe) {
            type = "";
        }
        try {
            currentType = getSyntacticType().split("\\'")[0];
        } catch (ArrayIndexOutOfBoundsException aioobe) {
            currentType = "";
        }
        // At this point, we should have split the mime types away from the
        // collection types, so the current type looks like, for example, l() or
        // s(l())

        // If the strings are the same then we return an iterator with a single
        // item
        // in it, namely the current DataContainer; this is needed where the enactor
        // has
        // detected that iteration is required somewhere else using the join
        // iterator
        if (type.equals(currentType)) {
            List<DataContainer> dataContainerList = new ArrayList<>();
            dataContainerList.add(this);
            return new BaclavaIterator(dataContainerList);
        }

        // Now need to check that the conversion is valid, so either the
        // input type is the empty string (iterate over everything to produce
        // leaf nodes)
        // or it is a substring of the collection type
        if (type.equals("") || currentType.endsWith(type)) {
            // See how deep the iterator needs to go.
            int iterationDepth = (currentType.length() - type.length()) / 2;
            // Now drill down into the data structure that number of levels,
            // build a list of all the items into a new collection, iterate over
            // this list building the DataContainer objects and return the iterator
            // over that list (and breathe...)
            List<Object> targetList = new ArrayList<>();
            List<int[]> indexList = new ArrayList<>();
            drill(iterationDepth, targetList, indexList, new int[0],
                    (Collection) theData);
            // Now iterate over the target list creating new DataContainer objects
            // from it
            /**
             * for (Iterator i = targetList.iterator(); i.hasNext(); ) {
             * DataContainer newData = new DataContainer(i.next()); // Copy any
             * metadata into the new DataContainer
             * newData.metadataMap.putAll(metadataMap);
             * newData.lsid.putAll(lsid); dataContainerList.add(newData); }
             */
            // return new BaclavaIterator(dataContainerList, indexList);
            return new BaclavaIterator(this, targetList, indexList);
        }
        throw new IntrospectionException(
                "Incompatible types for iterator, cannot extract " + type
                + " from " + getSyntacticType());

    }

//    /**
//     * Drill into a collection searching for a data-object that has the same
//     * data with oldDT and replace its data with newData.
//     *
//     * @return true if a dataContainer's data has been replaced with the
//     * newData.
//     */
//    @SuppressWarnings("unchecked")
//    public DataContainer drillAndSet(DataContainer oldDT, String newData) {
//        if (theData instanceof Collection) {
//            ArrayList<Object> dtList = new ArrayList<>((Collection) theData);
//            for (int i = 0; i < dtList.size(); i++) {
//                DataContainer tmp = new DataContainer(dtList.get(i));
//                tmp.metadataMap = metadataMap;
//                tmp.myMarkup = new SemanticMarkup(myMarkup);
//                tmp.lsid = lsid;
//                DataContainer newDT = tmp.drillAndSet(oldDT, newData);
//                if (newDT != null) {
//                    dtList.set(i, newDT.getData());
//                    setData(dtList);
//                    return this;
//                }
//            }
//            return null;
//        } else if (getData().equals(oldDT.getData())) {
//            setData(newData);
//            return this;
//        } else {
//            return null;
//        }
//    }

    /*
	 * public void drillAndPrint(){ System.out.print("-");
	 * System.out.println(toString()); for (Iterator
	 * iDT=childIterator();iDT.hasNext();){ System.out.print("-");
	 * ((DataContainer)iDT.next()).drillAndPrint(); } }
     */
    /**
     * Drill into a collection, adding items to the list if we're at the desired
     * depth, this makes the underlying assumption that the collection contains
     * either collections or objects, but never a mix of both.
     */
    private void drill(int iterationDepth, List<Object> targetList, List<int[]> indexList,
            int[] currentIndex, Collection theData) {
        if (iterationDepth == 1) {
            // Collecting items
            int localIndex = 0;
            for (Iterator i = theData.iterator(); i.hasNext();) {
                targetList.add(i.next());
                indexList.add(append(currentIndex, localIndex++));
            }
        } else {
            // Iterating further down
            int localIndex = 0;
            for (Iterator i = theData.iterator(); i.hasNext();) {
                Collection theCollection = (Collection) i.next();
                drill(iterationDepth - 1, targetList, indexList, append(
                        currentIndex, localIndex++), theCollection);
            }
        }
    }

    /**
     * Utility class, append an int onto an int array and return the new array
     */
    private int[] append(int[] list, int head) {
        int[] newlist = new int[list.length + 1];
        System.arraycopy(list, 0, newlist, 0, list.length);
        newlist[list.length] = head;
        return newlist;
    }

    @Override
    public String toString() {
        String datStr = theData.toString();
        boolean trimmed = false;
        int nl = datStr.indexOf('\n');
        if (nl > -1) {
            datStr = datStr.substring(0, nl);
            trimmed = true;
        }
        if (datStr.length() > 30) {
            datStr = datStr.substring(0, 27);
            trimmed = true;
        }
        if (trimmed) {
            datStr += "...";
        }
        return super.toString() + "\n\tValue=" + datStr + "\n\tLSID="
                + getLSID(theData) + "\n";
    }

    /**
     * Writes the contents of the DataContainer into the specified directory
     * using the given name. If there is only one item a single file is created
     * otherwise a directory structure mirroring the collection structure is
     * built. A File object representing the file or directory that has been
     * written is returned.
     *
     * @param destination
     * @param name
     * @return
     * @throws java.io.IOException
     */
    public File writeToFileSystem(File destination, String name)
            throws IOException {
        // Check for the most interesting type, if defined
        String interestingType = getMostInterestingMIMETypeForObject(theData);
        String fileExtension = ".text";
        if (interestingType != null
                && interestingType.equals("text/plain") == false) {
            // MIME types look like 'foo/bar'
            String lastPart = interestingType.split("/")[1];
            if (lastPart.startsWith("x-") == false) {
                fileExtension = "." + lastPart;
            }
        }
        File writtenFile = DataContainer.writeObjectToFileSystem(destination, name,
                theData, fileExtension);
        return writtenFile;
    }

    /**
     * Write a specific object to the filesystem this has no access to metadata
     * about the object and so is not particularly clever. A File object
     * representing the file or directory that has been written is returned.
     *
     * @param destination
     * @param name
     * @param o
     * @param defaultExtension
     * @return
     * @throws java.io.IOException
     */
    public static File writeObjectToFileSystem(File destination, String name,
            Object o, String defaultExtension) throws IOException {
        // If the destination is not a directory then set the destination
        // directory to the parent and the name to the filename
        // i.e. if the destination is /tmp/foo.text and this exists
        // then set destination to /tmp/ and name to 'foo.text'
        if (destination.exists() && destination.isFile()) {
            name = destination.getName();
            destination = destination.getParentFile();
        }
        if (destination.exists() == false) {
            // Create the directory structure if not already present
            destination.mkdirs();
        }
        File writtenFile = writeData(destination, name, o,
                defaultExtension);
        return writtenFile;
    }

    static char sep = File.separatorChar;

    private static File writeData(File destination, String name,
            Object o, String defaultExtension) throws IOException {
        if (o instanceof Collection) {
            // Create a new directory, iterate over the collection recursively
            // calling this method
            File targetDir = new File(destination.toString() + sep + name);
            targetDir.mkdir();
            int count = 0;
            Collection c = (Collection) o;
            for (Iterator i = c.iterator(); i.hasNext();) {
                writeData(targetDir, "" + count++, i.next(),
                        defaultExtension);
            }
            return targetDir;
        } else {
            // Write a single item
            if (o instanceof String && defaultExtension == null) {
                name = name + ".text";
            } else if (defaultExtension != null) {
                name = name + defaultExtension;
            }
            File targetFile = new File(destination.toString() + sep + name);
            FileOutputStream fos = new FileOutputStream(targetFile);
            if (o instanceof byte[]) {
                fos.write((byte[]) o);
                fos.flush();
                fos.close();
            } else if (o instanceof String) {
                PrintWriter out = new PrintWriter(new OutputStreamWriter(fos));
                out.print((String) o);
                out.flush();
                out.close();
            }
            return targetFile;
        }
    }

    @Override
    public Object clone() {
        ByteArrayOutputStream bin = new ByteArrayOutputStream();
        try {
            ObjectOutputStream out = new ObjectOutputStream(bin);
            out.writeObject(this);
            byte[] buf = bin.toByteArray();
            ObjectInputStream in = new ObjectInputStream(
                    new ByteArrayInputStream(buf));
            return in.readObject();
        } catch (IOException | ClassNotFoundException e) {
            LOGGER.log(Level.SEVERE, "Could not clone", e);
            return null;
        }
    }
}
