/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.data;

import java.awt.Image;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Map;
import org.biobeans.modules.bistro.data.factory.DataContainerFactory;

/**
 * This class is designed as a wrapper for input maps and output maps. Values
 * that are typically stored in these maps are usually stored as DataContainer. In
 * order to make the code cleaner, this wrapper insures that values inserted
 * into these maps are inserted as DataContainer.
 *
 * @author Mark
 */
public class DataContainerAdapter {

    private final Map<String, DataContainer> map;

    public DataContainerAdapter(Map<String, DataContainer> map) {
        this.map = map;
    }

    /**
     * This method gets a value out of the map and returns it as a string. It
     * performs the DataContainer conversion for you.
     *
     * @param key
     * @return
     */
    public String getString(String key) {
        DataContainer val = (DataContainer) map.get(key);
        if (val == null) {
            return null;
        }
        String strVal = (String) val.getData();

        return strVal;
    }

    /**
     * This method gets an array of objects as a string array.
     *
     * @param key
     * @return
     */
    public String[] getStringArray(String key) {
        DataContainer val = (DataContainer) map.get(key);
        if (val == null) {
            return null;
        }
        ArrayList list = (ArrayList) val.getData();
        String[] vals = new String[list.size()];
        return (String[]) list.toArray(vals);

        // return (String[])val.getData();
    }

    /**
     * This method wraps the string value in a DataContainer before storing it in
     * the hashmap.
     *
     * @param key The key for the value.
     * @param value The value to be stored.
     */
    public void putString(String key, String value) {
        DataContainer val = new DataContainer(value);
        this.map.put(key, val);
    }

    /**
     * This method puts a boolean value into the map.
     *
     * @param key
     * @param value
     */
    public void putBoolean(String key, boolean value) {
        DataContainer val = new DataContainer(String.valueOf(value));
        this.map.put(key, val);
    }

    /**
     * This method gets a boolean value from the hashmap
     *
     * @param key
     * @return False if null, otherwise returns the value indicated by the key.
     */
    public boolean getBoolean(String key) {
        DataContainer val = (DataContainer) map.get(key);
        if (val == null) {
            return false;
        }
        String strVal = (String) val.getData();

        return Boolean.parseBoolean(strVal);
    }

    /**
     * This method puts a string array into the DataContainer map.
     *
     * @param key The key for the value.
     * @param values The values to be stored.
     */
    public void putStringArray(String key, String[] values) {
        // ArrayList valueArray = new ArrayList();
        // valueArray = (ArrayList)Arrays.asList(values);
        DataContainer val = new DataContainer(values);
        this.map.put(key, val);
    }

    /**
     * This method gets an integer value from the DataContainer map
     *
     * @param key The key for the value.
     * @return
     */
    public int getInt(String key) {
        String strVal = getString(key);
        return Integer.parseInt(strVal);
    }

    /**
     * This method puts an integer value into the DataContainer map
     *
     * @param key The key for the value.
     * @param value The values to be stored.
     */
    public void putInt(String key, int value) {
        putString(key, String.valueOf(value));
    }

    /**
     * This method gets a two-dimensional array out of the DataContainer map.
     *
     * @param key
     * @return
     */
    public String[][] getArrayArray(String key) {
        String[][] values = null;
        DataContainer val = (DataContainer) map.get(key);
        return (String[][]) val.getData();
    }

    /**
     * This method puts a two-dimensional array into the DataContainer map.
     *
     * @param key
     * @param value
     */
    public void putArrayArray(String key, String[][] value) {
        map.put(key, DataContainerFactory.create(value));
    }

    /**
     * This method gets an ArrayList from the DataContainer map.
     *
     * @param key
     * @return
     */
    public ArrayList getArrayList(String key) {
        DataContainer val = (DataContainer) map.get(key);
        return (ArrayList) val.getData();
    }

    /**
     * This method puts an ArrayList into the DataContainer map
     *
     * @param key
     * @param list
     */
    public void putArrayList(String key, ArrayList list) {
        DataContainer val = new DataContainer(list);
        map.put(key, val);
    }

    /**
     * This method puts a Serializable object into the map.
     *
     * @param key
     * @param obj
     */
    public void putSerializable(String key, Serializable obj) {
        DataContainer val = new DataContainer(obj);
        map.put(key, val);
    }

    /**
     * This method gets a Serializable object from the map.
     *
     * @param key
     * @return
     */
    public Serializable getSerializable(String key) {
        DataContainer val = (DataContainer) map.get(key);
        return (Serializable) val.getData();
    }

    public Image getImage(String key) {
        DataContainer val = (DataContainer) map.get(key);
        return (Image) val.getData();
    }

    public void putImage(String key, Image img) {
        DataContainer val = new DataContainer(img);
        map.put(key, val);
    }

    public void putDouble(String key, Double db) {
        DataContainer val = new DataContainer(db);
        map.put(key, val);
    }

    public Double getDouble(String key) {
        DataContainer val = (DataContainer) map.get(key);
        return (Double) val.getData();
    }

    public double getDoubleVal(String key) {
        return getDouble(key);
    }

}
