/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.data.factory;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.modules.bistro.data.Base64;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.data.NoMetadataFoundException;
import org.biobeans.modules.bistro.dataflow.SemanticMarkup;
import org.biobeans.modules.bistro.dataflow.DataFlow;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;

/**
 * Performs the creation of XML elements from DataContainer objects and the
 * configuration of existing DataContainer objects from an XML representation.
 *
 * @author Tom Oinn
 * @author EMBL-EBI
 */
public class DataContainerXMLFactory {

    private static final Logger LOGGER = Logger.getLogger(DataContainerXMLFactory.class.getName());

    /**
     * The namespace used in the XML representation of a DataContainer object
     */
    public static Namespace namespace = Namespace.getNamespace("b",
            "http://org.biobeans.bistro/data/1.0");

    /**
     * Return a Document from a Map of String to DataContainer, this is the input
     * and output document format for the workflow enactment engine.
     * @param datas
     * @return 
     */
    public static Document getDataDocument(Map<String,DataContainer> datas) {
        Element rootElement = new Element("dataMap", namespace);
        Document theDocument = new Document(rootElement);
        for (String key : datas.keySet()) {
            DataContainer value = (DataContainer) datas.get(key);
            Element dataElement = new Element("data", namespace);
            dataElement.setAttribute("key", key);
            dataElement.addContent(value.getElement());
            rootElement.addContent(dataElement);
        }
        return theDocument;
    }

    /**
     * Parse a data document and return a Map of DataContainer objects, the keys in
     * the map being the string key attributes from the data document.
     * 
     * @param dataDocument
     * @return 
     */
    public static Map<String, DataContainer> parseDataDocument(Document dataDocument) {
        Map<String, DataContainer> result = new HashMap<>();
        Element rootElement = dataDocument.getRootElement();
        for (Iterator i = rootElement.getChildren("dataContainer", namespace)
                .iterator(); i.hasNext();) {
            Element e = (Element) i.next();
            String key = e.getAttributeValue("key");
            DataContainer d = new DataContainer(e.getChild("gridDataDocument",
                    namespace));
            result.put(key, d);
        }
        return result;
    }

    /**
     * Build a DataContainer from the supplied Element object, the Element in this
     * case is the 'gridDataDocument' element.
     * @param rootElement
     * @param theDataContainer
     * @return 
     */
    public static Object configureDataContainer(Element rootElement,
            DataContainer theDataContainer) {
        // Configure semantic markup object for the DataContainer
        Element semanticMarkupElement = rootElement.getChild("metadata",
                DataFlow.NAMESPACE);
        // Get the LSID for the entire DataContainer
        String lsid = rootElement.getAttributeValue("lsid");
        if (lsid != null) {
            theDataContainer.setLSID(theDataContainer, lsid);
        }
        if (semanticMarkupElement != null) {
            theDataContainer.getMetadata().configureFromElement(
                    semanticMarkupElement);
        }
        return objectForElement(rootElement, theDataContainer, rootElement);
    }

    /**
     * Handle either a leaf node - dataElement - or a partial order -
     * partialOrder, and delegate to the appropriate walker method. This
     * searches for the appropriate child of the supplied element and calls the
     * method on that.
     */
    private static Object objectForElement(Element e, DataContainer theDataContainer,
            Element rootElement) {
        Element dataElement = e.getChild("dataElement", namespace);
        if (dataElement != null) {
            return objectForDataElement(dataElement, theDataContainer, rootElement);
        }
        Element posetElement = e.getChild("partialOrder", namespace);
        if (posetElement != null) {
            return objectForCollectionElement(posetElement, theDataContainer,
                    rootElement);
        }
        // Fallback case, should never hit this.
        return null;
    }

    /**
     * The recursive tree walker to build objects from XML, this call handles
     * leaf nodes.
     */
    private static Object objectForDataElement(Element e,
            DataContainer theDataContainer, Element rootElement) {
        // Is there a hint in a metadata block?
        Element metadataElement = e.getChild("metadata", DataFlow.NAMESPACE);
        String mimeHint = "";
        if (metadataElement != null) {
            SemanticMarkup m = new SemanticMarkup(new Object());
            m.configureFromElement(metadataElement);
            mimeHint = m.getFirstMIMEType().toLowerCase();
        }
        if (mimeHint.equals("")) {
            String documentSyntacticType = rootElement
                    .getAttributeValue("syntactictype");
            if (documentSyntacticType.equals("") == false) {
                mimeHint = (documentSyntacticType.split("'"))[1].toLowerCase();
            }
        }
        // Default to assuming string data, which is probably wrong but what
        // else can we do here?
        if (mimeHint.equals("")) {
            mimeHint = "text/plain";
        }
        // If the major part is anything other than text then we have
        // binary data so it seems reasonable to ask for a byte array
        // back from the convertor.
        String mimeMajorType = mimeHint.split("/")[0];
        String encodedData = e.getChild("dataElementData", namespace)
                .getTextTrim();
        byte[] decodedData = Base64.decode(encodedData);
        Object result;
        if (mimeMajorType.equals("text")) {
            result = new String(decodedData);
        } // Handle java objects
        else if (mimeMajorType.equals("java")) {
            try {
                ObjectInputStream in = new ObjectInputStream(
                        new ByteArrayInputStream(decodedData));
                result = in.readObject();
                in.close();
            } catch (IOException | ClassNotFoundException ex) {
                LOGGER.log(Level.SEVERE, "Failed to deserialize java object type", ex);
                result = "Unable to decode " + mimeHint;
            }
        } else {
            result = decodedData;
        }
        if (metadataElement != null) {
            SemanticMarkup m = theDataContainer.getMetadataForObject(result, true);
            m.configureFromElement(metadataElement);
        }
        String lsid = e.getAttributeValue("lsid");
        if (lsid != null) {
            theDataContainer.setLSID(result, lsid);
        }
        return result;
    }

    /**
     * The recursive tree walker to build objects from XML, this call handles
     * branch nodes i.e. collections
     */
    private static Object objectForCollectionElement(Element e,
            DataContainer theDataContainer, Element rootElement) {
        // For now as we're only looking at sets and lists we can
        // actually ignore the ordering section. For tree and partial
        // order support we're going to need to consider it, but for
        // now we're good.
        String collectionType = e.getAttributeValue("type");
        Element itemListElement = e.getChild("itemList", namespace);
        Collection<Object> result = null;
        if (collectionType.equals("list")) {
            result = new ArrayList<>();
        } else {
            result = new HashSet<>();
        }

        for (Iterator i = itemListElement.getChildren().iterator(); i.hasNext();) {
            Element childElement = (Element) i.next();
            if (childElement.getNamespace().equals(namespace)) {
                if (childElement.getName().equals("dataElement")) {
                    result.add(objectForDataElement(childElement, theDataContainer,
                            rootElement));
                } else if (childElement.getName().equals("partialOrder")) {
                    result.add(objectForCollectionElement(childElement,
                            theDataContainer, rootElement));
                }
            }
        }
        String lsid = e.getAttributeValue("lsid");
        if (lsid != null) {
            theDataContainer.setLSID(result, lsid);
        }
        return result;
    }

    /**
     * Return an XML Jdom Element object for the DataContainer supplied as the
     * argument
     */
    public static Element getElement(DataContainer theDataContainer) {
        Element rootElement = new Element("gridDataDocument", namespace);
        rootElement.setAttribute("lsid", theDataContainer.getLSID(theDataContainer));
        rootElement.setAttribute("syntactictype", theDataContainer
                .getSyntacticType());
        rootElement.addContent(theDataContainer.getMetadata()
                .getConfigurationElement());
        rootElement.addContent(elementForObject(theDataContainer.getData(),
                theDataContainer));
        return rootElement;
    }

    /**
     * The recursive method to convert an arbitrary object into XML
     */
    private static Element elementForObject(Object o, DataContainer theDataContainer) {
        // Handle collections
        if (o instanceof Collection) {
            Element poElement = new Element("partialOrder", namespace);
            poElement.setAttribute("lsid", theDataContainer.getLSID(o));
            Element relationListElement = new Element("relationList", namespace);
            Element listElement = new Element("itemList", namespace);
            int currentIndex = 0;
            poElement.addContent(relationListElement);
            poElement.addContent(listElement);
            if (o instanceof List) {
                // Handle lists
                poElement.setAttribute("type", "list");
                // Iterate over the list, creating the new items
                if (!((List) o).isEmpty()) {
                    for (Iterator i = ((List) o).iterator(); i.hasNext();) {
                        Element listItemElement = elementForObject(i.next(),
                                theDataContainer);
                        listElement.addContent(listItemElement);
                        listItemElement
                                .setAttribute("index", "" + currentIndex);
                        // If the index is non zero create a new relation
                        if (currentIndex > 0) {
                            Element relationElement = new Element("relation",
                                    namespace);
                            relationElement.setAttribute("parent", ""
                                    + (currentIndex - 1));
                            relationElement.setAttribute("child", ""
                                    + currentIndex);
                            relationListElement.addContent(relationElement);
                        }
                        currentIndex++;
                    }
                }
            } else if (o instanceof Set) {
                // Handle sets
                poElement.setAttribute("type", "set");
                // Iterate over the set, creating new items
            }
            return poElement;
        }
        // Handle data elements
        Element dataElement = new Element("dataElement", namespace);
        dataElement.setAttribute("lsid", theDataContainer.getLSID(o));
        try {
            SemanticMarkup sm = theDataContainer.getMetadataForObject(o, false);
            dataElement.addContent(sm.getConfigurationElement());
        } catch (NoMetadataFoundException nmfe) {
            // No metadata available for this object
        }
        // Add the data itself
        Element realDataElement = new Element("dataElementData", namespace);
        byte[] theBytes = new byte[0];
        if (o instanceof String) {
            theBytes = ((String) o).getBytes();
        } else if (o instanceof byte[]) {
            theBytes = (byte[]) o;
        } // Handle java object
        else {
            try {
                ByteArrayOutputStream bos = new ByteArrayOutputStream();
                ObjectOutputStream oos = new ObjectOutputStream(bos);
                oos.writeObject(o);
                theBytes = bos.toByteArray();
            } catch (Exception ex) {
                LOGGER.log(Level.SEVERE, "Error serializing {0}", o.getClass().getName());
            }
        }
        // Convert to base64
        LOGGER.log(Level.FINE, "Size of binary data is {0} bytes.", theBytes.length);
        realDataElement.setText(Base64.encodeBytes(theBytes));
        dataElement.addContent(realDataElement);
        return dataElement;
    }

}
