/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.data.iterator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import org.biobeans.modules.bistro.data.DataContainer;

/**
 * This provides an Iterator interface with a single additional method to allow
 * reset of the iterator to its starting state. This is required to produce
 * orthogonal joins between iterators used in the implicit iteration mechanism
 * within the enactor.
 *
 * @author Tom Oinn
 */
public class BaclavaIterator implements ResumableIterator {

    private Collection underlyingCollection = null;

    private List indexList = new ArrayList();

    private Iterator indexIterator = indexList.iterator();

    private Iterator internalIterator = null;

    private int[] currentLocation = null;

    private DataContainer parent = null;

    public BaclavaIterator(Collection c) {
        this.underlyingCollection = c;
        this.internalIterator = c.iterator();
    }

    public BaclavaIterator(Collection c, List indexList) {
        this(c);
        this.indexList = indexList;
        this.indexIterator = indexList.iterator();
    }

    /**
     * Construct an iterator from a list of non data objects, a reference
     * to the parent DataContainer which contains them and an explicit index list.
     * This avoids the overhead of creating all the DataContainer objects at
     * iterator construction time, something that was causing some serious
     * performance issues, mostly because of the metadata copy operations
     */
    public BaclavaIterator(DataContainer parent, Collection c, List indexList) {
        this(c, indexList);
        this.parent = parent;
    }

    @Override
    public int[] getCurrentLocation() {
        if (currentLocation == null || indexList.isEmpty()) {
            // No location information available
            return new int[0];
        } else {
            return currentLocation;
        }
    }

    @Override
    public synchronized boolean hasNext() {
        if (underlyingCollection.isEmpty()) {
            return false;
        }
        return this.internalIterator.hasNext();
    }

    @Override
    public synchronized Object next() throws NoSuchElementException {
        if (indexIterator.hasNext()) {
            // Increment the current index
            currentLocation = (int[]) indexIterator.next();
        }
        if (parent == null) {
            return this.internalIterator.next();
        } else {
            // Construct a new DataContainer on the fly and return it
            DataContainer newData = new DataContainer(this.internalIterator.next());
            newData.linkMetadataFrom(parent);
            return newData;
        }
    }

    @Override
    public void remove() throws UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "Remove operation not allowed in a BaclavaIterator.");
    }

    @Override
    public synchronized void reset() {
        this.internalIterator = underlyingCollection.iterator();
        this.indexIterator = indexList.iterator();
    }

    @Override
    public int size() {
        return this.underlyingCollection.size();
    }

}
