/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.services;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.biobeans.modules.bistro.activity.ConfigurationBean;
import org.biobeans.modules.bistro.activity.SimpleConfigurationBean;
import org.openide.nodes.BeanNode;
import static org.openide.nodes.BeanNode.computeProperties;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.util.Lookup;
import org.openide.util.Utilities;

/**
 *
 */
public abstract class AbstractBioService implements BioService {

    private final List<ChangeListener> listeners
        = Collections.synchronizedList(new ArrayList<ChangeListener>());

    private String description = "";

    private String name = "";

    /** set of properties to use */
    private Sheet sheet;
    
    private List<ConfigurationBean> beanList;
    
    public AbstractBioService() {
        BioServiceRegistry registry = Lookup.getDefault().lookup(BioServiceRegistry.class);
        registry.register(this);
    }
    
    @Override
    public String getIconBase() {
        return "org/biobeans/modules/bistro/resources/bioservice.png";
    }
 
    public void setName(String n) {
        if (n == null) {
            throw new IllegalArgumentException("Name was null");
        }
        name = n;
    }

    @Override
    public String getName() {
        return name;
    }

    public void setDescription(String d) {
        description = d;
    }

    @Override
    public String getDescription() {
        if(description==null) {
            return getName();
        }
        return description;
    }

    @Override
    public void addChangeListener(ChangeListener listener) {
        listeners.add(listener);
    }

    @Override
    public void removeChangeListener(ChangeListener listener) {
        listeners.remove(listener);
    }

    protected void notifyChange() {
        for (ChangeListener listener : listeners) {
            listener.stateChanged(new ChangeEvent(this));
        }
    }
    
    
    /*========================================================================*/
    
    protected void addConfigurationBean(ConfigurationBean bean) {
        if(beanList==null) {
            beanList = new ArrayList<>();
        }
        beanList.add(bean);
        Class clazz = bean.getClass();
        while (!Modifier.isPublic(clazz.getModifiers()) && !hasExplicitBeanInfo(clazz)) {
            clazz = clazz.getSuperclass();

            if (clazz == null) {
                clazz = Object.class; // in case it was an interface
            }
        }
        try {
            BeanInfo beanInfo = Utilities.getBeanInfo(clazz);
            addProperties(bean, beanInfo);
        } catch (IntrospectionException ex) {
            //TODO
        }
    }
    
    @Override
    public Node.PropertySet[] getPropertySets() {
        Sheet s = getSheet();
        return s.toArray();
    }
    
    protected Sheet createSheet() {
        return new Sheet();
    }
    
    protected final synchronized Sheet getSheet() {
        if (sheet != null) {
            return sheet;
        }
        sheet = createSheet();
        return sheet;
    }
    
    
     /** Checks whether there is an explicit bean info for given class.
    * @param clazz the class to test
    * @return true if explicit bean info exists
    */
    private boolean hasExplicitBeanInfo(Class<?> clazz) {
        String className = clazz.getName();
        int indx = className.lastIndexOf('.');
        className = className.substring(indx + 1);

        String[] paths = Introspector.getBeanInfoSearchPath();
        for (int i = 0; i < paths.length; i++) {
            String s = paths[i] + '.' + className + "BeanInfo"; // NOI18N

            try {
                Class.forName(s);
                return true;
            } catch (ClassNotFoundException ex) {
            }
        }

        return false;
    }
    
    protected void addProperties(ConfigurationBean bean, BeanInfo info) {
        BeanNode.Descriptor d = computeProperties(bean, info);

        Sheet sets = getSheet();
        Sheet.Set pset = Sheet.createPropertiesSet();
        pset.setName(bean.getName());
        pset.setDisplayName(bean.getName() + " - " + Sheet.PROPERTIES);
        pset.put(d.property);
        sets.put(pset);

        if (d.expert.length != 0) {
            Sheet.Set eset = Sheet.createExpertSet();
            eset.setName(bean.getName() + " - " + Sheet.EXPERT);
            eset.setDisplayName(bean.getName() + " - " + Sheet.EXPERT);
            eset.put(d.expert);
            sets.put(eset);
        }
    }
    

}
