/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
*/
package org.biobeans.modules.bistro.task;

import java.util.EventListener;
import java.util.HashSet;
import java.util.Iterator;
import org.openide.util.Cancellable;

/**
 *
 * @author George Waldon
 */
public class ActivityTask implements Cancellable {
    
    /** flag if we have finished */
    private boolean finished;
    
    private boolean cancelled;
    
    private Termination tm;
    
    /** listeners for the finish of task (TaskListener) */
    private HashSet<ActivityTaskListener> list;

    /** Add a listener to the task.
     * 
     * @param l the listener to add
     */
    public void addTaskListener(ActivityTaskListener l) {
        synchronized (this) {
            if (list == null) {
                list = new HashSet<>();
            }
            list.add(l);
        }
    }

    /** Remove a listener from the task.
    * @param l the listener to remove
    */
    public synchronized void removeTaskListener(ActivityTaskListener l) {
        if (list == null) {
            return;
        }

        list.remove(l);
    }
    
    public void notifyActivity(ActivityEvent event) {
        Iterator it;

        synchronized (this) {
            // fire the listeners
            if (list == null) {
                return;
            }

            it = ((HashSet) list.clone()).iterator();
        }

        while (it.hasNext()) {
            ActivityTaskListener l = (ActivityTaskListener) it.next();
            l.activityUpdate(event);
        }
    }
    
    /** Test whether the task is finished.
    */
    public final boolean isFinished() {
        synchronized (this) {
            return finished;
        }
    }

    /** Notify all listener that this task has finished.
    */
    public final void notifyFinished() {
        Iterator it;

        synchronized (this) {
            finished = true;
            
            // fire the listeners
            if (list == null) {
                return;
            }

            it = ((HashSet) list.clone()).iterator();
        }

        while (it.hasNext()) {
            ActivityTaskListener l = (ActivityTaskListener) it.next();
            l.taskFinished(this);
        }
    }

    public final boolean isCancelled() {
        synchronized (this) {
            return cancelled;
        }
    }

    @Override
    public boolean cancel() {
        synchronized (this) {
            cancelled = true;
            notifyFinished();
            return true;
        }
    }
    
    public void setTermination(Termination tm) {
        this.tm = tm;
        tm.addTerminationListener(new TerminationListener() {
            
            @Override
            public void terminateTask() {
                cancel();
            }
        });
    }
    
    public static class Termination implements Cancellable {
   
        private HashSet<TerminationListener> list;
        
        public Termination() {
        }

        private void addTerminationListener(TerminationListener l) {
            synchronized (this) {
                if (list == null) {
                    list = new HashSet<>();
                }
                list.add(l);
            }
        }

        @Override
        public boolean cancel() {
            Iterator it;
            synchronized (this) {
                it = ((HashSet) list.clone()).iterator();
            }

            while (it.hasNext()) {
                TerminationListener l = (TerminationListener) it.next();
                l.terminateTask();
            }
            return true;
        }

    }
    
    private static class TerminationListener implements EventListener {
        
        public void terminateTask() {
            
        }
    }
}
