/*
 *                  BioJava development code
 *
 * This code may be freely distributed and modified under the
 * terms of the GNU Lesser General Public Licence.  This should
 * be distributed with the code.  If you do not have a copy,
 * see:
 *
 *      http://www.gnu.org/copyleft/lesser.html
 *
 * Copyright for this code is held jointly by the individual
 * authors.  These should be listed in @author doc comments.
 *
 * For more information on the BioJava project and its aims,
 * or to join the biojava-l mailing list, visit the home page
 * at:
 *
 *      http://www.biojava.org/
 * 
 * Created on Jun 25, 2008
 * 
 */
package org.biobeans.demo.biojava.bj1.alignment;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.NoSuchElementException;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.demo.biojava.DemUtils;
import org.biojava.bio.BioException;
import org.biojava.bio.alignment.AlignmentAlgorithm;
import org.biojava.bio.alignment.AlignmentPair;
import org.biojava.bio.alignment.SmithWaterman;
import org.biojava.bio.alignment.SubstitutionMatrix;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojavax.RichObjectFactory;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;
import org.biojavax.bio.seq.RichSequence.IOTools;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.modules.InstalledFileLocator;

/**
 * @author Andreas Dr&auml;ger (draeger) <andreas.draeger@uni-tuebingen.de>
 * @date Jun 25, 2008
 * @since 1.7
 * @author Andreas Prlic
 */
public class SmithWatermanExample implements Runnable {

    private static final Logger LOGGER = Logger.getLogger(SmithWatermanExample.class.getName());

    /**
     * This method computes a pairwise local alignment between two given sequences
     * and prints the result on the standard output stream. The sequences must be
     * genbank files and the substitution matrix must be defined for the same
     * alphabet than both sequences.
     *
     * @param args
     *          query sequence file (genbank), subject sequence file (genbank),
     *          substitution matrix file

     */
    public void run() {
        String file1 = "alignment/NM_134326.gb";
        String file2 = "alignment/NM_000477.gb";
        String file3 = "alignment/nuc-4_4.txt";


        long start = System.currentTimeMillis();
        long maxMem = 0;

        FileObject oFile = DemUtils.getDemoFile("SmithWaterman.html");
        PrintWriter oOut = null;
        try {
            oOut = new PrintWriter(oFile.getOutputStream());
            oOut.println("<body>");
            oOut.println("BioJava performance example - Smith Waterman");
            oOut.println("<h1>BioJava performance example</h1>");
            oOut.println("Align two sequences of approx 2000 nucleotides length using Smith Waterman.<br><br> ");
            oOut.println("Reading sequence 1...<br> ");
            
            BufferedReader b1 = getReader(file1);

            RichSequenceIterator rsiQuery = IOTools.readGenbankDNA(b1,
                    RichObjectFactory.getDefaultNamespace());

            oOut.println("Reading sequence 2...<br> ");

            BufferedReader b2 = getReader(file2);

            RichSequenceIterator rsiSubject = IOTools.readGenbankDNA(b2,
                    RichObjectFactory.getDefaultNamespace());

            if (rsiQuery.hasNext() && rsiSubject.hasNext()) {
                RichSequence query = rsiQuery.nextRichSequence();
                RichSequence subject = rsiSubject.nextRichSequence();

                String matrix = getMatrxiFromFile(file3);

                oOut.println("Using matrix: <pre>"+matrix+"</pre><br>");
                oOut.println("Calculating SmithWaterman of <br>");
                StringBuffer txt = new StringBuffer();
                txt.append("query  : ").append(query.getDescription()).append(" (").append(query.getName()).append(")<br>");
                txt.append("subject: ").append(subject.getDescription()).append(" (").append(subject.getName()).append(")<br>");
                short match = 0;
                short replace = 5;
                short insert = 2;
                short delete = 2;
                short gapExtend = 1;
                txt.append("using penalties: match (").append(match).append("), replace (").append(replace).append(")," + " insert (").append(insert).append(" ) delete (").append(delete).append(")," + " gapExtend (").append(gapExtend).append(") <br>");
                oOut.println(txt);

                SubstitutionMatrix smax = new SubstitutionMatrix(
                        (FiniteAlphabet) query.getAlphabet(), matrix, file3);

                long mem0 = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
                if (mem0 > maxMem) {
                    maxMem = mem0;
                }

                AlignmentAlgorithm sa = new SmithWaterman(match, replace, insert, delete, gapExtend, smax);

                mem0 = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
                if (mem0 > maxMem) {
                    maxMem = mem0;
                }

                AlignmentPair pair = sa.pairwiseAlignment(query, subject);

                mem0 = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
                if (mem0 > maxMem) {
                    maxMem = mem0;
                }

                txt = new StringBuffer();
                txt.append("done...<br> <pre>");
                txt.append(pair.formatOutput(80));

                long time = System.currentTimeMillis() - start;

                mem0 = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
                if (mem0 > maxMem) {
                    maxMem = mem0;
                }

                txt.append("<pre>");

                txt.append("total processing time: ").append(time / 1000).append(" sec. " + "<br>");
                txt.append("mamimum memory: ").append(maxMem / 1024 / 1024).append(" MB" + "<br>");


                oOut.println(txt);
            }
            oOut.println("</body>");

        } catch (FileNotFoundException e) {
            LOGGER.log(Level.INFO, "", e);
        } catch (NoSuchElementException e) {
            LOGGER.log(Level.INFO, "", e);
        } catch (BioException e) {
            LOGGER.log(Level.INFO, "", e);
        } catch (IOException e) {
            LOGGER.log(Level.INFO, "", e);
        } catch (Exception e) {
            LOGGER.log(Level.INFO, "", e);
        } finally {
            oOut.flush();
            oOut.close();
        }
        DemUtils.viewFile(oFile);
    }

    private static BufferedReader getReader(String fileName) throws IOException {
        File file = InstalledFileLocator.getDefault().locate("files/"+fileName, "", true);
        FileObject oFile = FileUtil.toFileObject(file);
        BufferedReader reader = null;
        try {
            reader = new BufferedReader(new InputStreamReader(oFile.getInputStream()));
        } catch (FileNotFoundException ex) {
            LOGGER.log(Level.INFO, "", ex);
            throw new IOException("",ex);
        }
        return reader;
    }

    private static String getMatrxiFromFile(String fileName) throws IOException {
        BufferedReader reader = getReader(fileName);
        StringBuilder matString = new StringBuilder();

        while (reader.ready()) {
            matString.append(reader.readLine()).append(System.getProperty("line.separator"));
        }
        reader.close();
        return matString.toString();

    }
}
