/*
 *                    BioJava development code
 *
 * This code may be freely distributed and modified under the
 * terms of the GNU Lesser General Public Licence.  This should
 * be distributed with the code.  If you do not have a copy,
 * see:
 *
 *      http://www.gnu.org/copyleft/lesser.html
 *
 * Copyright for this code is held jointly by the individual
 * authors.  These should be listed in @author doc comments.
 *
 * For more information on the BioJava project and its aims,
 * or to join the biojava-l mailing list, visit the home page
 * at:
 *
 *      http://www.biojava.org/
 *
 */
package org.biobeans.demo.biojava.bj1.eventbasedparsing;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;
import java.util.*;
import java.io.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.demo.biojava.DemUtils;
import org.biojava.bio.program.sax.BlastLikeSAXParser;
import org.biojava.bio.program.blast2html.*;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.modules.InstalledFileLocator;

/**
 * Example application for converting a Blast - like SAX stream
 * to HTML.
 * 
 * Primary author 
 *                 Colin Hardman      (CAT)
 * Other authors  -
 *                 Tim Dilks          (CAT)
 *                 Simon Brocklehurst (CAT)
 *                 Stuart Johnston    (CAT)
 *                 Lawerence Bower    (CAT)
 *                 Derek Crockford    (CAT)
 *                 Neil Benn          (CAT)
 *
 * Copyright 2001 Cambridge Antibody Technology Group plc.
 *
 *
 * This code released to the biojava project, May 2001
 * under the LGPL license.
 *
 * @author Cambridge Antibody Technology Group plc
 * @version 1.0
 */
public class Blast2HTML implements Runnable {

    private static final Logger LOGGER = Logger.getLogger(Blast2HTML.class.getName());
    private static String oStyleDefinition =
            ".footer { font-family: Arial, Helvetica, sans-serif; font-size: 12px; font-style: italic; font-weight: normal ; color: #CC0000}\n.alignment {  font-family: \"Courier New\", Courier, mono; font-size: 14px;}\n.dbRetrieve {  font-family: Arial, Helvetica, sans-serif; font-size: 14px; font-style: normal; font-weight: bold; color: #000000; line-height: normal}\n.titleLevel1 { font-family: Arial, Helvetica, sans-serif; font-size: 30px; font-style: normal; font-weight: bold; color: #000000 ; background-color: #CCCCFF; line-height: normal}\n.titleLevel1Sub { font-family: Arial, Helvetica, sans-serif; font-size: 18px; font-style: normal; font-weight: bold; color: #000000 ; background-color: #CCCCFF; line-height: normal}\n.titleLevel2 { font-family: Arial, Helvetica, sans-serif; font-size: 18px; font-style: normal; font-weight: bold; color: #000000 ; background-color: #99FFCC; line-height: normal}\n.titleLevel3 { font-family: Arial, Helvetica, sans-serif; font-size: 16px; font-style: normal; font-weight: bold; color: #000000 ; background-color: #FFFFCC; line-height: normal; }\n.titleLevel3Sub { font-family: Arial, Helvetica, sans-serif; font-size: 12px; font-style: normal; font-weight: bold; color: #000000 ; background-color: #FFFFCC; line-height: normal}\n.summaryBodyLineOdd { font-family: Arial, Helvetica, sans-serif; font-size: 12px; font-style: normal; font-weight: normal; color: #000000 ; clip: rect( ); line-height: normal; text-indent: 0pt; padding-top: 0px; padding-right: 0px; padding-bottom: 0px; padding-left: 0px; margin-top: 0px; margin-right: 0px; margin-bottom: 0px; margin-left: 0px }\n.summaryBodyLineEven { font-family: Arial, Helvetica, sans-serif; font-size: 12px; font-style: normal; font-weight: normal; color: #000000 ; background-color: #EEEEEE; line-height: normal }\n.titleLevel4 { font-family: Arial, Helvetica, sans-serif; font-size: 14px; font-style: normal; font-weight: bold; color: #000000 ; background-color: #FFFFEE; line-height: normal }\nbody {  font-family: Arial, Helvetica, sans-serif; font-size: 10px}\ntd {  font-family: Arial, Helvetica, sans-serif; font-size: 16px}\na {  text-decoration: none}\n";

    public static HTMLRenderer configureRenderer(String poType,
            PrintWriter poOut) {

        if (poType.equals("nucleic")) {
            return Blast2HTML.configureBlastN(poOut);
        } else {
            return Blast2HTML.configureBlastP(poOut);
        }
    }

    public static HTMLRenderer configureBlastN(PrintWriter poOut) {

        SimpleAlignmentStyler oStyler = new SimpleAlignmentStyler(SimpleAlignmentStyler.SHOW_ALL);
        String oRed = "FFA2A2";
        oStyler.addStyle("-", oRed);
        oStyler.addStyle("N", oRed);
        oStyler.addStyle("A", oRed);
        oStyler.addStyle("T", oRed);
        oStyler.addStyle("C", oRed);
        oStyler.addStyle("G", oRed);

        AlignmentMarker oAlignmentMarker = new AlignmentMarker(new ColourCommand() {

            /**
             * Highlight mismatches
             * NOTE: assuming the chars are the same case.
             */
            public boolean isColoured(String poFirst, String poSecond) {

                if (poFirst.equals(poSecond)) {
                    return false;
                } else {
                    return true;
                }
            }
        } // end ColourCommand
                , oStyler //	      new BlastMatrixAlignmentStyler()
                );

        Properties oProps = new Properties();
        oProps.put("db", "nucl");

        HTMLRenderer oRenderer = new HTMLRenderer(poOut,
                Blast2HTML.oStyleDefinition,
                50,
                new DefaultURLGeneratorFactory(),
                oAlignmentMarker,
                oProps);

        return oRenderer;
    }

    public static HTMLRenderer configureBlastP(PrintWriter poOut) {

        SimpleAlignmentStyler oStyler = new SimpleAlignmentStyler(SimpleAlignmentStyler.SHOW_SAME);

        oStyler.addStyle("A", "C8FFC8");
        oStyler.addStyle("C", "C8FFC8");
        oStyler.addStyle("L", "C8FFC8");
        oStyler.addStyle("I", "C8FFC8");
        oStyler.addStyle("V", "C8FFC8");
        oStyler.addStyle("M", "C8FFC8");
        oStyler.addStyle("G", "DCC0FF");
        oStyler.addStyle("P", "DCC0FF");
        oStyler.addStyle("S", "FFFCA0");
        oStyler.addStyle("T", "FFFCA0");
        oStyler.addStyle("N", "FFFCA0");
        oStyler.addStyle("Q", "FFFCA0");
        oStyler.addStyle("K", "FFA2A2");
        oStyler.addStyle("R", "FFA2A2");
        oStyler.addStyle("D", "A2E2FF");
        oStyler.addStyle("E", "A2E2FF");
        oStyler.addStyle("H", "50FF50");
        oStyler.addStyle("Y", "50FF50");
        oStyler.addStyle("W", "50FF50");
        oStyler.addStyle("F", "50FF50");

        AlignmentMarker oAlignmentMarker = new AlignmentMarker(new ColourCommand() {

            /**
             * Highlight mismatches
             * NOTE: assuming the chars are the same case.
             */
            public boolean isColoured(String poFirst, String poSecond) {

                if (poFirst.equals(poSecond)) {
                    return false;
                } else {
                    return true;
                }
            }
        } // end ColourCommand
                , oStyler //	      new BlastMatrixAlignmentStyler()
                );

        Properties oProps = new Properties();
        oProps.put("db", "Protein");

        HTMLRenderer oRenderer = new HTMLRenderer(poOut,
                Blast2HTML.oStyleDefinition,
                50,
                new DefaultURLGeneratorFactory(),
                oAlignmentMarker,
                oProps);

        return oRenderer;
    }

    /**
     * 
     *
     */
    public void run() {

        // output = blast2html.html
        FileObject oFile = DemUtils.getDemoFile("blast2html.html");
        PrintWriter oOut = null;
        try {
            oOut = new PrintWriter(oFile.getOutputStream());
        } catch (IOException ex) {
            LOGGER.log(Level.INFO, "", ex);
            return;
        }

        HTMLRenderer oRenderer = null;
        // Only nucleic renderer tested, see original demo for "protein"
        oRenderer = Blast2HTML.configureRenderer("nucleic", oOut);

        String oInput = "ncbiblast/blastn.out";
        File iFile = InstalledFileLocator.getDefault().locate("files/" + oInput, "", true);
        FileObject ioFile = FileUtil.toFileObject(iFile);

        /**
         * Create a SAX Parser
         */
        XMLReader oParser = (XMLReader) new BlastLikeSAXParser();
        ((BlastLikeSAXParser) oParser).setModeLazy();

        ContentHandler oHandler =
                (ContentHandler) new Blast2HTMLHandler(oRenderer);

        /*
         * Give the parser a reference to the ContentHandler
         * so that it can send SAX2 mesagges.
         */
        oParser.setContentHandler(oHandler);

        //Parsing using ByteSteam as InputSource
        // Open file and read all lines from file sequentially
        InputStream oInputStream;

        try {
            // create input stream
            oInputStream = ioFile.getInputStream();

            oOut.println("<HTML>\n<HEAD>");
            oOut.println(oRenderer.getHeaderDefinitions());
            oOut.println("</HEAD>");
            oOut.println("<body bgcolor=\"#FFFFFF\" alink=\"#33FFFF\" vlink=\"#CCCC99\">");
            oParser.parse(new InputSource(oInputStream));

            oOut.println("</BODY>\n</HTML>");

        } catch (SAXException ex) {
            LOGGER.log(Level.INFO, "", ex);
        } catch (java.io.FileNotFoundException x) {
            LOGGER.log(Level.INFO, "", x);
            return;
        } catch (IOException ex) {
            LOGGER.log(Level.INFO, "", ex);
        } finally {
            oOut.flush();
            oOut.close();
        }

        DemUtils.viewFile(oFile);
    }
}
