/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.activity.onlineactivities.ncbi;

import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.io.GetStreamProcessor;
import org.biobeans.modules.bistro.io.GetStreamTransmitter;
import org.biobeans.modules.bistro.io.PostStreamProcessor;
import org.biobeans.modules.bistro.io.PostStreamTransmitter;
import org.biobeans.modules.bistro.io.StreamProcessor;
import org.biobeans.modules.bistro.io.StreamTransmitter;
import org.biobeans.modules.bistro.io.TransmitterException;
import org.biobeans.modules.bistro.io.XSLTStreamProcessor;
import org.biobeans.modules.bistro.task.TaskExecutionException;
import org.biobeans.modules.bistro.activity.AbstractActivity;
import org.biobeans.modules.bistro.activity.ActivityConfigurationException;
import org.biobeans.modules.bistro.activity.SimpleOutputPort;

/**
 * This class provides the basis for all NCBI EFetch Tasks.
 *
 * @author Mark
 */
abstract class AbstractEFetchActivity<ConfigType> extends AbstractActivity<ConfigType> {

    private static final Logger LOGGER = Logger.getLogger(AbstractEFetchActivity.class.getName());

    public static final int GET_TRANSMITTER = 1;

    public static final int POST_TRANSMITTER = 2;
    
    /**
     * The return format. Values include (xml|html|text|asn.1)
     */
    protected String retmode;

    /**
     * Defines the output types based on database native Default format for
     * viewing sequences. fasta FASTA view of a sequence. gb GenBank view for
     * sequences, constructed sequences will be shown as contigs (by pointing to
     * its parts). Valid for nucleotides. gbc INSDSeq structured flat file for
     * Nucleotides. gbwithparts GenBank view for sequences, the sequence will
     * always be shown. Valid for Nucleotides est EST Report. Valid for
     * sequences from dbEST database. gss GSS Report. Valid for sequences from
     * dbGSS database. gp GenPept view. Valid for Proteins. gpc INSDSeq
     * structured flat file for Proteins. seqid To convert list of gis into list
     * of seqids. acc To convert list of gis into list of accessions. chr SNP
     * Chromosome Report. flt SNP Flat File report. rsr SNP RS Cluster report.
     * brief SNP ID list. docset SNP RS summary.
     */
    protected String rettype;

    /**
     * sequential number of the first id retrieved - default=0 which will
     * retrieve the first record)
     */
    protected String retstart = "0";

    /**
     * The number of items to be retrieved
     */
    protected String retmax = "1";

    protected String id;

    /**
     * The NCBI database to be queried. Values include genome nucleotide protein
     * popset snp sequences - Composite name including nucleotide, protein,
     * popset and genome.
     */
    protected String db;

    protected String url;

    public static final String NEWLINE = System.getProperty("line.separator");

    protected StreamProcessor streamProcessor = null;

    public AbstractEFetchActivity() {
        super();
    }
    
    protected void configureEFetchActivity(EFetchActivityConfigurationBean configbean)
            throws ActivityConfigurationException {
        removeInputPorts();
        removeOutputPorts();
        addOutputPort(new SimpleOutputPort("outputText","text/plain",0,this));
        addOutputPort(new SimpleOutputPort("id","text/plain",0,this));
        this.url = configbean.getURL();
        this.db = configbean.getDatabase();
        this.retmode = configbean.getRetmode();
        this.rettype = configbean.getRettype();
    }
    
    /**
     * This method is responsible for transmitting a map of values, and
     * processing the results
     *
     * @param transmitterMap
     * @param streamProcessor
     * @throws MalformedURLException
     * @throws TransmitterException
     */
    protected Map<String, DataContainer> transmit(Map<String, String> transmitterMap) throws MalformedURLException, TransmitterException {

        // initialise the transmitter
        int transmitterType = -1;
        String tType = transmitterMap.get("transmitterType");
        transmitterType = (tType == null) ? GET_TRANSMITTER : Integer.parseInt(tType);
        StreamTransmitter transmitter = getInstance(transmitterType);
        transmitter.setURL(url);

        String outputFile = transmitterMap.get("outputFile");
        String xslt = transmitterMap.get("xslt");
        String _retmode = transmitterMap.get("retmode");

        LOGGER.log(Level.FINE, "xslt: {0}", xslt);
        if (_retmode != null && _retmode.equalsIgnoreCase("xml")) {
            if (xslt == null) {
                streamProcessor = new GetStreamProcessor();
            } else {
                LOGGER.fine("Using XSLT Stream Processor");
                if (outputFile == null) {
                    throw new TransmitterException("The outputFile attribute was null");
                }
                Map<String, DataContainer> outputMap = new HashMap<>();
                streamProcessor = new XSLTStreamProcessor(outputMap, outputFile, xslt);
            }
        } else// otherwise treat it as plain text.
        {
            if (transmitterType == GET_TRANSMITTER) {
                streamProcessor = new GetStreamProcessor();
            } else {
                streamProcessor = new PostStreamProcessor();
            }
        }

        return transmitter.transmit(transmitterMap, streamProcessor);
    }

    /**
     * This method gets an instance of the stream transmitter.
     *
     * @param transmitterType
     * @return
     */
    protected StreamTransmitter getInstance(int transmitterType) {
        StreamTransmitter transmitter = null;
        switch (transmitterType) {
            case GET_TRANSMITTER: {
                transmitter = new GetStreamTransmitter();
                break;
            }
            case POST_TRANSMITTER: {
                transmitter = new PostStreamTransmitter();
                break;
            }
        }
        return transmitter;
    }

    @Override
    protected abstract Map<String, DataContainer> execute(Map<String, DataContainer> inputMap) throws TaskExecutionException;

}
