/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
*/
package org.biobeans.modules.bistro.activity.onlineactivities.ncbi;

import java.net.MalformedURLException;
import java.util.Map;
import java.util.logging.Logger;
import org.biobeans.modules.bistro.activity.AbstractActivity;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.io.GetStreamTransmitter;
import org.biobeans.modules.bistro.io.StreamProcessor;
import org.biobeans.modules.bistro.io.StreamTransmitter;
import org.biobeans.modules.bistro.io.TransmitterException;
import org.biobeans.modules.bistro.task.TaskExecutionException;
import org.biobeans.modules.bistro.activity.SimpleInputPort;
import org.biobeans.modules.bistro.activity.SimpleOutputPort;

/**
 * This class provides basic common functionality needed by all NCBI activities.
 * 
 * Help: http://www.ncbi.nlm.nih.gov/books/NBK25501/
 *
 * @author mfortner
 */
abstract class AbstractNCBIActivity<ConfigType> extends AbstractActivity<ConfigType> {

    private static final Logger LOGGER = Logger.getLogger(AbstractNCBIActivity.class.getName());
    
    protected StreamProcessor streamProcessor = null;

    protected String rettype;
    protected String id;
    protected String queryKey;
    protected String displayOption;
    protected String originalDb;
    protected int maxRecords = 1;
    protected String startTag;
    protected String endTag;
    protected String url = "https://www.ncbi.nlm.nih.gov/entrez/query.fcgi";
    protected String cmd;
    protected String cmdCurrent;

    public AbstractNCBIActivity () {
        initialize();
    }
    
    private void initialize() {
        addInputPort(new SimpleInputPort("id","text/plain",0,this));
        addInputPort(new SimpleInputPort("maxRecords","text/plain",0,this));
        addInputPort(new SimpleInputPort("xslt","text/plain",0,this));
        addOutputPort(new SimpleOutputPort("resultsXml","text/xml",0,this));
    }
    
    @Override
    public abstract Map<String, DataContainer> execute(Map<String, DataContainer> inputMap) throws TaskExecutionException;

    /**
     * This method is responsible for transmitting a map of values, and
     * processing the results
     *
     * @param transmitterMap
     * @param streamProcessor
     * @throws MalformedURLException
     * @throws TransmitterException
     */
    protected Map<String, DataContainer> transmit(Map<String, String> transmitterMap) throws MalformedURLException, TransmitterException {
        StreamTransmitter transmitter = new GetStreamTransmitter();
        transmitter.setURL(this.url);
        String xslt = transmitterMap.get("xslt");
        if (xslt == null || xslt.equals("")) {
            streamProcessor = new NCBIXMLStreamProcessor(this.startTag, this.endTag);

        } else {
            LOGGER.fine("Using XSLT Stream Processor");
            streamProcessor = new NCBIXSLTStreamProcessor(this.startTag, this.endTag, xslt);
        }

        return transmitter.transmit(transmitterMap, streamProcessor);
    }

    public void setStreamProcessor(StreamProcessor streamProc) {
        this.streamProcessor = streamProc;
    }
}
