/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
*/
package org.biobeans.modules.bistro.activity.onlineactivities.ncbi;

import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.Map;
import org.biobeans.modules.bistro.activity.ActivityConfigurationException;
import org.biobeans.modules.bistro.data.DataContainer;
import org.biobeans.modules.bistro.data.DataContainerAdapter;
import org.biobeans.modules.bistro.io.TransmitterException;
import org.biobeans.modules.bistro.task.TaskExecutionException;

/**
 * This processor is responsible for fetching an Entrez Gene record in XML
 * format. It can also transform the resulting XML document.
 *
 * @tavinput term The search term (usually a gene symbol).
 * @tavinput maxRecords The maximum number of records to be returned.
 * @tavinput xslt A complete path to the XSLT used to transform the results. (optional)
 *
 * @tavoutput resultsXml A string containing the resultant XML.
 *
 * @author mfortner
 */
public class EntrezGeneActivity extends AbstractNCBIActivity<EntrezGeneActivityConfigurationBean> {

    private EntrezGeneActivityConfigurationBean configbean;
    
    public EntrezGeneActivity() {
        this.url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi";
        this.startTag = "<Entrezgene>";
        this.endTag = "</Entrezgene>";
        this.originalDb = "gene";
    }

    @Override
    public void configure(EntrezGeneActivityConfigurationBean conf) throws ActivityConfigurationException {
        configbean = conf;
    }

    @Override
    public EntrezGeneActivityConfigurationBean getConfiguration() {
        return configbean;
    }

    @Override
    public Map<String, DataContainer> execute(Map<String, DataContainer> inputMap) throws TaskExecutionException {
        //this.inputMap = inputMap;
        DataContainerAdapter inAdapter = new DataContainerAdapter(inputMap);

        this.id = inAdapter.getString("id");
        String maxRecs = inAdapter.getString("maxRecords");

        this.maxRecords = (maxRecs == null) ? 1 : Integer.parseInt(maxRecs);

        Map<String, String> transmitterMap = new HashMap<>();
        transmitterMap.put("db", this.originalDb);
        transmitterMap.put("retmode", "xml");
        transmitterMap.put("id", this.id);

        Map<String, DataContainer> outputMap = new HashMap<>();
        try {
            outputMap = this.transmit(transmitterMap);
        } catch (MalformedURLException | TransmitterException e) {
            throw new TaskExecutionException(e);
        }
        outputMap.put("id", inputMap.get("id")); 
        return outputMap;
    }

}
