/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
 */
package org.biobeans.modules.bistro.activity.onlineactivities.ncbi;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import org.biobeans.modules.bistro.io.StreamProcessor;
import org.biobeans.modules.bistro.data.DataContainer;

/**
 * This class cleans up NCBI pseudoXML embedded in HTML pages, and turns it into
 * valid XML.
 *
 * @author mfortner
 */
class NCBIXMLStreamProcessor extends ReplacementStreamProcessor implements StreamProcessor {

    private static Logger LOGGER = Logger.getLogger(NCBIXMLStreamProcessor.class.getName());

    protected String startTag;

    protected String endTag;

    public NCBIXMLStreamProcessor() {
    }

    public NCBIXMLStreamProcessor(String startTag, String endTag) {
        this.oldText = new String[]{"&lt;", "&gt;", "&quot;", "<b>", "</b>", "<font color=\"#C00000\">", "</font>",
            "<font color=\"#4040FF\">", "<font color=\"#C00000\">", "<font color=\"#C04000\">",
            "<font color=\"#FF0000\">", "&amp;amp;"};
        this.replacementText = new String[]{"<", ">", "\"", "", "", "", "", "", "", "", "", "&amp;"};
        this.setOldText(oldText);
        this.setReplacementText(replacementText);
        this.startTag = startTag;
        this.endTag = endTag;
    }

    /**
     * This method processes a stream of text from a URL endpoint.
     *
     * @param stream
     */
    @Override
    public Map<String, DataContainer> processStream(InputStream stream) throws IOException {
        BufferedReader in = new BufferedReader(new InputStreamReader(stream), this.bufferSize);
        String str;
        String lineEnding = System.getProperty("line.separator");
        StringWriter sw = new StringWriter();
        BufferedWriter out = new BufferedWriter(sw, this.bufferSize);
        String[] oldText = this.getOldText();
        String[] replacementText = this.getReplacementText();
        out.write("<?xml version=\"1.0\"?>");
        out.write("<searchResults>");
        boolean startTagFound = false;
        boolean endTagFound = false;


        while ((str = in.readLine()) != null) {
            int startIndex = str.indexOf(startTag);
            int endIndex = str.indexOf(endTag);

            startTagFound = (startIndex != -1) || startTagFound;

            if (!startTagFound) {
                continue;
            }
            endTagFound = (endIndex != -1);
            if (startTagFound) {

                if (startIndex != -1 && str.length() > startTag.length()) {
                    str = str.substring(startIndex);
                }
            }

            if (endTagFound) {
                if (endIndex == 0) {
                    str = endTag;
                } else if (endIndex > 0) {
                    str = str.substring(0, endIndex);
                }
                startTagFound = false;
                endTagFound = false;
            }

            out.write(str);
            out.write(lineEnding);
        }
        out.write("</searchResults>");
        out.close();
        String outStr = sw.toString();

        // now that we have all of the raw text,
        // replace all of the escaped characters with real characters
        // in order to make the XML parseable
        for (int i = 0; i < oldText.length; i++) {
            outStr = outStr.replaceAll(oldText[i], replacementText[i]);
        }

        Map<String, DataContainer> outputMap = new HashMap<>();
        outputMap.put("resultsXml", new DataContainer(outStr));
        return outputMap;
    }

    /**
     * @return Returns the endTag.
     */
    public String getEndTag() {
        return endTag;
    }

    /**
     * @param endTag The endTag to set.
     */
    public void setEndTag(String endTag) {
        this.endTag = endTag;
    }

    /**
     * @return Returns the startTag.
     */
    public String getStartTag() {
        return startTag;
    }

    /**
     * @param startTag The startTag to set.
     */
    public void setStartTag(String startTag) {
        this.startTag = startTag;
    }

}
